/****************************************************************

  Cluster.c

  Ismo Krkkinen

  Last changed 2.11.2001

****************************************************************/

#define FACTFILE "cluster.fac"
#include "parametr.c"
#include "swis_cb.h"
#include <windows.h>
#include "swi_util.h"
#include "swis_ImageView.h"
#include "ls.h"
#include "interfc.h"
#include "memctrl.h"
#include <stdio.h>
#include "sortcb.h"
#include "random.h"
#include "cef.h"

/* global vars */
SWI_ID CBClassID, TSClassID, PAClassID, GClassID;

static int NewDataSetIndex = 1;

// add by Jian Pu, 7,6,2007
// for Stepwise K-means
CODEBOOK* g_CB = NULL;


/* brute force local search */

int PreBFLocalSearch(int InputCount, SWI_ID* InputIDs, void** InputExternals,
	int OutputCount, SWI_ID* OutputIDs, void** OutputExternals, void** Aux) 
{
	TRAININGSET* TS;
	*Aux = (void*)1;
	TS = (TRAININGSET*) InputExternals[0];
	if (Value(MaxBookSize) == 0) SetValue(MaxBookSize, Value(MinBookSize));
	if (BookSize(TS) < Value(MaxBookSize)) {
		ErrorMessage(
			"Data set size (%i) is less than "
			"requested largest number of centroids (%i).",
			BookSize(TS), Value(MaxBookSize));
		return 1;
	}
	if (Value(MaxBookSize) < Value(MinBookSize)) {
		ErrorMessage(
			"Maximum number of centroids (%i) is less than "
			"minimum number of centroids (%i).",
			Value(MaxBookSize), Value(MinBookSize));
		return 2;
	}
	if (Value(EvaluationFunction) == 0) {
		if (TS->MinValue != 0 && TS->MaxValue != 1) {
			ErrorMessage(
				"Currently the stochastic complexity evaluation function"
				"is only for binary data.");
			return 3;
		}
	}
	return 0;
}

int MainBFLocalSearch(int InputCount, void** InputExternals, 
	int OutputCount, void** OutputExternals, void** Aux)
{
	int ReturnValue;
	TRAININGSET* TSIn;
	CODEBOOK* CBOut;
	PARTITIONING* PAOut;
	TSIn = (TRAININGSET*) InputExternals[0];
	OutputExternals[0] = CBOut = (CODEBOOK*) new_swiCODEBOOK();
	CreateNewCodebook(CBOut, Value(MinBookSize), TSIn);
	OutputExternals[1] = PAOut = (PARTITIONING*) new_swiPARTITIONING(0);
	CreateNewPartitioning(PAOut, TSIn, Value(MinBookSize));
	ReturnValue = BruteForce(TSIn, CBOut, PAOut, Value(EvaluationFunction),
		Value(MinBookSize), Value(MaxBookSize), Value(BFLSQuietLevel),
		Value(SwappingMethod), Value(GLAIterations), Value(LSIterations),
		0, 0, 0);
	/*ReturnValue = IterateFromMinToMaxByLocalSearch(TSIn, CBOut, PAOut, 
		Value(EvaluationFunction), Value(MinBookSize), Value(MaxBookSize), 
		Value(BFLSQuietLevel), Value(SwappingMethod), 
		Value(GLAIterations), Value(LSIterations), 0);*/
	if (!ReturnValue) *Aux = 0;
	else {
		FreeCodebook(CBOut);
		delete_swiCODEBOOK((swiCODEBOOK*) OutputExternals[0]);
		OutputExternals[0] = 0;
		FreePartitioning(PAOut);
		delete_swiPARTITIONING((swiPARTITIONING*) OutputExternals[1]);
		OutputExternals[1] = 0;
	}
	return ReturnValue;
}


int PostBFLocalSearch(int InputCount, SWI_ID* InputIDs, void** InputExternals,
	int OutputCount, SWI_ID* OutputIDs, void** OutputExternals, void** Aux) 
{
	int ndups;
	char Name[20];
	if (*Aux) return 1;
	ndups = DuplicatesInCodebook((CODEBOOK*) OutputExternals[0]);
	if (ndups && Value(BFLSQuietLevel)>=2)
		PrintMessage("WARNING: Final CB contains %i duplicates.\n", ndups);
	SortCodebook((CODEBOOK*) OutputExternals[0], DATA_DESCENDING);
	sprintf(Name, "New%i.cb", NewDataSetIndex);
	OutputIDs[0] = swiDSNew(CBClassID, OutputExternals[0], Name);
	if (Value(SavePartition)) {
		SetswiPARTITIONING_TS_ID((swiPARTITIONING*) OutputExternals[1], InputIDs[0]);
		sprintf(Name, "New%i.pa", NewDataSetIndex);
		OutputIDs[1] = swiDSNew(PAClassID, OutputExternals[1], Name);
	} else {
		delete_swiPARTITIONING((swiPARTITIONING*) OutputExternals[1]);
		OutputExternals[1] = 0;
	}
	NewDataSetIndex++;
	return 0;
}


static int Params[8] = {
	_BFLSQuietLevel, _SavePartition, _MinBookSize, _MaxBookSize, 
	_LSIterations, _SwappingMethod, _GLAIterations, _EvaluationFunction
};

static SWI_ID swisLocalSearchFunction(SWI_ID TrainingSetID, 
	SWI_ID CodebookID, SWI_ID PartitioningID, SWI_ID AfterProcessInMenu)
{
	SWI_ID InputReq, OutputReq, Process, Options;
	InputReq = swiRQNew();
	swiRQAdd(InputReq, TrainingSetID, 1);
	OutputReq = swiRQNew();
	swiRQAdd(OutputReq, CodebookID, 1);
	swiRQAdd(OutputReq, PartitioningID, 1);
	Options = swiPSNew(8, Params);
	Process = swiPCNew(0, "Local search (Brute Force)", 
		&PreBFLocalSearch, &MainBFLocalSearch, &PostBFLocalSearch,
		Options, InputReq, OutputReq);
	swiPCSetMenuInfo(Process, 'B', 1, 1, 0, 0, AfterProcessInMenu);
	return Process;
}


int MainSWLocalSearch(int InputCount, void** InputExternals, 
	int OutputCount, void** OutputExternals, void** Aux)
{
	int ReturnValue;
	TRAININGSET* TSIn;
	CODEBOOK* CBOut;
	PARTITIONING* PAOut;
	TSIn = (TRAININGSET*) InputExternals[0];
	OutputExternals[0] = CBOut = (CODEBOOK*) new_swiCODEBOOK();
	CreateNewCodebook(CBOut, Value(MinBookSize), TSIn);
	OutputExternals[1] = PAOut = (PARTITIONING*) new_swiPARTITIONING(0);
	CreateNewPartitioning(PAOut, TSIn, Value(MinBookSize));
	ReturnValue = Stepwise(TSIn, CBOut, PAOut, Value(EvaluationFunction),
		Value(MinBookSize), Value(MaxBookSize), Value(BFLSQuietLevel),
		Value(SwappingMethod), Value(GLAIterations),
		Value(LSIterations), Value(LSStepwiseIterations),
		0, 0, Value(SearchDirectionDecreasing), 0);
	/*ReturnValue = IterateFromMinToMaxByLocalSearch(TSIn, CBOut, PAOut, 
		Value(EvaluationFunction), Value(MinBookSize), Value(MaxBookSize), 
		Value(BFLSQuietLevel), Value(SwappingMethod), 
		Value(GLAIterations), Value(LSIterations), 0);*/
	if (!ReturnValue) *Aux = 0;
	else {
		FreeCodebook(CBOut);
		delete_swiCODEBOOK((swiCODEBOOK*) OutputExternals[0]);
		OutputExternals[0] = 0;
		FreePartitioning(PAOut);
		delete_swiPARTITIONING((swiPARTITIONING*) OutputExternals[1]);
		OutputExternals[1] = 0;
	}
	return ReturnValue;
}


static int StepwiseParams[10] = {
	_BFLSQuietLevel, _SavePartition, _MinBookSize, _MaxBookSize, 
	_LSIterations, _SwappingMethod, _GLAIterations, _EvaluationFunction,
	_LSStepwiseIterations, _SearchDirectionDecreasing
};

static SWI_ID swisStepwiseFunction(SWI_ID TrainingSetID, 
	SWI_ID CodebookID, SWI_ID PartitioningID, SWI_ID AfterProcessInMenu)
{
	SWI_ID InputReq, OutputReq, Process, Options;
	InputReq = swiRQNew();
	swiRQAdd(InputReq, TrainingSetID, 1);
	OutputReq = swiRQNew();
	swiRQAdd(OutputReq, CodebookID, 1);
	swiRQAdd(OutputReq, PartitioningID, 1);
	Options = swiPSNew(10, StepwiseParams);
	Process = swiPCNew(0, "Local Search (Stepwise)", 
		&PreBFLocalSearch, &MainSWLocalSearch, &PostBFLocalSearch,
		Options, InputReq, OutputReq);
	swiPCSetMenuInfo(Process, 'W', 1, 1, 0, 0, AfterProcessInMenu);
	return Process;
}


/* miscellaneous */

int PreGenerateOptimalPartitioning(int InputCount, SWI_ID* InputIDs, 
	void** InputExternals, int OutputCount, SWI_ID* OutputIDs, 
	void** OutputExternals, void** Aux) 
{
	TRAININGSET* TS;
	CODEBOOK* CB;
	*Aux = (void*)1;
	TS = InputExternals[0];
	CB = InputExternals[1];
	if (VectorSize(TS) != VectorSize(CB)) {
		ErrorMessage("Vector sizes don't match, %i for data set, "
			"%i for centroids.\n",
			VectorSize(TS), VectorSize(CB));
		return 1;
	}
	if (TS->MinValue != CB->MinValue || TS->MaxValue != CB->MaxValue) {
		ErrorMessage(
			"Scalar ranges don't match, for data set it is [%i, %i]"
			" and for centroids [%i, %i].\n",
			TS->MinValue, TS->MaxValue, CB->MinValue, CB->MaxValue);
		return 3;
	}
	switch (Value(ErrorFunction)) {
	case efMSE: break;
	case efSC: break;
	case efJACCARD: break;
	case efDICE: break;
	default: 
		ErrorMessage("Unknown error metric (%i)", Value(ErrorFunction));
		return 2;
	}
	return 0;
}

int MainGenerateOptimalPartitioning(int InputCount, void** InputExternals, 
	int OutputCount, void** OutputExternals, void** Aux)
{
	TRAININGSET* TS;
	PARTITIONING* PA;
	CODEBOOK* CB;
	int ErrorFunctionType;
	TS = InputExternals[0];
	CB = InputExternals[1];
	PA = OutputExternals[0] = new_swiPARTITIONING(0);
	CreateNewPartitioning(PA, TS, BookSize(CB));
	switch (Value(ErrorFunction)) {
	case efMSE: ErrorFunctionType = MSE; break;
	case efSC: ErrorFunctionType = SC; break;
	case efJACCARD: ErrorFunctionType = JACCARD; break;
	case efDICE: ErrorFunctionType = DICE; break;
	}
	GenerateOptimalPartitioningGeneral(TS, CB, PA, ErrorFunctionType);
	*Aux = 0;
	return 0;
}

int PostGenerateOptimalPartitioning(int InputCount, SWI_ID* InputIDs,
	void** InputExternals, int OutputCount, SWI_ID* OutputIDs, 
	void** OutputExternals, void** Aux) 
{
	char foo[40];
	if (*Aux) return 1;
	SetswiPARTITIONING_TS_ID(OutputExternals[0], InputIDs[0]);
	sprintf(foo, "New%i.pa", NewDataSetIndex++);
	OutputIDs[0] = swiDSNew(PAClassID, OutputExternals[0], foo);
	return 0;
}

static GOPParams[1] = { _ErrorFunction };

static SWI_ID swisGenerateOptimalPartitioning(SWI_ID AfterProcessInMenu) {
	SWI_ID In, Out, Options, ID;
	In = swiRQNew();
	swiRQAdd(In, TSClassID, 1);
	swiRQAdd(In, CBClassID, 1);
	Out = swiRQNew();
	swiRQAdd(Out, PAClassID, 1);
	Options = swiPSNew(1, GOPParams);
	ID = swiPCNew(0, "Optimal partition",
		&PreGenerateOptimalPartitioning,
		&MainGenerateOptimalPartitioning,
		&PostGenerateOptimalPartitioning,
		Options, In, Out);
	swiPCSetMenuInfo(ID, 'P', 1, 1, 0, 0, AfterProcessInMenu);
	return ID;
}

// added by Jian Pu, 30,5,2007
int PrePreformCompleteKMeans(int InputCount, SWI_ID* InputIDs, 
	void** InputExternals, int OutputCount, SWI_ID* OutputIDs, 
	void** OutputExternals, void** Aux) 
{
	TRAININGSET* TS;
	CODEBOOK* CB;
	*Aux = (void*)1;
	TS = InputExternals[0];
	CB = InputExternals[1];
	if (VectorSize(TS) != VectorSize(CB)) {
		ErrorMessage("Vector sizes don't match, %i for data set, "
			"%i for centroids.\n",
			VectorSize(TS), VectorSize(CB));
		return 1;
	}
	if (TS->MinValue != CB->MinValue || TS->MaxValue != CB->MaxValue) {
		ErrorMessage(
			"Scalar ranges don't match, for data set it is [%i, %i]"
			" and for centroids [%i, %i].\n",
			TS->MinValue, TS->MaxValue, CB->MinValue, CB->MaxValue);
		return 3;
	}
	switch (Value(ErrorFunction)) {
	case efMSE: break;
	case efSC: break;
	case efJACCARD: break;
	case efDICE: break;
	default: 
		ErrorMessage("Unknown error metric (%i)", Value(ErrorFunction));
		return 2;
	}
	return 0;
}

// added by Jian Pu, 30,5,2007
int MainPreformCompleteKMeans(int InputCount, void** InputExternals, 
	int OutputCount, void** OutputExternals, void** Aux)
{
	TRAININGSET* TS;
	CODEBOOK* iCB;
	CODEBOOK* oCB;
	PARTITIONING* oPA;
	int ErrorFunctionType;
	
	TS = InputExternals[0];
	iCB = InputExternals[1];
	oCB = OutputExternals[0] = new_swiCODEBOOK();
	oPA = OutputExternals[1] = new_swiPARTITIONING(0);

	switch (Value(ErrorFunction)) {
	case efMSE: ErrorFunctionType = MSE; break;
	case efSC: ErrorFunctionType = SC; break;
	case efJACCARD: ErrorFunctionType = JACCARD; break;
	case efDICE: ErrorFunctionType = DICE; break;
	}

	CreateNewPartitioning(oPA, TS, BookSize(iCB));
	GenerateOptimalPartitioningGeneral(TS, iCB, oPA, ErrorFunctionType);

	CreateNewCodebook(oCB, oPA->PartitionCount, TS);
	GenerateOptimalCodebook(TS, oCB, oPA);

	iCB = new_swiCODEBOOK();
	CreateNewCodebook(iCB, oPA->PartitionCount, TS);
	do{
		// to swap pointers
		CODEBOOK* tmp = iCB;
		iCB = oCB;
		oCB = tmp;
		
		GenerateOptimalPartitioningGeneral(TS, iCB, oPA, ErrorFunctionType);
		GenerateOptimalCodebook(TS, oCB, oPA);
	}while( !swi_EqualCodeBook(iCB, oCB) );
	*Aux = 0;
	return 0;
}

// added by Jian Pu, 30,5,2007
int PostPreformCompleteKMeans(int InputCount, SWI_ID* InputIDs,
	void** InputExternals, int OutputCount, SWI_ID* OutputIDs, 
	void** OutputExternals, void** Aux) 
{
	char foo[40];
	if (*Aux) return 1;

	sprintf(foo, "New%i.cb", NewDataSetIndex++);
	OutputIDs[1] = swiDSNew(CBClassID, OutputExternals[0], foo);

	SetswiPARTITIONING_TS_ID(OutputExternals[1], InputIDs[0]);
	sprintf(foo, "New%i.pa", NewDataSetIndex++);
	OutputIDs[0] = swiDSNew(PAClassID, OutputExternals[1], foo);

	return 0;
}

// added by Jian Pu, 30.5.2007
static SWI_ID swisPreformCompleteKMeans(SWI_ID AfterProcessInMenu)
{
	SWI_ID In, Out, Options, ID;
	In = swiRQNew();
	swiRQAdd(In, TSClassID, 1);
	swiRQAdd(In, CBClassID, 1);
	Out = swiRQNew();
	swiRQAdd(Out, CBClassID, 1);
	swiRQAdd(Out, PAClassID, 1);
	Options = swiPSNew(1, GOPParams);
	ID = swiPCNew(0, "K-means( complete )",
		&PrePreformCompleteKMeans,
		&MainPreformCompleteKMeans,
		&PostPreformCompleteKMeans,
		Options, In, Out);
	swiPCSetMenuInfo(ID, 'K', 1, 1, 0, 0, AfterProcessInMenu);
	return ID;
}

// added by Jian Pu, 7,6,2007
int PrePreformStepKMeans(int InputCount, SWI_ID* InputIDs, 
	void** InputExternals, int OutputCount, SWI_ID* OutputIDs, 
	void** OutputExternals, void** Aux) 
{
	TRAININGSET* TS;
	CODEBOOK* CB;
	*Aux = (void*)1;
	TS = InputExternals[0];
	CB = InputExternals[1];
	if (VectorSize(TS) != VectorSize(CB)) {
		ErrorMessage("Vector sizes don't match, %i for data set, "
			"%i for centroids.\n",
			VectorSize(TS), VectorSize(CB));
		return 1;
	}
	if (TS->MinValue != CB->MinValue || TS->MaxValue != CB->MaxValue) {
		ErrorMessage(
			"Scalar ranges don't match, for data set it is [%i, %i]"
			" and for centroids [%i, %i].\n",
			TS->MinValue, TS->MaxValue, CB->MinValue, CB->MaxValue);
		return 3;
	}
	switch (Value(ErrorFunction)) {
	case efMSE: break;
	case efSC: break;
	case efJACCARD: break;
	case efDICE: break;
	default: 
		ErrorMessage("Unknown error metric (%i)", Value(ErrorFunction));
		return 2;
	}
	return 0;
}

// added by Jian Pu, 7,6,2007
int MainPreformStepKMeans(int InputCount, void** InputExternals, 
	int OutputCount, void** OutputExternals, void** Aux)
{
	TRAININGSET* TS;
	CODEBOOK* iCB;
	CODEBOOK* oCB;
	PARTITIONING* oPA;
	int ErrorFunctionType;
	
	TS = InputExternals[0];
	if( g_CB == NULL )
		iCB = InputExternals[1];
	else
		iCB = g_CB;
	oCB = OutputExternals[0] = new_swiCODEBOOK();
	oPA = OutputExternals[1] = new_swiPARTITIONING(0);

	switch (Value(ErrorFunction)) {
	case efMSE: ErrorFunctionType = MSE; break;
	case efSC: ErrorFunctionType = SC; break;
	case efJACCARD: ErrorFunctionType = JACCARD; break;
	case efDICE: ErrorFunctionType = DICE; break;
	}

	CreateNewPartitioning(oPA, TS, BookSize(iCB));
	GenerateOptimalPartitioningGeneral(TS, iCB, oPA, ErrorFunctionType);

	CreateNewCodebook(oCB, oPA->PartitionCount, TS);
	GenerateOptimalCodebook(TS, oCB, oPA);

	if( swi_EqualCodeBook(iCB, oCB) )
	{
		swiInformationMessageBox("Stepwise K-means finished.");
		g_CB = NULL;
	}
	else
	{
		g_CB = oCB;
	}
	*Aux = 0;
	return 0;
}

// added by Jian Pu, 7,6,2007
int PostPreformStepKMeans(int InputCount, SWI_ID* InputIDs,
	void** InputExternals, int OutputCount, SWI_ID* OutputIDs, 
	void** OutputExternals, void** Aux) 
{
	char foo[40];
	if (*Aux) return 1;
	if( !g_CB ) return 1;

	sprintf(foo, "New%i.cb", NewDataSetIndex++);
	OutputIDs[1] = swiDSNew(CBClassID, OutputExternals[0], foo);

	SetswiPARTITIONING_TS_ID(OutputExternals[1], InputIDs[0]);
	sprintf(foo, "New%i.pa", NewDataSetIndex++);
	OutputIDs[0] = swiDSNew(PAClassID, OutputExternals[1], foo);

	return 0;
}

// added by Jian Pu, 7,6,2007
static SWI_ID swisPreformStepKMeans(SWI_ID AfterProcessInMenu)
{
	SWI_ID In, Out, Options, ID;
	In = swiRQNew();
	swiRQAdd(In, TSClassID, 1);
	swiRQAdd(In, CBClassID, 1);
	Out = swiRQNew();
	swiRQAdd(Out, CBClassID, 1);
	swiRQAdd(Out, PAClassID, 1);
	Options = swiPSNew(1, GOPParams);
	ID = swiPCNew(0, "K-means( stepwise )",
		&PrePreformStepKMeans,
		&MainPreformStepKMeans,
		&PostPreformStepKMeans,
		Options, In, Out);
	swiPCSetMenuInfo(ID, 'U', 1, 1, 0, 0, AfterProcessInMenu);
	return ID;
}

int PreGenerateOptimalCodebook(int InputCount, SWI_ID* InputIDs,
	void** InputExternals, int OutputCount, SWI_ID* OutputIDs, 
	void** OutputExternals, void** Aux) 
{
	TRAININGSET* TS;
	PARTITIONING* PA;
	*Aux = (void*)1;
	TS = InputExternals[0];
	PA = InputExternals[1];
	if (BookSize(TS) != PA->TSsize) {
		ErrorMessage("Input sizes do not match, %i != %i",
			BookSize(TS), PA->TSsize);
		return 1;
	}
	return 0;
}

int MainGenerateOptimalCodebook(int InputCount, void** InputExternals, 
	int OutputCount, void** OutputExternals, void** Aux)
{
	TRAININGSET* TS;
	PARTITIONING* PA;
	CODEBOOK* CB;
	TS = InputExternals[0];
	PA = InputExternals[1];
	CB = OutputExternals[0] = new_swiCODEBOOK();
	CreateNewCodebook(CB, PA->PartitionCount, TS);
	GenerateOptimalCodebook(TS, CB, PA);
	*Aux = 0;
	return 0;
}

int PostGenerateOptimalCodebook(int InputCount, SWI_ID* InputIDs,
	void** InputExternals, int OutputCount, SWI_ID* OutputIDs, 
	void** OutputExternals, void** Aux) 
{
	char foo[40];
	if (*Aux) return 1;
	sprintf(foo, "New%i.cb", NewDataSetIndex++);
	OutputIDs[0] = swiDSNew(CBClassID, OutputExternals[0], foo);
	return 0;
}

static SWI_ID swisGenerateOptimalCodebook(SWI_ID AfterProcessInMenu) {
	SWI_ID In, Out, ID;
	In = swiRQNew();
	swiRQAdd(In, TSClassID, 1);
	swiRQAdd(In, PAClassID, 1);
	Out = swiRQNew();
	swiRQAdd(Out, CBClassID, 1);
	ID = swiPCNew(0, "Optimal codebook",
		&PreGenerateOptimalCodebook,
		&MainGenerateOptimalCodebook,
		&PostGenerateOptimalCodebook,
		0, In, Out);
	swiPCSetMenuInfo(ID, 'D', 1, 1, 0, 0, AfterProcessInMenu);
	return ID;
}

/* add random centroid by increasing codebook size by one */

int PreAddCodevectorToCodebook(int InputCount, SWI_ID* InputIDs,
	void** InputExternals, int OutputCount, SWI_ID* OutputIDs, 
	void** OutputExternals, void** Aux) 
{
	TRAININGSET* TS;
	CODEBOOK* CB;
	*Aux = (void*)1;
	TS = InputExternals[0];
	CB = InputExternals[1];
	if (VectorSize(TS) != VectorSize(CB)) {
		ErrorMessage("Vector sizes don't match, %i for data set, "
			"%i for centroids.\n",
			VectorSize(TS), VectorSize(CB));
		return 1;
	}
	if (TS->MinValue != CB->MinValue || TS->MaxValue != CB->MaxValue) {
		ErrorMessage(
			"Scalar ranges don't match, for data set it is [%i, %i]"
			" and for centroids [%i, %i].\n",
			TS->MinValue, TS->MaxValue, CB->MinValue, CB->MaxValue);
		return 2;
	}
	if (BookSize(CB) == BookSize(TS)) {
		ErrorMessage(
			"Can't make codebook larger than data set.\n");
		return 3;
	}
	return 0;
}

int MainAddCodevectorToCodebook(int InputCount, void** InputExternals, 
	int OutputCount, void** OutputExternals, void** Aux)
{
	TRAININGSET* TS;
	CODEBOOK* CB;
	CODEBOOK* CB2;
	int x, n, Unique;
	TS = InputExternals[0];
	CB = InputExternals[1];
	CB2 = OutputExternals[0] = new_swiCODEBOOK();
	CreateNewCodebook(CB2, BookSize(CB) + 1, TS);
	SetExitRequestHonored(1);
	for (n = 0; n < BookSize(CB); ++n)
		CopyNode(&Node(CB, n), &Node(CB2, n), VectorSize(CB));
	do {
        Unique = 1;
        x = irand(0, BookSize(TS) - 1);
        for (n = 0; (n < BookSize(CB2) - 1) && Unique; n++) Unique = 
            !EqualVectors(Vector(TS, x), Vector(CB2, n), VectorSize(CB2));
		if (ExitRequested()) {
			FreeCodebook(CB2);
			delete_swiCODEBOOK(OutputExternals[0]);
			return EXIT_REQUEST_HONOURED;
		}
    } while (!Unique);
    CopyVector(Vector(TS, x), Vector(CB2, BookSize(CB2) - 1), VectorSize(CB2));
    VectorFreq(CB2, BookSize(CB2) - 1) = 0;
	*Aux = 0;
	return 0;
}

int PostAddCodevectorToCodebook(int InputCount, SWI_ID* InputIDs,
	void** InputExternals, int OutputCount, SWI_ID* OutputIDs, 
	void** OutputExternals, void** Aux) 
{
	char foo[40];
	if (*Aux) return 1;
	sprintf(foo, "Add%i.cb", NewDataSetIndex++);
	OutputIDs[0] = swiDSNew(CBClassID, OutputExternals[0], foo);
	return 0;
}

static SWI_ID swisAddCodevectorToCodebook(SWI_ID AfterProcessInMenu) {
	SWI_ID In, Out, ID;
	In = swiRQNew();
	swiRQAdd(In, TSClassID, 1);
	swiRQAdd(In, CBClassID, 1);
	Out = swiRQNew();
	swiRQAdd(Out, CBClassID, 1);
	ID = swiPCNew(0, "Add centroid (random)",
		&PreAddCodevectorToCodebook,
		&MainAddCodevectorToCodebook,
		&PostAddCodevectorToCodebook,
		0, In, Out);
	swiPCSetMenuInfo(ID, 'L', 1, 1, 0, 0, AfterProcessInMenu);
	return ID;
}

static SWI_ID swisAddCodevectorToCodebookWPA() {
	SWI_ID In, Out, ID;
	In = swiRQNew();
	swiRQAdd(In, TSClassID, 1);
	swiRQAdd(In, CBClassID, 1);
	swiRQAdd(In, PAClassID, 1);
	Out = swiRQNew();
	swiRQAdd(Out, CBClassID, 1);
	ID = swiPCNew(0, "Add centroid (random)",
		&PreAddCodevectorToCodebook,
		&MainAddCodevectorToCodebook,
		&PostAddCodevectorToCodebook,
		0, In, Out);
	return ID;
}

/* add codevector without data set */

int PreAddCodevectorToCodebookWOTS(int InputCount, SWI_ID* InputIDs,
	void** InputExternals, int OutputCount, SWI_ID* OutputIDs, 
	void** OutputExternals, void** Aux) 
{
	*Aux = (void*)1;
	return 0;
}

int MainAddCodevectorToCodebookWOTS(int InputCount, void** InputExternals, 
	int OutputCount, void** OutputExternals, void** Aux)
{
	CODEBOOK* CB;
	CODEBOOK* CB2;
	int n;
	CB = InputExternals[0];
	CB2 = OutputExternals[0] = new_swiCODEBOOK();
	CreateNewCodebook(CB2, BookSize(CB) + 1, CB);
	for (n = 0; n < BookSize(CB); ++n)
		CopyNode(&Node(CB, n), &Node(CB2, n), VectorSize(CB));
	for (n = 0; n < VectorSize(CB2); ++n)
		VectorScalar(CB2, BookSize(CB2) - 1, n) 
			= irand(CB2->MinValue, CB2->MaxValue);
    VectorFreq(CB2, BookSize(CB2) - 1) = 0;
	*Aux = 0;
	return 0;
}

int PostAddCodevectorToCodebookWOTS(int InputCount, SWI_ID* InputIDs,
	void** InputExternals, int OutputCount, SWI_ID* OutputIDs, 
	void** OutputExternals, void** Aux) 
{
	char foo[40];
	if (*Aux) return 1;
	sprintf(foo, "Add%i.cb", NewDataSetIndex++);
	OutputIDs[0] = swiDSNew(CBClassID, OutputExternals[0], foo);
	return 0;
}

static SWI_ID swisAddCodevectorToCodebookWOTS() {
	SWI_ID In, Out, ID;
	In = swiRQNew();
	swiRQAdd(In, CBClassID, 1);
	Out = swiRQNew();
	swiRQAdd(Out, CBClassID, 1);
	ID = swiPCNew(0, "Add random centroid without data set",
		&PreAddCodevectorToCodebookWOTS,
		&MainAddCodevectorToCodebookWOTS,
		&PostAddCodevectorToCodebookWOTS,
		0, In, Out);
	return ID;
}

static SWI_ID swisAddCodevectorToCodebookWOTSWPA() {
	SWI_ID In, Out, ID;
	In = swiRQNew();
	swiRQAdd(In, CBClassID, 1);
	swiRQAdd(In, PAClassID, 1);
	Out = swiRQNew();
	swiRQAdd(Out, CBClassID, 1);
	ID = swiPCNew(0, "Add random centroid without data set with partitioning",
		&PreAddCodevectorToCodebookWOTS,
		&MainAddCodevectorToCodebookWOTS,
		&PostAddCodevectorToCodebookWOTS,
		0, In, Out);
	return ID;
}


/* remove random centroid by decreasing codebook size by one */

int PreRemoveCodevectorFromCodebook(int InputCount, SWI_ID* InputIDs,
	void** InputExternals, int OutputCount, SWI_ID* OutputIDs, 
	void** OutputExternals, void** Aux) 
{
	TRAININGSET* TS;
	CODEBOOK* CB;
	*Aux = (void*)1;
	TS = InputExternals[0];
	CB = InputExternals[1];
	if (VectorSize(TS) != VectorSize(CB)) {
		ErrorMessage("Vector sizes don't match, %i for data set, "
			"%i for centroids.\n",
			VectorSize(TS), VectorSize(CB));
		return 1;
	}
	if (TS->MinValue != CB->MinValue || TS->MaxValue != CB->MaxValue) {
		ErrorMessage(
			"Scalar ranges don't match, for data set it is [%i, %i]"
			" and for centroids [%i, %i].\n",
			TS->MinValue, TS->MaxValue, CB->MinValue, CB->MaxValue);
		return 2;
	}
	if (BookSize(CB) == 1) {
		ErrorMessage("Only one codevector, can't remove.\n");
		return 3;
	}
	return 0;
}

int MainRemoveCodevectorFromCodebook(int InputCount, void** InputExternals, 
	int OutputCount, void** OutputExternals, void** Aux)
{
	TRAININGSET* TS;
	CODEBOOK* CB;
	CODEBOOK* CB2;
	int x, n;
	TS = InputExternals[0];
	CB = InputExternals[1];
	CB2 = OutputExternals[0] = new_swiCODEBOOK();
	CreateNewCodebook(CB2, BookSize(CB) - 1, TS);
	x = irand(0, BookSize(CB) - 2);
	for (n = 0; n < x; ++n)
		CopyNode(&Node(CB, n), &Node(CB2, n), VectorSize(CB));
	VectorFreq(CB2, 0) += VectorFreq(CB, x);
	for (n = x + 1; n < BookSize(CB); ++n)
		CopyNode(&Node(CB, n), &Node(CB2, n - 1), VectorSize(CB));
	*Aux = 0;
	return 0;
}

int PostRemoveCodevectorFromCodebook(int InputCount, SWI_ID* InputIDs,
	void** InputExternals, int OutputCount, SWI_ID* OutputIDs, 
	void** OutputExternals, void** Aux) 
{
	char foo[40];
	if (*Aux) return 1;
	sprintf(foo, "Remove%i.cb", NewDataSetIndex++);
	OutputIDs[0] = swiDSNew(CBClassID, OutputExternals[0], foo);
	return 0;
}

static SWI_ID swisRemoveCodevectorFromCodebook(SWI_ID AfterProcessInMenu) {
	SWI_ID In, Out, ID;
	In = swiRQNew();
	swiRQAdd(In, TSClassID, 1);
	swiRQAdd(In, CBClassID, 1);
	Out = swiRQNew();
	swiRQAdd(Out, CBClassID, 1);
	ID = swiPCNew(0, "Remove centroid (random)",
		&PreRemoveCodevectorFromCodebook,
		&MainRemoveCodevectorFromCodebook,
		&PostRemoveCodevectorFromCodebook,
		0, In, Out);
	swiPCSetMenuInfo(ID, 'R', 1, 1, 0, 0, AfterProcessInMenu);
	return ID;
}

/* Remove with partitioning present */

static SWI_ID swisRemoveCodevectorFromCodebookWPA() {
	SWI_ID In, Out, ID;
	In = swiRQNew();
	swiRQAdd(In, TSClassID, 1);
	swiRQAdd(In, CBClassID, 1);
	swiRQAdd(In, PAClassID, 1);
	Out = swiRQNew();
	swiRQAdd(Out, CBClassID, 1);
	ID = swiPCNew(0, "Remove random centroid with partitioning",
		&PreRemoveCodevectorFromCodebook,
		&MainRemoveCodevectorFromCodebook,
		&PostRemoveCodevectorFromCodebook,
		0, In, Out);
	return ID;
}

/* remove without data set */

int PreRemoveCodevectorFromCodebookWOTS(int InputCount, SWI_ID* InputIDs,
	void** InputExternals, int OutputCount, SWI_ID* OutputIDs, 
	void** OutputExternals, void** Aux) 
{
	CODEBOOK* CB;
	*Aux = (void*)1;
	CB = (CODEBOOK*) InputExternals[0];
	if (BookSize(CB) == 1) {
		ErrorMessage("Only one codevector, can't remove.\n");
		return 1;
	}
	return 0;
}

int MainRemoveCodevectorFromCodebookWOTS(int InputCount, void** InputExternals, 
	int OutputCount, void** OutputExternals, void** Aux)
{
	CODEBOOK* CB;
	CODEBOOK* CB2;
	int x, n;
	CB = InputExternals[0];
	CB2 = OutputExternals[0] = new_swiCODEBOOK();
	CreateNewCodebook(CB2, BookSize(CB) - 1, CB);
	x = irand(0, BookSize(CB) - 2);
	for (n = 0; n < x; ++n)
		CopyNode(&Node(CB, n), &Node(CB2, n), VectorSize(CB));
	VectorFreq(CB2, 0) += VectorFreq(CB, x);
	for (n = x + 1; n < BookSize(CB); ++n)
		CopyNode(&Node(CB, n), &Node(CB2, n - 1), VectorSize(CB));
	*Aux = 0;
	return 0;
}

int PostRemoveCodevectorFromCodebookWOTS(int InputCount, SWI_ID* InputIDs,
	void** InputExternals, int OutputCount, SWI_ID* OutputIDs, 
	void** OutputExternals, void** Aux) 
{
	char foo[40];
	if (*Aux) return 1;
	sprintf(foo, "Remove%i.cb", NewDataSetIndex++);
	OutputIDs[0] = swiDSNew(CBClassID, OutputExternals[0], foo);
	return 0;
}

static SWI_ID swisRemoveCodevectorFromCodebookWOTS() {
	SWI_ID In, Out, ID;
	In = swiRQNew();
	swiRQAdd(In, CBClassID, 1);
	Out = swiRQNew();
	swiRQAdd(Out, CBClassID, 1);
	ID = swiPCNew(0, "Remove random centroid without data set",
		&PreRemoveCodevectorFromCodebookWOTS,
		&MainRemoveCodevectorFromCodebookWOTS,
		&PostRemoveCodevectorFromCodebookWOTS,
		0, In, Out);
	return ID;
}

static SWI_ID swisRemoveCodevectorFromCodebookWOTSWPA() {
	SWI_ID In, Out, ID;
	In = swiRQNew();
	swiRQAdd(In, CBClassID, 1);
	swiRQAdd(In, PAClassID, 1);
	Out = swiRQNew();
	swiRQAdd(Out, CBClassID, 1);
	ID = swiPCNew(0, "Remove random centroid without data set",
		&PreRemoveCodevectorFromCodebookWOTS,
		&MainRemoveCodevectorFromCodebookWOTS,
		&PostRemoveCodevectorFromCodebookWOTS,
		0, In, Out);
	return ID;
}


/* random swap */

int PreRandomSwap(int InputCount, SWI_ID* InputIDs,
	void** InputExternals, int OutputCount, SWI_ID* OutputIDs, 
	void** OutputExternals, void** Aux) 
{
	TRAININGSET* TS;
	CODEBOOK* CB;
	*Aux = (void*)1;
	TS = InputExternals[0];
	CB = InputExternals[1];
	if (VectorSize(TS) != VectorSize(CB)) {
		ErrorMessage("Vector sizes don't match, %i for data set, "
			"%i for centroids.\n",
			VectorSize(TS), VectorSize(CB));
		return 1;
	}
	if (TS->MinValue != CB->MinValue || TS->MaxValue != CB->MaxValue) {
		ErrorMessage(
			"Scalar ranges don't match, for data set it is [%i, %i]"
			" and for centroids [%i, %i].\n",
			TS->MinValue, TS->MaxValue, CB->MinValue, CB->MaxValue);
		return 2;
	}
	if (BookSize(CB) == BookSize(TS)) {
		ErrorMessage(
			"Can't swap if data set size and centroid count are the same.\n");
		return 3;
	}
	return 0;
}

int MainRandomSwap(int InputCount, void** InputExternals, 
	int OutputCount, void** OutputExternals, void** Aux)
{
	TRAININGSET* TS;
	CODEBOOK* CB;
	CODEBOOK* CB2;
	int x, n, Unique, Index;
	TS = InputExternals[0];
	CB = InputExternals[1];
	CB2 = OutputExternals[0] = new_swiCODEBOOK();
	CreateNewCodebook(CB2, BookSize(CB), TS);
	CopyCodebook(CB, CB2);
	SetExitRequestHonored(1);
	Index = irand(0, BookSize(CB2) - 1);
	do {
        Unique = 1;
        x = irand(0, BookSize(TS) - 1);
        for (n = 0; (n < BookSize(CB2) - 1) && Unique; n++) Unique = 
            !EqualVectors(Vector(TS, x), Vector(CB2, n), VectorSize(CB2));
		if (ExitRequested()) {
			FreeCodebook(CB2);
			delete_swiCODEBOOK(OutputExternals[0]);
			return EXIT_REQUEST_HONOURED;
		}
    } while (!Unique);
    CopyVector(Vector(TS, x), Vector(CB2, Index), VectorSize(CB2));
    VectorFreq(CB2, Index) = 0;
	*Aux = 0;
	return 0;
}

int PostRandomSwap(int InputCount, SWI_ID* InputIDs,
	void** InputExternals, int OutputCount, SWI_ID* OutputIDs, 
	void** OutputExternals, void** Aux) 
{
	char foo[40];
	if (*Aux) return 1;
	sprintf(foo, "Swap%i.cb", NewDataSetIndex++);
	OutputIDs[0] = swiDSNew(CBClassID, OutputExternals[0], foo);
	return 0;
}

static SWI_ID swisRandomSwap(SWI_ID AfterProcessInMenu) {
	SWI_ID In, Out, ID;
	In = swiRQNew();
	swiRQAdd(In, TSClassID, 1);
	swiRQAdd(In, CBClassID, 1);
	Out = swiRQNew();
	swiRQAdd(Out, CBClassID, 1);
	ID = swiPCNew(0, "Random swap",
		&PreRandomSwap,
		&MainRandomSwap,
		&PostRandomSwap,
		0, In, Out);
	swiPCSetMenuInfo(ID, 'M', 1, 1, 0, 0, AfterProcessInMenu);
	return ID;
}

/* generate random solution */

int PreRandomSolution(int InputCount, SWI_ID* InputIDs,
	void** InputExternals, int OutputCount, SWI_ID* OutputIDs, 
	void** OutputExternals, void** Aux) 
{
	TRAININGSET* TS;
	*Aux = (void*)1;
	TS = InputExternals[0];
	if (Value(BookSize) > BookSize(TS)) {
		ErrorMessage("Too many clusters (%i), only %i vectors.\n",
			Value(BookSize), BookSize(TS));
		return 1;
	}
	return 0;
}

int MainRandomSolution(int InputCount, void** InputExternals, 
	int OutputCount, void** OutputExternals, void** Aux)
{
	TRAININGSET* TS;
	CODEBOOK* CB;
	int x, n, Unique, Index;
	TS = InputExternals[0];
	CB = OutputExternals[0] = new_swiCODEBOOK();
	CreateNewCodebook(CB, Value(BookSize), TS);
	SetExitRequestHonored(1);
	if (BookSize(TS) == BookSize(CB)) CopyCodebook(TS, CB);
	else for (Index = 0; Index < BookSize(CB); ++Index) {
		do {
			Unique = 1;
			x = irand(0, BookSize(TS) - 1);
			for (n = 0; (n < Index) && Unique; ++n) Unique = 
				!EqualVectors(Vector(TS, x), Vector(CB, n), VectorSize(CB));
			if (ExitRequested()) {
				FreeCodebook(CB);
				delete_swiCODEBOOK(OutputExternals[0]);
				return EXIT_REQUEST_HONOURED;
			}
		} while (!Unique);
		CopyVector(Vector(TS, x), Vector(CB, Index), VectorSize(CB));
		VectorFreq(CB, Index) = 0;
	}
	*Aux = 0;
	return 0;
}

int PostRandomSolution(int InputCount, SWI_ID* InputIDs,
	void** InputExternals, int OutputCount, SWI_ID* OutputIDs, 
	void** OutputExternals, void** Aux) 
{
	char foo[40];
	if (*Aux) return 1;
	sprintf(foo, "Random%i.cb", NewDataSetIndex++);
	OutputIDs[0] = swiDSNew(CBClassID, OutputExternals[0], foo);
	return 0;
}

static int RandomSolutionParams[] = { _BookSize };

static SWI_ID swisRandomSolution(SWI_ID AfterProcessInMenu) {
	SWI_ID In, Out, ID, Options;
	In = swiRQNew();
	swiRQAdd(In, TSClassID, 1);
	Out = swiRQNew();
	swiRQAdd(Out, CBClassID, 1);
	Options = swiPSNew(1, RandomSolutionParams);
	ID = swiPCNew(0, "Random codebook",
		&PreRandomSolution,
		&MainRandomSolution,
		&PostRandomSolution,
		Options, In, Out);
	swiPCSetMenuInfo(ID, 'N', 1, 1, 0, 0, AfterProcessInMenu);
	return ID;
}

/* evaluation functions */

int PreEvaluate(int InputCount, SWI_ID* InputIDs, void** InputExternals,
	int OutputCount, SWI_ID* OutputIDs, void** OutputExternals, void** Aux)
{
	TRAININGSET* TS;
	CODEBOOK* CB;
	PARTITIONING* PA;
	TS = InputExternals[0];
	CB = InputExternals[1];
	PA = InputExternals[2];
	if (BookSize(TS) != PA->TSsize) {
		ErrorMessage(
			"Data set size (%i) doesn't equal partition size (%i).\n",
			BookSize(TS), PA->TSsize);
		return 1;
	}
	if (BookSize(CB) != PartitionCount(PA)) {
		ErrorMessage(
			"Cluster centroid count (%i) "
			"doesn't equal partition count (%i).\n",
			BookSize(CB), PartitionCount(PA));
		return 2;
	}
	if (VectorSize(CB) != VectorSize(TS)) {
		ErrorMessage("Vector sizes don't match, %i for data set, "
			"%i for centroids.\n",
			VectorSize(TS), VectorSize(CB));
		return 3;
	}
	if (TS->MinValue != CB->MinValue || TS->MaxValue != CB->MaxValue) {
		ErrorMessage(
			"Scalar ranges don't match, for data set it is [%i, %i]"
			" and for centroids [%i, %i].\n",
			TS->MinValue, TS->MaxValue, CB->MinValue, CB->MaxValue);
		return 5;
	}
	if (TS->MaxValue == 1 
		&& Value(EvaluateFunction) != eStochasticComplexity
		&& Value(EvaluateFunction) != eAll) 
	{
		ErrorMessage(
			"This evaluation function doesn't apply to binary data.");
		return 6;
	}
	if (TS->MaxValue != 1 
		&& Value(EvaluateFunction) == eStochasticComplexity)
	{
		ErrorMessage(
			"Stochastic complexity doesn't yet apply to non-binary data.");
		return 7;
	}
	return 0;
}

int MainEvaluate(int InputCount, void** InputExternals, 
	int OutputCount, void** OutputExternals, void** Aux)
{
	TRAININGSET* TS;
	CODEBOOK* CB;
	PARTITIONING* PA;
	TS = InputExternals[0];
	CB = InputExternals[1];
	PA = InputExternals[2];
	switch (Value(EvaluateFunction)) {
	case 0: /* all */
		if (TS->MaxValue == 1) {
			PrintMessage(
				"Only stochastic complexity applies to binary data.\n");
			PrintMessage(
				"Stochastic complexity: %.6f\n", 
				EvaluatePartitioning(TS, CB, PA, 1));
			return 0;
		} else {
			PrintMessage(
				"Mean square error: %.6f\n", 
				EvaluatePartitioning(TS, CB, PA, 0));
			PrintMessage(
				"Stochastic complexity for non-binary data "
				"is not yet implemented.\n");
			PrintMessage(
				"Davies-Bouldin index: %.6f\n", 
				EvaluatePartitioning(TS, CB, PA, 2));
			return 0;
		}
	case 1: /* MSE */
		PrintMessage("Mean square error: %.6f\n", 
			EvaluatePartitioning(TS, CB, PA, 0));
		return 0;
	case 2: /* SC */
		PrintMessage("Stochastic complexity: %.6f\n", 
			EvaluatePartitioning(TS, CB, PA, 1));
		return 0;
	case 3: /* DBI */
		PrintMessage("Davies-Bouldin index: %.6f\n", 
			EvaluatePartitioning(TS, CB, PA, 2));
		return 0;
	}
	ErrorMessage("Unknown evaluation function: %s\n", 
		NameOfValue(EvaluateFunction, Value(EvaluateFunction)));
	return RECOVERABLE_ERROR;
}
	
static int EvaluateParams[1] = { _EvaluateFunction };

static SWI_ID swisEvaluate(SWI_ID AfterProcessInMenu) {
	SWI_ID In, ID, Options;
	In = swiRQNew();
	swiRQAdd(In, TSClassID, 1);
	swiRQAdd(In, CBClassID, 1);
	swiRQAdd(In, PAClassID, 1);
	Options = swiPSNew(1, EvaluateParams);
	ID = swiPCNew(0, "Evaluate", &PreEvaluate, &MainEvaluate, 
		0, Options, In, 0);
	swiPCSetMenuInfo(ID, 'E', 1, 1, 0, 0, AfterProcessInMenu);
	return ID;
}

int PreEvaluateWithoutPartitions(int InputCount, SWI_ID* InputIDs, 
	void** InputExternals, int OutputCount, SWI_ID* OutputIDs, 
	void** OutputExternals, void** Aux) 
{
	TRAININGSET* TS;
	CODEBOOK* CB;
	TS = InputExternals[0];
	CB = InputExternals[1];
	if (VectorSize(CB) != VectorSize(TS)) {
		ErrorMessage("Vector sizes don't match, %i for data set, "
			"%i for centroids.\n",
			VectorSize(TS), VectorSize(CB));
		return 3;
	}
	if (TS->MinValue != CB->MinValue || TS->MaxValue != CB->MaxValue) {
		ErrorMessage(
			"Scalar ranges don't match, for data set it is [%i, %i]"
			" and for centroids [%i, %i].\n",
			TS->MinValue, TS->MaxValue, CB->MinValue, CB->MaxValue);
		return 5;
	}
	if (TS->MaxValue == 1) {
		ErrorMessage(
			"No evaluation functions for binary data without "
			"partitioning available.\n");
		return 6;
	}
	return 0;
}

int MainEvaluateWithoutPartitions(int InputCount, void** InputExternals, 
	int OutputCount, void** OutputExternals, void** Aux)
{
	TRAININGSET* TS;
	CODEBOOK* CB;
	PARTITIONING PA;
	TS = InputExternals[0];
	CB = InputExternals[1];
	CreateNewPartitioning(&PA, TS, BookSize(CB));
	GenerateOptimalPartitioningGeneral(TS, CB, &PA, MSE);
	switch (Value(EvaluateFunction)) {
	case epAll: /* all */
		PrintMessage("Mean square error: %.6f\n",
			EvaluatePartitioning(TS, CB, &PA, 0));
		PrintMessage("Davies-Bouldin index: %.6f\n",
			EvaluatePartitioning(TS, CB, &PA, 2));
		break;
	case epMSE: /* MSE */
		PrintMessage("Mean square error: %.6f\n", 
			EvaluatePartitioning(TS, CB, &PA, 0));
		break;
	case epDaviesBouldinIndex: /* DBI */
		PrintMessage("Davies-Bouldin index: %.6f\n", 
			EvaluatePartitioning(TS, CB, &PA, 2));
		break;
	default:
		ErrorMessage("Unknown evaluation function: %s\n", 
			NameOfValue(EvaluateFunction, Value(EvaluateFunction)));
		FreePartitioning(&PA);
		return RECOVERABLE_ERROR;
	}
	FreePartitioning(&PA);
	return 0;
}

static int EvaluateWOPParams[1] = { _EvaluateFunctionWOP };

static SWI_ID swisEvaluateWithoutPartitions() {
	SWI_ID In, ID, Options;
	In = swiRQNew();
	swiRQAdd(In, TSClassID, 1);
	swiRQAdd(In, CBClassID, 1);
	Options = swiPSNew(1, EvaluateWOPParams);
	ID = swiPCNew(0, "Evaluate without partition", 
		&PreEvaluateWithoutPartitions, 
		&MainEvaluateWithoutPartitions, 0,
		Options, In, 0);
	return ID;
}


int PreEvaluateWithoutCodebook(int InputCount, SWI_ID* InputIDs, 
	void** InputExternals, int OutputCount, SWI_ID* OutputIDs, 
	void** OutputExternals, void** Aux) 
{
	TRAININGSET* TS;
	PARTITIONING* PA;
	TS = InputExternals[0];
	PA = InputExternals[1];
	if (BookSize(TS) != PA->TSsize) {
		ErrorMessage(
			"Data set size (%i) doesn't equal partition size (%i).\n",
			BookSize(TS), PA->TSsize);
		return 1;
	}
	if (TS->MaxValue == 1 
		&& Value(EvaluateFunction) != eStochasticComplexity
		&& Value(EvaluateFunction) != eAll) 
	{
		ErrorMessage(
			"This evaluation function doesn't apply to binary data.");
		return 6;
	}
	if (TS->MaxValue != 1 
		&& Value(EvaluateFunction) == eStochasticComplexity)
	{
		ErrorMessage(
			"Stochastic complexity doesn't yet apply to non-binary data.");
		return 7;
	}
	return 0;
}

int MainEvaluateWithoutCodebook(int InputCount, void** InputExternals, 
	int OutputCount, void** OutputExternals, void** Aux)
{
	TRAININGSET* TS;
	CODEBOOK CB;
	PARTITIONING* PA;
	TS = InputExternals[0];
	PA = InputExternals[1];
	CreateNewCodebook(&CB, PA->PartitionCount, TS);
	GenerateOptimalCodebook(TS, &CB, PA);
	switch (Value(EvaluateFunction)) {
	case 0: /* all */
		if (TS->MaxValue == 1) {
			PrintMessage(
				"Only stochastic complexity applies to binary data.\n");
			PrintMessage("Stochastic complexity: %.6f\n", 
				EvaluatePartitioning(TS, &CB, PA, 1));
			break;
		} else {
			PrintMessage("Mean square error: %.6f\n",
				EvaluatePartitioning(TS, &CB, PA, 0));
			PrintMessage("Stochastic complexity for non-binary "
				"data not yet implemented.\n");
			PrintMessage("Davies-Bouldin index: %.6f\n",
				EvaluatePartitioning(TS, &CB, PA, 2));
			break;
		}
	case 1: /* MSE */
		PrintMessage("Mean square error: %.6f\n", 
			EvaluatePartitioning(TS, &CB, PA, 0));
		break;
	case 2: /* SC */
		PrintMessage("Stochastic complexity: %.6f\n",
			EvaluatePartitioning(TS, &CB, PA, 1));
		break;
	case 3: /* DBI */
		PrintMessage("Davies-Bouldin index: %.6f\n",
			EvaluatePartitioning(TS, &CB, PA, 2));
		break;
	default:
		ErrorMessage("Unknown evaluation function: %s\n", 
			NameOfValue(EvaluateFunction, Value(EvaluateFunction)));
		FreeCodebook(&CB);
		return RECOVERABLE_ERROR;
	}
	FreeCodebook(&CB);
	return 0;
}
	
static SWI_ID swisEvaluateWithoutCodebook() {
	SWI_ID In, ID, Options;
	In = swiRQNew();
	swiRQAdd(In, TSClassID, 1);
	swiRQAdd(In, PAClassID, 1);
	Options = swiPSNew(1, EvaluateParams);
	ID = swiPCNew(0, "Evaluate without cluster centroids",
		&PreEvaluateWithoutCodebook, 
		&MainEvaluateWithoutCodebook, 0, Options, In, 0);
	return ID;
}


/*
int MainMemoryStress(int InputCount, void** InputExternals, 
	int OutputCount, void** OutputExternals, void** Aux)
{
	int k, slot;
	char* foo[128];
	initrandom(0);
	SetExitRequestHonored(1);
	OpenProgressWindow();
	for (k = 0; k < 128; k++) foo[k] = allocate(irand(2, 20480));
	for (k = 0; k < 10000000; k++) {
		slot = irand(0, 127);
		deallocate(foo[slot]);
		foo[slot] = allocate(irand(2, 20480));
		if (!(k % 1000)) Sleep(0);
		if (!(k % 1000000)) PrintMessage("%i\n", k / 1000000);
		if (!(k % 100000)) UpdateProgressBar(k, 10000000);
		if (ExitRequested()) break;
	}
	for (k = 0; k < 128; k++) deallocate(foo[k]);
	PrintMessage("Possible maximum allocated %i\n", 128 * 20480);
	CloseProgressWindow();
	return 0;
}

static SWI_ID swisMemoryStress(SWI_ID AfterProcessInMenu)
{
	SWI_ID Process;
	Process = swiPCNew(0, "Memory allocation/deallocation stress test", 
		0, &MainMemoryStress, 0, 0, 0, 0);
	swiPCSetMenuInfo(Process, (int)'M', 0, 1, 0, 1, AfterProcessInMenu);
	return Process;
}
*/

int __stdcall WinMain(HINSTANCE ThisInst, HINSTANCE Prev, 
	LPSTR CmdLine, int ShowState)
{
	SWI_ID SW, BFLS, Part, Book, Eval, EvalPart, EvalBook, Add, Remove,
		RandomSwap, RandomSolution, TmpID, TmpID2, KmeansComplete, KmeansStep;
	int ReturnValue;
	initrandom(0);	// add by Jian Pu to initialize random seed. 29.5.2007
	swiInitialize();
	swiUtil_Parameters();
	TSClassID = swisCreateTrainingSetClass();
	CBClassID = swisCreateCodebookClass();
	PAClassID = swisCreatePartitioningClass(TSClassID);
	GClassID = swisCreateGraphClass();
	RandomSolution = swisRandomSolution(0);
	Book = swisGenerateOptimalCodebook(RandomSolution);
	Part = swisGenerateOptimalPartitioning(Book);

	// added by Jian Pu, 30.5.2007
	KmeansComplete = swisPreformCompleteKMeans(Part); 
	KmeansStep = swisPreformStepKMeans(KmeansComplete);

	Add = swisAddCodevectorToCodebook(KmeansStep);
	Remove = swisRemoveCodevectorFromCodebook(Add);
	RandomSwap = swisRandomSwap(Remove);
	BFLS = swisLocalSearchFunction(
		TSClassID, CBClassID, PAClassID, RandomSwap);
	SW = swisStepwiseFunction(
		TSClassID, CBClassID, PAClassID, BFLS);
	Eval = swisEvaluate(SW);
	EvalPart = swisEvaluateWithoutPartitions();
	EvalBook = swisEvaluateWithoutCodebook();
	swiPCLink(Eval, EvalPart);
	swiPCLink(EvalPart, EvalBook);
	TmpID = swisRemoveCodevectorFromCodebookWPA();
	swiPCLink(Remove, TmpID);
	TmpID2 = swisRemoveCodevectorFromCodebookWOTSWPA();
	swiPCLink(TmpID, TmpID2);
	TmpID = swisRemoveCodevectorFromCodebookWOTS();
	swiPCLink(TmpID2, TmpID);
	TmpID = swisAddCodevectorToCodebookWPA();
	swiPCLink(Add, TmpID);
	TmpID2 = swisAddCodevectorToCodebookWOTSWPA();
	swiPCLink(TmpID, TmpID2);
	TmpID = swisAddCodevectorToCodebookWOTS();
	swiPCLink(TmpID2, TmpID);
	swisCreateImageView();
	ReturnValue = swiRun("Cluster", "Cluster v0.21 (19.6.2007)", "Cluster.hlp");
	swiFinalize();
	return ReturnValue;
}
