/****************************************************************

  swis_ImageView.c

  Ismo Krkkinen

  Last changed 17.11.1999

****************************************************************/

#include "swis_ImageView.h"
#include "interfc.h"
#include "memctrl.h"
#include "cb.h"
#include <windows.h>
#include "swi.h"
#include "ClusterGlobals.h"
#include "voronoi.h"
#include "cnvxhull.h"
#include <stdio.h>
#include <limits.h>


typedef struct {
	int TS_ID;
	int TSSize;
	int AddedCBs;
	SWI_ID* CBs;
	int AddedPAs;
	SWI_ID* PAs;
	int AddedGs;
	SWI_ID* Gs;
	//int MaxVal, MinVal;
} ImageViewData;

void* swis_IVCreate() {
	ImageViewData* D = allocate(sizeof(ImageViewData));
	D->TS_ID = 0;
	D->TSSize = 0;
	D->AddedCBs = 0;
	D->CBs = 0;
	D->AddedPAs = 0;
	D->PAs = 0;
	D->AddedGs = 0;
	D->Gs = 0;
	/*D->MaxVal = 0;
	D->MinVal = 0;*/
	return D;
}

void swis_IVDestroy(void* ImageData) {
	ImageViewData* D = (ImageViewData*) ImageData;
	deallocate(D->CBs);
	deallocate(D->PAs);
	deallocate(D->Gs);
	deallocate(D);
}

static int CBColor = 0;

swiDrawingCommands* swis_IVAdd(AddInfo* AI) {
	int k, n, Pindex;
	swiDrawingCommands* DCs;
	TRAININGSET UseTS;
	TRAININGSET* TS;
	PARTITIONING UsePA;
	PARTITIONING* PA;
	CONVEXHULLSET* CHS;
	CONVEXVERTEX* CV1, *CV2;
	CODEBOOK* CB;
	Graph* G;
	GraphVector* GV;
	GraphVector* GVTo;
	int* Coords;
	int* CoordsTo;
	ImageViewData* D = AI->ViewData;
	char Buffer[256];
	if (AI->Class == TSClassID) {
		EnterCriticalSection(AI->CS);
		if (D->TS_ID) {
			LeaveCriticalSection(AI->CS);
			AI->ErrorString = allocate(
				strlen("A data set has already been added.") + 1);
			strcpy(AI->ErrorString, 
				"A data set has already been added.");
			return 0;
		}
		TS = (TRAININGSET*) AI->External;
		if (VectorSize(TS) < 2) {
			sprintf(Buffer, "Cannot plot one-dimensional data.");
			LeaveCriticalSection(AI->CS);
			AI->ErrorString = allocate(strlen(Buffer) + 1);
			strcpy(AI->ErrorString, Buffer);
			return 0;
		}
		/*if (D->AddedCBs && ((D->MaxVal != TS->MaxValue) 
			|| (D->MinVal != TS->MinValue)))
		{
			sprintf(Buffer, "Scalar value limits do not match.\n"
				"Values should be within [%i, %i], "
				"for data set were within [%i, %i].",
				D->MinVal, D->MaxVal, TS->MinValue, TS->MaxValue);
			LeaveCriticalSection(AI->CS);
			AI->ErrorString = allocate(strlen(Buffer) + 1);
			strcpy(AI->ErrorString, Buffer);
			return 0;
		}*/
		D->TS_ID = AI->ID;
		/*if (!D->AddedGs) {
			D->MinVal = TS->MinValue;
			D->MaxVal = TS->MaxValue;
		} else {
			if (D->MinVal > TS->MinValue) D->MinVal = TS->MinValue;
			if (D->MaxVal < TS->MaxValue) D->MaxVal = TS->MaxValue;
		}*/
		D->TSSize = BookSize(TS);
		/* convert to drawing commands */
		DCs = swiDRNew(D->TS_ID);
		for (k = 0; k < BookSize(TS); k++) swiDRCircle(DCs, 
			VectorScalar(TS, k, 0), VectorScalar(TS, k, 1));
		if (D->AddedCBs) swiIVUpdateSpecial(AI->Handle);
		LeaveCriticalSection(AI->CS);
		return DCs;
	}
	if (AI->Class == PAClassID) { 
		EnterCriticalSection(AI->CS);
		if (!D->TS_ID) {
			LeaveCriticalSection(AI->CS);
			AI->ErrorString = allocate(1 + strlen(
				"You must add a data set before you can add a partition."));
			strcpy(AI->ErrorString, 
				"You must add a data set before you can add a partition.");
			return 0;
		}
		PA = (PARTITIONING*) AI->External;
		if (D->TSSize != PA->TSsize) {
			sprintf(Buffer, 
				"Sizes of data set and partitions don't match.\n"
				"Partitions size == %i, data set size == %i.",
				PA->TSsize, D->TSSize);
			LeaveCriticalSection(AI->CS);
			AI->ErrorString = allocate(strlen(Buffer) + 1);
			strcpy(AI->ErrorString, Buffer);
			return 0;
		}
		for (k = 0; k < D->AddedPAs; k++) if (AI->ID == D->PAs[k]) {
			LeaveCriticalSection(AI->CS);
			return 0;
		}
		D->AddedPAs++;
		D->PAs = reallocate(D->PAs, sizeof(SWI_ID) * D->AddedPAs);
		D->PAs[D->AddedPAs - 1] = AI->ID;
		/* convert to drawing commands */
		TS = (TRAININGSET*) swiDSExternal(D->TS_ID);
		CreateNewTrainingSet(&UseTS, BookSize(TS), 
			TS->BlockSizeX, TS->BlockSizeY, 
			TS->BytesPerElement, TS->MinValue, TS->MaxValue, 
			TS->Preprocessing, TS->GenerationMethod);
		CopyCodebook(TS, &UseTS);
		CreateNewPartitioning(&UsePA, &UseTS, PA->PartitionCount);
		CopyPartitioning(PA, &UsePA);
		LeaveCriticalSection(AI->CS);
		k = 0;
		while (k < BookSize(&UseTS)) {
			n = 0;
			while (n < k) {
				if (VectorScalar(&UseTS, k, 0) 
					== VectorScalar(&UseTS, n, 0)
					&& VectorScalar(&UseTS, k, 1) 
					== VectorScalar(&UseTS, n, 1))
				{
					Pindex = Map(&UsePA, BookSize(&UseTS) - 1);
					ChangePartition(&UseTS, &UsePA, ENDPARTITION, n);
					ChangePartition(&UseTS, &UsePA, ENDPARTITION,
						BookSize(&UseTS)-1);
					CopyNode(&Node(&UseTS, BookSize(&UseTS) - 1),
						&Node(&UseTS, k), 2);
					ChangePartition(&UseTS, &UsePA, Pindex, k);
					BookSize(&UseTS)--;
					n = k;
					k--;
				} else n++;
			}
			k++;
		}
		ConstructConvexHulls(&UseTS, &UsePA, &CHS);
		FreePartitioning(&UsePA);
		DCs = swiDRNew(AI->ID);
		for (k = 0; k < CHS_size(CHS); k++) {
			if (CHS_hullsize(CHS, k) > 1) {
				CV1 = CHS_hullfirst(CHS, k);
				CV2 = CV_next(CV1);
				while (CV2) {
					swiDRLine(DCs, 
						VectorScalar(&UseTS, CV_index(CV1), 0),
						VectorScalar(&UseTS, CV_index(CV1), 1),
						VectorScalar(&UseTS, CV_index(CV2), 0),
						VectorScalar(&UseTS, CV_index(CV2), 1));
					CV1 = CV2;
					CV2 = CV_next(CV2);
				}
			}
		}
		FreeCodebook(&UseTS);
		FreeConvexHulls(CHS);
		return DCs;
	}
	if (AI->Class == CBClassID) {
		CB = (CODEBOOK*) AI->External;
		EnterCriticalSection(AI->CS);
		if (VectorSize(CB) < 2) {
			sprintf(Buffer, "Cannot plot one-dimensional data.");
			LeaveCriticalSection(AI->CS);
			AI->ErrorString = allocate(strlen(Buffer) + 1);
			strcpy(AI->ErrorString, Buffer);
			return 0;
		}
		/*if ((D->AddedCBs || D->TS_ID) 
			&& ((D->MaxVal != CB->MaxValue) || (D->MinVal != CB->MinValue)))
		{
			sprintf(Buffer, "Scalar value limits do not match.\n"
				"Values should be within [%i, %i], "
				"for centroids were within [%i, %i].",
				D->MinVal, D->MaxVal, CB->MinValue, CB->MaxValue);
			LeaveCriticalSection(AI->CS);
			AI->ErrorString = allocate(strlen(Buffer) + 1);
			strcpy(AI->ErrorString, Buffer);
			return 0;
		}*/
		if (D->TS_ID && (BookSize(CB) > D->TSSize)) {
			sprintf(Buffer, "Number of cluster centroids is greater "
				"than the data set size.\n"
				"Number of centroids == %i, data set size == %i.", 
				BookSize(CB), D->TSSize);
			LeaveCriticalSection(AI->CS);
			AI->ErrorString = allocate(strlen(Buffer) + 1);
			strcpy(AI->ErrorString, Buffer);
			return 0;
		}
		for (k = 0; k < D->AddedCBs; k++) if (AI->ID == D->CBs[k]) {
			LeaveCriticalSection(AI->CS);
			return 0;
		}
		D->AddedCBs++;
		D->CBs = reallocate(D->CBs, D->AddedCBs * sizeof(SWI_ID));
		D->CBs[D->AddedCBs - 1] = AI->ID;
		/*if (!D->AddedGs) {
			D->MinVal = CB->MinValue;
			D->MaxVal = CB->MaxValue;
		} else {
			if (D->MinVal > CB->MinValue) D->MinVal = CB->MinValue;
			if (D->MaxVal < CB->MaxValue) D->MaxVal = CB->MaxValue;
		}*/
		DCs = swiDRNew(AI->ID);
		LeaveCriticalSection(AI->CS);
		AI->ParamIndex = 4;
		if (++CBColor > 9) CBColor = 0;
		AI->ParamValue = CBColor;
		for (k = 0; k < BookSize(CB); k++) swiDRCircle(DCs, 
			VectorScalar(CB, k, 0), VectorScalar(CB, k, 1));
		swiIVUpdateSpecial(AI->Handle);
		return DCs;
	}
	if (AI->Class == GClassID) {
		G = *((Graph**) AI->External);
		EnterCriticalSection(AI->CS);
		if (GraphGetDim(G) < 2) {
			sprintf(Buffer, "Cannot plot one-dimensional graph.");
			LeaveCriticalSection(AI->CS);
			AI->ErrorString = allocate(strlen(Buffer) + 1);
			strcpy(AI->ErrorString, Buffer);
			return 0;
		}
		D->AddedGs++;
		D->Gs = reallocate(D->Gs, sizeof(SWI_ID) * D->AddedGs);
		D->Gs[D->AddedGs - 1] = AI->ID;
		LeaveCriticalSection(AI->CS);
		/*if (!D->AddedCBs && !D->TS_ID) {
			D->MinVal = GraphGetMinCoord(G);
			D->MaxVal = GraphGetMaxCoord(G);
		} else {
			if (D->MinVal > GraphGetMinCoord(G))
				D->MinVal = GraphGetMinCoord(G);
			if (D->MaxVal < GraphGetMaxCoord(G))
				D->MaxVal = GraphGetMaxCoord(G);
		}*/
		DCs = swiDRNew(AI->ID);
		for (n = 0; n < GraphGetNumberVectors(G); ++n) {
			GV = GraphGetVector(G, n);
			Coords = GraphGetVectorCoord(GV);
			swiDRCircle(DCs, Coords[0], Coords[1]);
			k = 0;
			GVTo = GraphGetNNVector(G, GV, 0);
			while (GVTo) {
				if (GraphGetVectorIndex(GVTo) < n 
					|| (GraphGetVectorIndex(GVTo) > n 
						&& !IsGraphEdge(G, GVTo, GV)))
				{
					CoordsTo = GraphGetVectorCoord(GVTo);
					swiDRArrow(DCs,
						Coords[0], Coords[1], CoordsTo[0], CoordsTo[1],
						1, (IsGraphEdge(G, GVTo, GV) ? 1 : 0));
				}
				++k;
				GVTo = GraphGetNNVector(G, GV, k);
			}
		}
		return DCs;
	}
	strcpy(Buffer, "Data type not supported.");
	AI->ErrorString = allocate(strlen(Buffer) + 1);
	strcpy(AI->ErrorString, Buffer);
	return 0;
}

void swis_IVRemove(AddInfo* AI) {
	int k;
	ImageViewData* IVD = AI->ViewData;
	EnterCriticalSection(AI->CS);
	if ((int)AI->ID == IVD->TS_ID) {
		for (k = 0; k < IVD->AddedPAs; k++)
			swiIVRemove(AI->Handle, IVD->PAs[k]);
		IVD->TS_ID = 0;
		IVD->TSSize = 0;
		//if (!IVD->AddedCBs) IVD->MinVal = IVD->MaxVal = 0;
		LeaveCriticalSection(AI->CS);
		return;
	}
	for (k = 0; k < IVD->AddedPAs; k++) if (IVD->PAs[k] == AI->ID) {
		IVD->PAs[k] = IVD->PAs[IVD->AddedPAs - 1];
		IVD->AddedPAs--;
		LeaveCriticalSection(AI->CS);
		return;
	}
	for (k = 0; k < IVD->AddedCBs; k++) if (IVD->CBs[k] == AI->ID) {
		IVD->CBs[k] = IVD->CBs[IVD->AddedCBs - 1];
		IVD->AddedCBs--;
		//if (!IVD->AddedCBs && !IVD->TS_ID) IVD->MinVal = IVD->MaxVal = 0;
		LeaveCriticalSection(AI->CS);
		swiIVUpdateSpecial(AI->Handle);
		return;
	}
	for (k = 0; k < IVD->AddedGs; ++k) if (IVD->Gs[k] == AI->ID) {
		IVD->Gs[k] = IVD->Gs[IVD->AddedGs - 1];
		IVD->AddedGs--;
		LeaveCriticalSection(AI->CS);
		swiIVUpdateSpecial(AI->Handle);
		return;
	}
}

swiDrawingCommands* swis_IVVoronoi(ImageViewData* D, CRITICAL_SECTION* CS) {
	swiDrawingCommands* DCs;
	int k, j, v1, v2;
	//float Factor;
	VoronoiGraph* VG;
	CODEBOOK Union;
	TRAININGSET* TS;
	CODEBOOK* CB;
	EnterCriticalSection(CS);
	if (!D->TS_ID || !D->AddedCBs) {
		LeaveCriticalSection(CS);
		return 0;
	}
	TS = swiDSExternal(D->TS_ID);
	if (VectorSize(TS) != 2) {
		LeaveCriticalSection(CS);
		return 0;
	}
	CB = swiDSExternal(D->CBs[0]);
	CreateNewCodebook(&Union, BookSize(CB), TS);
	CopyCodebook(CB, &Union);
	for (k = 1; k < D->AddedCBs; k++) {
		CB = swiDSExternal(D->CBs[k]);
		AddCodebook(&Union, CB);
	}
	//Factor = 1000000.0f / (D->MaxVal - D->MinVal);
	LeaveCriticalSection(CS);
	k = 0;
	while (k < BookSize(&Union)) {
		j = 0;
		while (j < k) {
			if (VectorScalar(&Union, k, 0) == VectorScalar(&Union, j, 0) &&
				VectorScalar(&Union, k, 1) == VectorScalar(&Union, j, 1) )
			{
				BookSize(&Union)--;
				CopyVector(Vector(&Union, BookSize(&Union)), 
					Vector(&Union, k), 2);
				j = k;
				k--;
			} else j++;
		}
		k++;
    }
	VG = VG_make(&Union);
	VG_generate(VG, &Union);
	FreeCodebook(&Union);
	DCs = swiDRNew(0);
	for (k = 0; k < VG_nEdge(VG); k++) {
		switch (VG_EdgeStatus(VG, k)) {
		case CONNECTION:
			v1 = VG_Vertex(VG, k, 0);
			v2 = VG_Vertex(VG, k, 1);
			swiDRLine(DCs, 
				VG_VertexCoordinate(VG, v1)->x,
				VG_VertexCoordinate(VG, v1)->y,
				VG_VertexCoordinate(VG, v2)->x,
				VG_VertexCoordinate(VG, v2)->y);
			break;
		case DIRECTION:
			break;
		case BORDER:
			break;
		}
	}
	VG_kill(VG);
	return DCs;
}

static int Options[1] = { 1 };

void swisCreateImageView() {
	SWI_ID Params;
	swiIVFunctions(&swis_IVCreate, &swis_IVDestroy, &swis_IVAdd, 
		&swis_IVRemove, &swis_IVVoronoi);
	Params = swiPSNew(1, Options);
	swiIVSetParameters(Params, 
		"Draw a Voronoi diagram using union of all cluster centroids.");
}