/****************************************************************

  swis_cb.c

  Ismo Krkkinen

  Last changed 4.9.2003

****************************************************************/

#include "cb.h"
#include "swis_cb.h"
#include "memctrl.h"
#include <math.h>
#include <stdio.h>
#include <string.h>
#include "interfc.h"
#include <limits.h>


swiCODEBOOK* new_swiCODEBOOK() {
	swiCODEBOOK* sCB = allocate(sizeof(swiCODEBOOK));
	sCB->LastLine = 0;
	sCB->TextCount = 0;
	return sCB;
}

void delete_swiCODEBOOK(swiCODEBOOK* sCB) {
	if (sCB->LastLine && (sCB->LastLine != sCB->Buffer))
		deallocate(sCB->LastLine);
	deallocate(sCB);
}

int swis_CBDestroy(void* External) {
	swiCODEBOOK* sCB = (swiCODEBOOK*) External;
	if (sCB) {
		FreeCodebook(&(sCB->CB));
		delete_swiCODEBOOK(sCB);
	}
	return 0;
}

int swis_CBLoad(void** External, char* Filename, int Count, SWI_ID* Reqs) {
	swiCODEBOOK* sCB = new_swiCODEBOOK();
	ReadCodebook(Filename, &(sCB->CB));
	*External = sCB;
	return 0;
}

int swis_CBSave(void* External, char* Filename) {
	swiCODEBOOK* sCB = (swiCODEBOOK*) External;
	WriteCodebook(Filename, &(sCB->CB), 1);
	return 0;
}

// add by Jian Pu, 7.6.2007
int swi_EqualCodeBook(CODEBOOK* CB1, CODEBOOK* CB2)
{
	float EPSILON = 0.01;
	int i;

	if( CB1->CodebookSize != CB2->CodebookSize )
		return 0;
	for( i=0; i<CB1->CodebookSize; ++i )//for( int i=0; i<CB1->CodebookSize; ++i )
	{
		if( abs(CB1->Book[i].vmean-CB2->Book[i].vmean) > EPSILON )
			return 0;
	}
	return 1;
}


int swis_CBLineCount(void* External, int* ApproximateMaxLineLength) {
	int Dim, NumSize;
	swiCODEBOOK* sCB = (swiCODEBOOK*) External;
	Dim = VectorSize(&(sCB->CB));
	if (sCB->LastLine == 0) {
		NumSize = (sCB->CB.MaxValue) ? (int)log10(sCB->CB.MaxValue) + 1 : 1;
		sCB->LineLength = NumSize * Dim + NumSize + 14;
		if (sCB->LineLength > SWIS_CB_BUFFER_LENGTH) 
			sCB->LastLine = allocate(sCB->LineLength);
		else sCB->LastLine = sCB->Buffer;
		sCB->Format[0] = '%';
		if (NumSize > 9) {
			sCB->Format[1] = '0' + NumSize / 10;
			sCB->Format[2] = '0' + NumSize % 10;
			sCB->Format[3] = 'i';
			sCB->Format[4] = ' ';
			sCB->Format[5] = 0;
		} else {
			sCB->Format[1] = '0' + NumSize;
			sCB->Format[2] = 'i';
			sCB->Format[3] = ' ';
			sCB->Format[4] = 0;
		}
	}
	*ApproximateMaxLineLength = sCB->LineLength;
	return 4 + BookSize(&(sCB->CB));
}

char* swis_CBLine(void* External, int Line) {
	int k, n, Dim;
	char Buffer[15];
	swiCODEBOOK* sCB = (swiCODEBOOK*) External;
	Dim = VectorSize(&(sCB->CB));
	sCB->LastLine[0] = 0;
	if (Line >= 5) {
		Line -= 5;
		if ((sCB->CB.MinValue == 0) && (sCB->CB.MaxValue == 1)) {
			for (k = 0; k < Dim; k++) 
				if (VectorScalar(&(sCB->CB), Line, k)) 
					sCB->LastLine[k] = '1';
				else sCB->LastLine[k] = '0';
			sCB->LastLine[Dim] = 0;
		} else for (k = 0; k < Dim; k++) {
			n = VectorScalar(&(sCB->CB), Line, k);
			sprintf(Buffer, sCB->Format, n);
			strcat(sCB->LastLine, Buffer);
		}
		sprintf(Buffer, " (%i)", VectorFreq(&(sCB->CB), Line));
		strcat(sCB->LastLine, Buffer);
	} else {
		sCB->Buffer[0] = 0;
		switch (Line) {
		case 1:
			strcpy(sCB->Buffer, sCB->CB.Versionstr);
			break;
		case 2:
			sprintf(sCB->Buffer, 
				"%i vectors, total frequency %i, vector size %i",
				BookSize(&(sCB->CB)), TotalFreq(&(sCB->CB)), 
				VectorSize(&(sCB->CB)));
			break;
		case 3:
			if (sCB->CB.BytesPerElement == 1)
				sprintf(sCB->Buffer, 
					"Values: %i...%i (1 byte per element)", 
					sCB->CB.MinValue, sCB->CB.MaxValue);
			else sprintf(sCB->Buffer, 
				"Values: %i...%i (%i bytes per element)", 
				sCB->CB.MinValue, sCB->CB.MaxValue,
				sCB->CB.BytesPerElement);
			break;
		case 4:
			if (sCB->CB.Preprocessing == 1 || sCB->CB.Preprocessing == 0) 
				sprintf(sCB->Buffer, 
					"Generated by %s", sCB->CB.GenerationMethod);
			else sprintf(sCB->Buffer,
				"Generated by %s (preprocess = %i)", 
				sCB->CB.GenerationMethod, sCB->CB.Preprocessing);
			break;
		}
		return sCB->Buffer;
	}
	return sCB->LastLine;
}

static int CBImageParams[3] = { 4, 5, 6 };

SWI_ID swisCreateCodebookClass() {
	SWI_ID Class, Filter;
	Filter = swiFLNew();
	swiFLAddOpenStrings(Filter, "Cluster centroids", "*.cb");
	swiFLAddStrings(Filter, "All files", "*");
	swiFLDefaultExtensionForSave(Filter, "cb");
	Class = swiDCNew("Cluster centroids", 0, SWI_ICON_CODEBOOK);
	swiDCDestroyFunc(Class, &swis_CBDestroy);
	swiDCLoadFunc(Class, &swis_CBLoad);
	swiDCSaveFunc(Class, &swis_CBSave);
	swiDCTextFunc(Class, &swis_CBLineCount, &swis_CBLine);
	swiDCAskSaveOnExit(Class, 1);
	swiDCSetFilter(Class, Filter);
	swiDCSetViewParameters(Class, swiPSNew(3, CBImageParams));
	return Class;
}

static int TSImageParams[2] = { 7, 8 };

SWI_ID swisCreateTrainingSetClass() {
	SWI_ID Class, Filter;
	Filter = swiFLNew();
	swiFLAddOpenStrings(Filter, "Data set", "*.ts");
	swiFLAddStrings(Filter, "All files", "*");
	swiFLDefaultExtensionForSave(Filter, "ts");
	Class = swiDCNew("Data set", 0, SWI_ICON_TRAININGSET);
	swiDCDestroyFunc(Class, &swis_CBDestroy);
	swiDCLoadFunc(Class, &swis_CBLoad);
	swiDCSaveFunc(Class, &swis_CBSave);
	swiDCTextFunc(Class, &swis_CBLineCount, &swis_CBLine);
	swiDCAskSaveOnExit(Class, 1);
	swiDCSetFilter(Class, Filter);
	swiDCSetViewParameters(Class, swiPSNew(2, TSImageParams));
	return Class;
}

void SetswiPARTITIONING_TS_ID(swiPARTITIONING* P, SWI_ID NewTS_ID) {
	if (P->TS_ID) swiDSUnuse(P->TS_ID);
	P->TS_ID = NewTS_ID;
	if (P->TS_ID) {
		swiDSUse(P->TS_ID);
		P->TS = &(((swiCODEBOOK*)swiDSExternal(P->TS_ID))->CB);
	} else P->TS = 0;
}

swiPARTITIONING* new_swiPARTITIONING(SWI_ID TS_ID) {
	swiPARTITIONING* P = allocate(sizeof(swiPARTITIONING));
	P->TS_ID = 0;
	SetswiPARTITIONING_TS_ID(P, TS_ID);
	return P;
}

void delete_swiPARTITIONING(swiPARTITIONING* P) {
	if (P->TS_ID) swiDSUnuse(P->TS_ID);
	deallocate(P);
}

int swis_PALoad(void** External, char* Filename, int Count, SWI_ID* Reqs) {
	swiPARTITIONING* P = new_swiPARTITIONING(Reqs[0]);
	*External = P;
	ReadPartitioning(Filename, &(P->PA), P->TS);
	return 0;
}

int swis_PASave(void* External, char* Filename) {
	swiPARTITIONING* P = (swiPARTITIONING*) External;
	WritePartitioning(Filename, &(P->PA), P->TS, 1);
	return 0;
}

int swis_PADestroy(void* External) {
	swiPARTITIONING* P = (swiPARTITIONING*) External;
	if (P) {
		FreePartitioning(&(P->PA));
		delete_swiPARTITIONING(P);
	}
	return 0;
}

int swis_PALineCount(void* External, int* ApproximateMaxLineLength) {
	swiPARTITIONING* P = (swiPARTITIONING*) External;
	*ApproximateMaxLineLength = 26;
	return 2 + P->PA.TSsize;
}

char* swis_PALine(void* External, int Line) {
	swiPARTITIONING* P = (swiPARTITIONING*) External;
	static char Buffer[80];
	if (Line >= 3) {
		Line -= 3;
		sprintf(Buffer, "%i -> %i", Line + 1, P->PA.Map[Line]);
	} else {
		switch (Line) {
		case 1:
			strcpy(Buffer, P->PA.Versionstr);
			break;
		case 2:
			sprintf(Buffer, "%i vectors, %i clusters.",
				P->PA.TSsize, P->PA.PartitionCount);
			break;
		}
	}
	return Buffer;
}

static int PAImageParams[1] = { 3 };

SWI_ID swisCreatePartitioningClass(SWI_ID TrainingSetClass) {
	SWI_ID Filter, Class, Req;
	Filter = swiFLNew();
	swiFLAddOpenStrings(Filter, "Partitions", "*.pa");
	swiFLAddStrings(Filter, "All files", "*");
	swiFLDefaultExtensionForSave(Filter, "pa");
	Req = swiRQNew();
	swiRQAdd(Req, TrainingSetClass, 1);
	Class = swiDCNew("Partitions", Req, SWI_ICON_PARTITIONING);
	swiDCLoadFunc(Class, &swis_PALoad);
	swiDCSaveFunc(Class, &swis_PASave);
	swiDCDestroyFunc(Class, &swis_PADestroy);
	swiDCTextFunc(Class, &swis_PALineCount, &swis_PALine);
	swiDCAskSaveOnExit(Class, 1);
	swiDCSetFilter(Class, Filter);
	swiDCSetViewParameters(Class, swiPSNew(1, PAImageParams));
	return Class;
}



swiGRAPH* new_swiGRAPH() {
	swiGRAPH* sG = allocate(sizeof(swiGRAPH));
	sG->G = (Graph*)0;
	sG->Line = (char*) 0;
	sG->LineLength = 0;
	return sG;
}

void delete_swiGRAPH(swiGRAPH* sG) {
	deallocate(sG->Line);
	deallocate(sG);
}

int swis_GDestroy(void* External) {
	swiGRAPH* sG = (swiGRAPH*) External;
	if (sG) {
		FreeGraph(sG->G);
		delete_swiGRAPH(sG);
	}
	return 0;
}

int swis_GLoad(void** External, char* Filename, int Count, SWI_ID* Reqs) {
	swiGRAPH* sG;
	if (!IsGraphFile(Filename)) return 1;
	sG = new_swiGRAPH();
	sG->G = GraphRead(Filename);
	if (!sG->G) {
		delete_swiGRAPH(sG);
		*External = 0;
		return 2;
	}
	*External = sG;
	return 0;
}

int swis_GSave(void* External, char* Filename) {
	swiGRAPH* sG = (swiGRAPH*) External;
	GraphWrite(Filename, sG->G, 1);
	return 0;
}


int swis_GLineCount(void* External, int* ApproximateMaxLineLength) {
	int Dim, NumSize, MaxVal, MinVal, k;
	swiGRAPH* sG = (swiGRAPH*) External;
	Dim = GraphGetDim(sG->G);
	MaxVal = GraphGetMaxCoord(sG->G);
	MinVal = GraphGetMinCoord(sG->G);
	if (MaxVal > 0) NumSize = 1 + (int)log10(MaxVal);
	else if (MaxVal < 0) NumSize = 2 + (int)log10(-MaxVal);
	else NumSize = 1;
	if (MinVal > 0) k = 1 + (int)log10(MinVal);
	else if (MinVal < 0) k = 2 + (int)log10(-MinVal);
	else k = 1;
	if (k > NumSize) NumSize = k;
	k = GraphGetK(sG->G) + 1;
	sG->LineLength = (NumSize + 1) * Dim + k
		* (2 + (int)log10(GraphGetNumberVectors(sG->G))) + 10;
	if (sG->LineLength < 50) sG->LineLength = 50;
	*ApproximateMaxLineLength = sG->LineLength;
	if (sG->Line) deallocate(sG->Line);
	sG->Line = allocate(sG->LineLength + 1);
	return 1 + GraphGetNumberVectors(sG->G);
}

char* swis_GLine(void* External, int Line) {
	int k, Dim;
	char Buffer[32];
	int* Coords;
	GraphVector* From;
	GraphVector* To;
	swiGRAPH* sG = (swiGRAPH*) External;
	Dim = GraphGetDim(sG->G);
	if (Line > 1) {
		Line -= 2;
		sprintf(sG->Line, "%i (", Line);
		Dim = GraphGetDim(sG->G);
		Coords = GraphGetVectorCoord(GraphGetVector(sG->G, Line));
		for (k = 0; k < Dim - 1; ++k) {
			sprintf(Buffer, "%i, ", Coords[k]);
			strcat(sG->Line, Buffer);
		}
		sprintf(Buffer, "%i) ", Coords[Dim - 1]);
		strcat(sG->Line, Buffer);
		k = 0;
		From = GraphGetVector(sG->G, Line);
		To = GraphGetNNVector(sG->G, From, 0);
		while (To) {
			sprintf(Buffer, (k ? ", %i": "%i"), GraphGetVectorIndex(To));
			strcat(sG->Line, Buffer);
			++k;
			To = GraphGetNNVector(sG->G, From, k);
		}
	} else {
		sprintf(sG->Line, "%i nodes of dimensionality %i, max k: %i",
			GraphGetNumberVectors(sG->G), GraphGetDim(sG->G),
			GraphGetK(sG->G));
	}
	return sG->Line;
}


static int GImageParams[7] = { 4, 5, 6, 9, 10, 11, 12 };

SWI_ID swisCreateGraphClass() {
	SWI_ID Class, Filter;
	Filter = swiFLNew();
	swiFLAddOpenStrings(Filter, "k-NN graph", "*.dgr");
	swiFLAddStrings(Filter, "All files", "*");
	swiFLDefaultExtensionForSave(Filter, "dgr");
	Class = swiDCNew("k-NN graph", 0, SWI_ICON_GRAPH);
	swiDCDestroyFunc(Class, &swis_GDestroy);
	swiDCLoadFunc(Class, &swis_GLoad);
	swiDCSaveFunc(Class, &swis_GSave);
	swiDCTextFunc(Class, &swis_GLineCount, &swis_GLine);
	swiDCAskSaveOnExit(Class, 0);
	swiDCSetFilter(Class, Filter);
	swiDCSetViewParameters(Class, swiPSNew(7, GImageParams));
	return Class;
}
