/*---------------------------------------------------------------------

  TRACEPRG.C	Ismo Krkkinen

	General routines to save intermediate results from clustering 
	algorithms.

  -------------------------------------------------------------------*/

#define ProgName        "TRACEPRG"
#define VersionNumber   "Version 0.01"
#define LastUpdated     "4.5.2000"



#include "traceprg.h"
#include "stdio.h"

static int ClMin = -2;
static int ClMax = -2;
static int ClCount = -1;
static int ItMin = -2;
static int ItMax = -2;
static int ItCount = -1;
static char Prefix[128];

void InitializeTrace(int ClusterMin, int ClusterMax,
	int IterationMin, int IterationMax, char* FilenamePrefix, int TraceID)
{
	ClMin = ClusterMin;
	ClMax = ClusterMax;
	ItMin = IterationMin;
	ItMax = IterationMax;
	if (FilenamePrefix) sprintf(Prefix, "%s_%i", FilenamePrefix, TraceID);
	else sprintf(Prefix, "%i", TraceID);
}

void SetIteration(int Iter) {
	ItCount = Iter;
}

void SetClusterCount(int Count) {
	ClCount = Count;
}

static char buffer[512];

void SaveTraceData(TRAININGSET* TS, PARTITIONING* PA, CODEBOOK* CB,
	int PhaseNumber, char* PhaseName)
{
	if (ClMin <= ClCount && ClCount <= ClMax
		&& ItMin <= ItCount && ItCount <= ItMax)
	{
		if (TS && PA) {
			sprintf(buffer, "%s_%04i_%05i_%02i%s.pa", 
				Prefix, ClCount, ItCount, PhaseNumber, PhaseName);
			WritePartitioning(buffer, PA, TS, 1);
		}
		if (CB) {
			sprintf(buffer, "%s_%04i_%05i_%02i%s.cb", 
				Prefix, ClCount, ItCount, PhaseNumber, PhaseName);
			WriteCodebook(buffer, CB, 1);
		}
	}
}
