/********************************************************************

	idxqueue.c

	Ismo Krkkinen

	Last changed 30.11.1998

	Indexed queue, one can fetch items by using index. Items can be
	removed from or left to the queue when doing so.

********************************************************************/

#include "idxqueue.h"
#include "memctrl.h"

void InitIndexedQueue(INDEXEDQUEUE* IQ) {
	IQ->Head = 0;
	IQ->Tail = 0;
	IQ->Count = 0;
}

void DeleteIndexedQueue(INDEXEDQUEUE* IQ) {
	void* Out = DequeueFromIndexedQueue(IQ);
	while (Out) Out = DequeueFromIndexedQueue(IQ);
	IQ->Head = 0;
	IQ->Tail = 0;
	IQ->Count = 0;
}

void EnqueueIntoIndexedQueue(INDEXEDQUEUE* IQ, void* Item) {
	IDXQUEUEELEMENT* IQE, *PrevTail;
	IQE = allocate(sizeof(IDXQUEUEELEMENT));
	IQE->Data = Item;
	IQE->Next = 0;
	if (IQ->Count) {
		PrevTail = IQ->Tail;
		IQE->Prev = PrevTail;
		PrevTail->Next = IQE;
		IQ->Tail = IQE;
	} else {
		IQ->Head = IQ->Tail = IQE;
		IQE->Prev = 0;
	}
	IQ->Count++;
}

int NumberOfItemsInIndexedQueue(INDEXEDQUEUE* IQ) {
	return IQ->Count;
}

void* IndexedQueueItem(INDEXEDQUEUE* IQ, int Index, int Remove) {
	void* Item;
	IDXQUEUEELEMENT* Away;
	int Curr;
	if ((Index >= 0) && (IQ->Count > Index)) {
		Away = IQ->Head;
		Curr = 0;
		while (Curr++ < Index) Away = Away->Next;
		if (Remove) {
			if (Away->Prev) 
				((IDXQUEUEELEMENT*)Away->Prev)->Next = Away->Next;
			if (Away->Next) 
				((IDXQUEUEELEMENT*)Away->Next)->Prev = Away->Prev;
			if (Away == IQ->Head) {
				IQ->Head = Away->Next;
			}
			if (Away == IQ->Tail) {
				IQ->Tail = IQ->Tail->Prev;
			}
			IQ->Count--;
		}
		Item = Away->Data;
		if (Remove) deallocate(Away);
		return Item;
	} else return 0;
}
