#ifndef SWI_H
#define SWI_H

#include <windows.h>

typedef unsigned int SWI_ID;

/* Note: icons are loaded in swi_MainWindow.c, order must match. */
typedef enum {
	SWI_ICON_TRAININGSET = 0,
	SWI_ICON_CODEBOOK,
	SWI_ICON_PARTITIONING,
	SWI_ICON_GRAPH,
	SWI_ICON_COUNT
} swiIcons;

typedef struct {
	void* ViewData;
	SWI_ID ID;
	SWI_ID Class;
	void* External;
	HWND Handle;
	SWI_ID Params;
	char* ErrorString;
	CRITICAL_SECTION* CS;
	int ParamIndex;
	int ParamValue;
} AddInfo;

typedef struct {
	int x1, y1, x2, y2;
} swiDrawLineData;

typedef struct {
	int x, y;
} swiDrawCircleData;

typedef struct {
	int x1, y1, x2, y2;
	int Heads;
} swiDrawArrowData;

typedef struct {
	SWI_ID Params;
	SWI_ID ID;
	int LineCount, LineAllocated;
	swiDrawLineData* Lines;
	HGDIOBJ LinePen;
	int CircleCount, CircleAllocated;
	swiDrawCircleData* Circles;
	HGDIOBJ CirclePen, CircleBrush;
	int CircleRadius, CircleAbsoluteRadius;
	int ArrowCount, ArrowAllocated;
	swiDrawArrowData* Arrows;
	int ArrowLocation, ArrowHeadSize;
	HGDIOBJ ArrowPen;
} swiDrawingCommands;

typedef int (*SWI_INT_PVOID_PCHAR)(void*, char*);
typedef int (*SWI_INT_PPVOID_PCHAR)(void**, char*);
typedef int (*SWI_INT_PVOID)(void*);
typedef int (*SWI_INT_PVOID_INT)(void*, int);
typedef char* (*SWI_PCHAR_PVOID_INT)(void*, int);
typedef void (*SWI_INT_PVOID_PINT_PINT)(void*, int*, int*);
typedef unsigned int (*SWI_UINT_PVOID_INT_INT)(void*, int, int);
typedef int (*SWI_INT_SWI_ID)(SWI_ID);
typedef char* (*SWI_PCHAR_INT)(int);
typedef int (*SWI_INT_INT)(int);
typedef void (*SWI_VOID_INT_INT)(int, int);
typedef char* (*SWI_PCHAR_INT_INT)(int, int);
typedef int (*SWI_INT_INT_PSWI_ID_PPVOID_INT_PSWI_ID_PPVOID_PPVOID)
				(int, SWI_ID*, void**, int, SWI_ID*, void**, void**);
typedef int (*SWI_INT_INT_PPVOID_INT_PPVOID_PPVOID)(int, void**, int, void**, void**);
typedef void* (*SWI_PVOID_VOID)();
typedef void (*SWI_VOID_PVOID)(void*);
typedef int (*SWI_INT_PVOID_PVOID)(void*, void*);
typedef int (*SWI_INT_PPVOID_PCHAR_INT_PSWI_ID)(void**, char*, int, SWI_ID*);
typedef int (*SWI_INT_PVOID_PINT)(void*, int*);
typedef void (*SWI_VOID_PVOID_PVOID_SWI_ID)(void*, void*, SWI_ID);
typedef swiDrawingCommands* (*SWI_PSWIDRAWINGCOMMANDS_PVOID_PCRITICAL_SECTION)(void*, CRITICAL_SECTION*);
typedef swiDrawingCommands* (*SWI_PSWIDRAWINGCOMMANDS_PADDINFO)(AddInfo*);
typedef void (*SWI_VOID_PADDINFO)(AddInfo*);


extern void swiInitialize();
extern int swiRun(char* MainWindowTitle, char* ApplicationInfo, char* HelpFilename);
extern void swiFinalize();




/* data class functions */
extern SWI_ID swiDCNew(char* ClassName, SWI_ID Requirements, int Icon);
extern void swiDCLoadFunc(SWI_ID DataClassID, SWI_INT_PPVOID_PCHAR_INT_PSWI_ID LoadFunc);
extern void swiDCSaveFunc(SWI_ID DataClassID, SWI_INT_PVOID_PCHAR SaveFunc);
extern void swiDCDestroyFunc(SWI_ID DataClassID, SWI_INT_PVOID DestroyFunc);
extern void swiDCTextFunc(SWI_ID DataClassID, 
	SWI_INT_PVOID_PINT LineCountFunc, SWI_PCHAR_PVOID_INT LineFunc);
extern void swiDCAskSaveOnExit(SWI_ID DataClassID, int YesOrNo);

extern void swiDCSetFilter(SWI_ID DataClassID, SWI_ID FilterID);
extern void swiDCSetViewParameters(SWI_ID DataClassID, SWI_ID Params);

/* data set functions */
extern SWI_ID swiDSNew(SWI_ID DataClassID, void* External, char* SetName);
extern void swiDSUse(SWI_ID DataID);
extern void swiDSUnuse(SWI_ID DataID);
extern void* swiDSExternal(SWI_ID DataID);

/* filename filter functions */
extern SWI_ID swiFLNew();
extern void swiFLAddStrings(SWI_ID FilterID, char* Description, char* FilterString);
extern void swiFLAddOpenStrings(SWI_ID FilterID, char* Description, char* FilterString);
extern void swiFLDefaultExtensionForSave(SWI_ID FilterID, char* Extension);

/* requirement functions */

extern SWI_ID swiRQNew();
extern void swiRQAdd(SWI_ID RequirementID, SWI_ID DataClassID, int Count);


/* process type functions */

extern SWI_ID swiPTNew(SWI_ID ParentTypeID, char* Description);

/* process functions */

extern SWI_ID swiPCNew(
	SWI_ID ProcessTypeID,
	char* Description,
	SWI_INT_INT_PSWI_ID_PPVOID_INT_PSWI_ID_PPVOID_PPVOID PreprocessFunc,
	SWI_INT_INT_PPVOID_INT_PPVOID_PPVOID ProcessFunc,
	SWI_INT_INT_PSWI_ID_PPVOID_INT_PSWI_ID_PPVOID_PPVOID PostprocessFunc,
	SWI_ID ParameterSetID,
	SWI_ID InputRequirements,
	SWI_ID OutputRequirements);

extern int swiPCSetMenuInfo(SWI_ID ProcessClassID, int Key, int Special, 
	int Ctrl, int Alt, int Shift, SWI_ID InMenuAfterProcess);

extern void swiPCLink(SWI_ID ProcessClassID, SWI_ID LinkTo);


/* parameter set functions */

extern void swiFullParameterSet(
	int FirstOption, int LastOption,
	SWI_PCHAR_INT DocumentFunc,
	SWI_INT_INT TypeFunc,
	SWI_INT_INT GetValueFunc,
	SWI_VOID_INT_INT SetValueFunc,
	SWI_INT_INT MaxValueFunc,
	SWI_INT_INT MinValueFunc,
	SWI_INT_INT DefaultValueFunc,
	SWI_PCHAR_INT_INT ValueNameFunc);

extern SWI_ID swiPSNew(int ParameterCount, int* Parameters);


/* image view functions and types */

extern swiDrawingCommands* swiDRNew(SWI_ID ID);
extern void swiDRDelete(swiDrawingCommands* DC);
extern void swiDRLine(swiDrawingCommands* DC, int x1, int y1, int x2, int y2);
extern void swiDRCircle(swiDrawingCommands* DC, int x, int y);
extern void swiDRArrow(swiDrawingCommands* DC, int x1, int y1, int x2, int y2,
	int ForwardArrow, int BackwardArrow);

extern void swiIVFunctions(SWI_PVOID_VOID CreateFunc, SWI_VOID_PVOID DestroyFunc,
	SWI_PSWIDRAWINGCOMMANDS_PADDINFO AddFunc, SWI_VOID_PADDINFO RemoveFunc,
	SWI_PSWIDRAWINGCOMMANDS_PVOID_PCRITICAL_SECTION SpecialFunc);
extern void swiIVSetParameters(SWI_ID ParameterSetID, char* SpecialDescription);
extern void swiIVRemove(HWND ImageViewWindow, SWI_ID DataSetID);
extern void swiIVUpdateSpecial(HWND ImageViewWindow);

#endif
