/********************************************************************

	swi_DataSettextWindow.c

	Ismo Krkkinen

	Last changed 16.11.1999

********************************************************************/

#include "swi_WindowsControls.h"
#include "swi_DataSetTextWindow.h"
#include "swi_WindowsText.h"
#include "memctrl.h"
#include "swi_Internal.h"
#include "swi_DragAndDrop.h"
#include "swi_DataClass.h"
#include "swi_WindowsText.h"
/*#include "swi_OwnedWindow.h"*/

typedef struct {
	SWI_ID ID;
/*	SWI_ID OWID;*/
	int VertBar;
	int HorzBar;
	int Width, Height;
	int MHeight, MWidth;
	int VSMax, HSMax;
	int VSPos, HSPos;
	int VPage, HPage;
	int MaxExtent, MEAdd;
	int ApproxMaxLineLength;
} DSTWExtra;

static SWI_ID NewWindowDataID = 0;

static void FillDSTWExtra(HWND Handle, DSTWExtra* E) {
	HDC DC;
	TEXTMETRIC tm;
	RECT Area;
	int k, Width;
	E->ID = NewWindowDataID;
	E->VertBar = E->HorzBar = 1;
	GetClientRect(Handle, &Area);
	E->Width = Area.right; /* + GetSystemMetrics(SM_CXVSCROLL);*/
	E->Height = Area.bottom; /* + GetSystemMetrics(SM_CYHSCROLL);*/
	E->VSPos = E->HSPos = 1;
	DC = GetDC(Handle);
	SetMapMode(DC, MM_TEXT);
	SetBkMode(DC, TRANSPARENT);
	swiWTAttachFont(DC, swiWTStockFont(1));
	GetTextMetrics(DC, &tm);
	E->MHeight = tm.tmHeight + tm.tmExternalLeading;
	E->MWidth = tm.tmAveCharWidth;
	E->VSMax = swiDSLineCount(E->ID, &(E->ApproxMaxLineLength)) + 2;
	E->HSMax = 1;
	E->MaxExtent = 0;
	E->MEAdd = 2 * E->MWidth + E->MWidth / 2;
	/* might cause problems with delay with large number of lines */
	for (k = 1; k < E->VSMax - 2; k++) {
		swiWTExtent(DC, swiDSLine(E->ID, k), 0, &Width);
		if (Width > E->MaxExtent) E->MaxExtent = Width;
	}
}

static void SetBars(HWND Window, DSTWExtra* E) {
	int sbh, sbw, Ext;
	HDC DC;
	DC = GetDC(Window);
	sbh = GetSystemMetrics(SM_CYHSCROLL);
	sbw = GetSystemMetrics(SM_CXVSCROLL);
	if (E->VertBar) E->Width += sbw;
	if (E->HorzBar) E->Height += sbh;
	if (E->VSMax >= (E->Height + E->MHeight - 1) / E->MHeight) {
		E->Width -= sbw;
		E->VertBar = 1;
	} else E->VertBar = 0;
	Ext = E->MaxExtent + E->MEAdd;
	if (Ext > E->Width) {
		E->HSMax = (Ext + E->MWidth - 1) / E->MWidth;
		if (!E->HorzBar) E->HSPos = 1;
		E->HorzBar = 1;
		E->Height -= sbh;
		/* now shallower, check if vertical bar needed now */
		if (!E->VertBar 
			&& (E->VSMax >= (E->Height + E->MHeight - 1) / E->MHeight))
		{
			E->Width -= sbw;
			E->VertBar = 1;
		}
	} else {
		E->HorzBar = 0;
		E->HSMax = 1;
	}
	if (E->VSMax * E->MHeight < E->Height) E->VPage = E->VSMax;
	else E->VPage = E->Height / E->MHeight;
	E->VSPos = max(1, E->VSPos);
	E->VSPos = max(E->VSPos, min(E->VSPos, E->VSMax - max(E->VPage, 1)));
	swiSBSize(Window, SB_VERT, 1, max(1, E->VSMax), E->VPage);
	if (E->HSMax * E->MWidth < E->Width) E->HPage = E->HSMax;
	else E->HPage = E->Width / E->MWidth;
	E->HSPos = max(1, E->HSPos);
	E->HSPos = max(E->HSPos, min(E->HSPos, E->HSMax - max(E->HPage, 1)));
	swiSBSize(Window, SB_HORZ, 1, max(1, E->HSMax), E->HPage);
}

LRESULT CALLBACK TextWindowProcedure(HWND Handle, UINT Message,
	WPARAM wParam, LPARAM lParam)
{
	RECT Area, WorkArea;
	int OldPos, k, Begin, End, Top;
	int Width, Height, newX, newY, newWidth, newHeight;
	HDC DC;
	WORD ScrollNotify;
	UINT ScrollMessage;
	DSTWExtra* E = (DSTWExtra*) GetWindowLong(Handle, 0);
	switch (Message) {
	case WM_CREATE:
		E = allocate(sizeof(DSTWExtra));
/*		E->OWID = swiOWNew(Handle);*/
		FillDSTWExtra(Handle, E);
		SetWindowLong(Handle, 0, (LONG) E);
		if (SystemParametersInfo(SPI_GETWORKAREA, 0, &WorkArea, 0)) {
			Width = WorkArea.right - (WorkArea.left + 50);
			Height = WorkArea.bottom - (WorkArea.top + 50);
			newWidth = max(GetSystemMetrics(SM_CXMINTRACK),
				min(E->MWidth * E->ApproxMaxLineLength + 30, Width));
			newHeight = min(Height,
				E->VSMax * E->MHeight 
				+ GetSystemMetrics(SM_CYMINTRACK) + 20);
			GetWindowRect(Handle, &Area);
			newX = (Area.left + newWidth > Width) ? 
				Width - newWidth : Area.left;
			newY = (Area.top + newHeight > Height) ? 
				Height - newHeight : Area.top;
			MoveWindow(Handle, newX, newY, newWidth, newHeight, FALSE);
		}
		SetBars(Handle, E);
		SendMessage(swiMainWindowHandle(), 
			SWI_WM_ADD_TO_WINDOW_MENU, 0, (LPARAM) Handle);
		return 0;
	case WM_ACTIVATE:
		swiDDEnd();
		return 0;
	case WM_MOUSEMOVE:
		if (swiDDDragging() && (wParam & MK_LBUTTON != MK_LBUTTON))
			swiDDEnd();
		break;
	case WM_PAINT:
		if (GetUpdateRect(Handle, &Area, TRUE)) {
			DC = GetDC(Handle);
			Begin = E->VSPos + Area.top / E->MHeight;
			End = E->VSPos + (Area.bottom + E->MHeight - 1) / E->MHeight;
			Top = Area.top - Area.top % E->MHeight;
			ShowCursor(FALSE);
			for (k = Begin; k <= min(End, E->VSMax - 2); k++) swiWTText(DC,
				swiDSLine(E->ID, k), Top + (k - Begin) * E->MHeight,
				(E->HSPos - 1) * E->MWidth - E->MWidth / 2,
				E->MHeight, E->Width);
			ValidateRect(Handle, &Area);
			ShowCursor(TRUE);
		}
		return 0;
	case WM_SIZE:
		if ((wParam != SIZE_MAXHIDE) && (wParam != SIZE_MINIMIZED)) {
			E->Width = LOWORD(lParam);
			E->Height = HIWORD(lParam);
			SetBars(Handle, E);
		}
		return 0;
	case WM_VSCROLL:
		OldPos = E->VSPos;
		E->VSPos = swiSBMove(Handle, SB_VERT, LOWORD(wParam), 1, E->VPage,
			HIWORD(wParam));
		if (E->VSPos != OldPos) {
			GetClientRect(Handle, &Area);
			ScrollWindowEx(Handle, 0, E->MHeight * (OldPos - E->VSPos),
				0, &Area, 0, 0, SW_ERASE | SW_INVALIDATE);
		}
		return 0;
	case WM_HSCROLL:
		OldPos = E->HSPos;
		E->HSPos = swiSBMove(Handle, SB_HORZ, LOWORD(wParam), 1, E->HPage,
			HIWORD(wParam));
		if (E->HSPos != OldPos) {
			GetClientRect(Handle, &Area);
			ScrollWindowEx(Handle, E->MWidth * (OldPos - E->HSPos), 0,
				0, &Area, 0, 0, SW_ERASE | SW_INVALIDATE);
		}
		return 0;
	case WM_KEYDOWN:
		switch (wParam) {
		case VK_UP:
			ScrollNotify = SB_LINEUP;
			ScrollMessage = WM_VSCROLL;
			PostMessage(Handle, 
				ScrollMessage, MAKELONG(ScrollNotify, 0), 0);
			break;
		case VK_PRIOR:
			ScrollNotify = SB_PAGEUP;
			ScrollMessage = WM_VSCROLL;
			PostMessage(Handle, 
				ScrollMessage, MAKELONG(ScrollNotify, 0), 0);
			break;
		case VK_NEXT:
			ScrollNotify = SB_PAGEDOWN;
			ScrollMessage = WM_VSCROLL;
			PostMessage(Handle, 
				ScrollMessage, MAKELONG(ScrollNotify, 0), 0);
			break;
		case VK_DOWN:
			ScrollNotify = SB_LINEDOWN;
			ScrollMessage = WM_VSCROLL;
			PostMessage(Handle, 
				ScrollMessage, MAKELONG(ScrollNotify, 0), 0);
			break;
		case VK_HOME:
			ScrollNotify = SB_TOP;
			ScrollMessage = WM_VSCROLL;
			PostMessage(Handle, 
				ScrollMessage, MAKELONG(ScrollNotify, 0), 0);
			ScrollNotify = SB_TOP;
			ScrollMessage = WM_HSCROLL;
			PostMessage(Handle, 
				ScrollMessage, MAKELONG(ScrollNotify, 0), 0);
			break;
		case VK_END:
			ScrollNotify = SB_BOTTOM;
			ScrollMessage = WM_VSCROLL;
			PostMessage(Handle, 
				ScrollMessage, MAKELONG(ScrollNotify, 0), 0);
			ScrollNotify = SB_BOTTOM;
			ScrollMessage = WM_HSCROLL;
			PostMessage(Handle, 
				ScrollMessage, MAKELONG(ScrollNotify, 0), 0);
			break;
		case VK_LEFT:
			ScrollNotify = SB_LINELEFT;
			ScrollMessage = WM_HSCROLL;
			PostMessage(Handle, 
				ScrollMessage, MAKELONG(ScrollNotify, 0), 0);
			break;
		case VK_RIGHT:
			ScrollNotify = SB_LINERIGHT;
			ScrollMessage = WM_HSCROLL;
			PostMessage(Handle, 
				ScrollMessage, MAKELONG(ScrollNotify, 0), 0);
			break;
		default: return DefWindowProc(Handle, Message, wParam, lParam);
		}
		return 0;
	case WM_DESTROY:
/*		swiOWDelete(E->OWID);*/
		swiDSTextWindowClosed(E->ID);
		deallocate(E);
		SendMessage(swiMainWindowHandle(), 
			SWI_WM_REMOVE_FROM_WINDOW_MENU, 0, (LPARAM) Handle);
		return 0;
	case SWI_WM_RENAMED:
		SendMessage(swiMainWindowHandle(), 
			SWI_WM_REMOVE_FROM_WINDOW_MENU, 0, (LPARAM) Handle);
		SetWindowText(Handle, swiFilename(swiDSName(E->ID)));
		SendMessage(swiMainWindowHandle(), 
			SWI_WM_ADD_TO_WINDOW_MENU, 0, (LPARAM) Handle);
		return 0;
	}
	return DefWindowProc(Handle, Message, wParam, lParam);
}

static void RegisterTextWindowClass() {
	WNDCLASS W;
	W.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	W.lpfnWndProc = &TextWindowProcedure;
	W.cbClsExtra = 0;
	W.cbWndExtra = sizeof(DSTWExtra*);
	W.hInstance = GetModuleHandle(0);
	W.hIcon = swiAppIcon();
	W.hCursor = 0; /*LoadCursor(NULL, IDC_ARROW);*/
	W.hbrBackground = (HBRUSH) (COLOR_WINDOW + 1);
	W.lpszMenuName = 0;
	W.lpszClassName = "swiDataSetTextWindow";
	RegisterClass(&W);
}


static int DSTWRegistered = 0;

HWND swiDSTWCreate(SWI_ID DataID) {
	HWND Handle;
	if (!DSTWRegistered) {
		DSTWRegistered = 1;
		RegisterTextWindowClass();
	}
	NewWindowDataID = DataID;
	Handle = CreateWindow("swiDataSetTextWindow", 
		swiFilename(swiDSName(DataID)),
		WS_OVERLAPPEDWINDOW | WS_HSCROLL | WS_VSCROLL,
		CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
		swiMainWindowHandle(), 0, GetModuleHandle(0), 0);
	ShowWindow(Handle, SW_SHOW);
	UpdateWindow(Handle);
	return Handle;
}

void swiDSTWDestroy(HWND Handle) {
	DestroyWindow(Handle);
}
