/********************************************************************

	swi_ImageView.c

	Ismo Krkkinen

	Last changed 4.9.2003

********************************************************************/

#include "swi_ImageView.h"
#include "swi_DragAndDrop.h"
#include "swi_WindowsControls.h"
#include "memctrl.h"
#include "swi_Internal.h"
#include "swi_Parameters.h"
#include <string.h>
#include "image.h"
#include "swi_SelectedData.h"
#include "swi_WindowsDialogs.h"
#include "swi_MessageBoxes.h"
#include <commctrl.h>
#include "resource.h"
#include "swi_Application.h"
#include <process.h>
#include "swi_DataClass.h"
#include <math.h>
#include <limits.h>

#define FWD_ARROW 1
#define BWD_ARROW 2

#define ZOOM_IN_SCALE 0.666666f
#define ZOOM_OUT_SCALE 1.5f
#define VIEW_MOVE_AMOUNT 0.125f

#define ListViewWidth 90

static void MoveVisibleRelative(DrawArea* Visible, float DX, float DY) {
	DX *= Visible->Width;
	DY *= Visible->Height;
	Visible->Left += DX;
	if (Visible->Left + Visible->Width > 1.0f)
		Visible->Left = 1.0f - Visible->Width;
	if (Visible->Left < 0.0f) Visible->Left = 0.0f;
	Visible->Bottom += DY;
	if (Visible->Bottom + Visible->Height > 1.0f)
		Visible->Bottom = 1.0f - Visible->Height;
	if (Visible->Bottom < 0.0f) Visible->Bottom = 0.0f;
}

static void ScaleVisible(DrawArea* Visible, float Scale) {
	float tmp = Visible->Height;
	Visible->Height *= Scale;
	if (Visible->Height > 1.0f) Visible->Height = 1.0f;
	if (Visible->Height < 1e-5f) Visible->Height = 1e-5f;
	tmp = (tmp - Visible->Height) * 0.5f;
	Visible->Bottom += tmp;
	tmp = Visible->Width;
	Visible->Width *= Scale;
	if (Visible->Width > 1.0f) Visible->Width = 1.0f;
	if (Visible->Width < 1e-5f) Visible->Width = 1e-5f;
	tmp = (tmp - Visible->Width) * 0.5f;
	Visible->Left += tmp;
	if (Visible->Left < 0) Visible->Left = 0;
	if (Visible->Left + Visible->Width > 1.0f)
		Visible->Left = 1.0f - Visible->Width;
	if (Visible->Bottom < 0) Visible->Bottom = 0;
	if (Visible->Bottom + Visible->Height > 1.0f)
		Visible->Bottom = 1.0f - Visible->Height;
}

static void AdjustVisible(DrawArea* Visible, DrawArea* Adjusted, int W, int H)
{
	float Ratio, tmp;
	Adjusted->Left = Visible->Left;
	Adjusted->Bottom = Visible->Bottom;
	Adjusted->Width = Visible->Width;
	Adjusted->Height = Visible->Height;
	if (W > H) {
		Ratio = (float)W/(float)H;
		tmp = Adjusted->Width;
		Adjusted->Width *= Ratio;
		Adjusted->Left -= (Adjusted->Width - tmp) * 0.5f;
	} else {
		Ratio = (float)H/(float)W;
		tmp = Adjusted->Height;
		Adjusted->Height *= Ratio;
		Adjusted->Bottom -= (Adjusted->Height - tmp) * 0.5f;
	}
}

static void ComputeScale(DrawArea* Adjusted, DrawArea* Full,
	DrawArea* ShiftScale, int W, int H)
{
	ShiftScale->Left = -(Full->Left + Adjusted->Left * Full->Width);
	ShiftScale->Bottom = -(Full->Bottom + Adjusted->Bottom * Full->Height);
	ShiftScale->Width = W / (Full->Width * Adjusted->Width);
	ShiftScale->Height = H / (Full->Height * Adjusted->Height);
}

static void FindMinMax(swiDrawingCommands* DC, int* MinVal, int* MaxVal) {
	int k;
	if (DC->LineCount)
		for (k = 0; k < DC->LineCount; k++) {
			if (DC->Lines[k].x1 < *MinVal) *MinVal = DC->Lines[k].x1;
			if (DC->Lines[k].y1 < *MinVal) *MinVal = DC->Lines[k].y1;
			if (DC->Lines[k].x2 < *MinVal) *MinVal = DC->Lines[k].x2;
			if (DC->Lines[k].y2 < *MinVal) *MinVal = DC->Lines[k].y2;
			if (DC->Lines[k].x1 > *MaxVal) *MaxVal = DC->Lines[k].x1;
			if (DC->Lines[k].y1 > *MaxVal) *MaxVal = DC->Lines[k].y1;
			if (DC->Lines[k].x2 > *MaxVal) *MaxVal = DC->Lines[k].x2;
			if (DC->Lines[k].y2 > *MaxVal) *MaxVal = DC->Lines[k].y2;
		}
	if (DC->ArrowCount)
		for (k = 0; k < DC->ArrowCount; ++k) {
			if (DC->Arrows[k].x1 < *MinVal) *MinVal = DC->Arrows[k].x1;
			if (DC->Arrows[k].y1 < *MinVal) *MinVal = DC->Arrows[k].y1;
			if (DC->Arrows[k].x2 < *MinVal) *MinVal = DC->Arrows[k].x2;
			if (DC->Arrows[k].y2 < *MinVal) *MinVal = DC->Arrows[k].y2;
			if (DC->Arrows[k].x1 > *MaxVal) *MaxVal = DC->Arrows[k].x1;
			if (DC->Arrows[k].y1 > *MaxVal) *MaxVal = DC->Arrows[k].y1;
			if (DC->Arrows[k].x2 > *MaxVal) *MaxVal = DC->Arrows[k].x2;
			if (DC->Arrows[k].y2 > *MaxVal) *MaxVal = DC->Arrows[k].y2;
		}
	if (DC->CircleCount)
		for (k = 0; k < DC->CircleCount; k++) {
			if (DC->Circles[k].x < *MinVal) *MinVal = DC->Circles[k].x;
			if (DC->Circles[k].y < *MinVal) *MinVal = DC->Circles[k].y;
			if (DC->Circles[k].x > *MaxVal) *MaxVal = DC->Circles[k].x;
			if (DC->Circles[k].y > *MaxVal) *MaxVal = DC->Circles[k].y;
		}
}

static void FindLimits(swiDrawingCommands* DC1, int Count, 
	swiDrawingCommands** DCs, DrawArea* Full)
{
	int MinVal, MaxVal, k, VMinVal, VMaxVal;
	MinVal = VMinVal = INT_MAX;
	MaxVal = VMaxVal = INT_MIN;
	if (DC1) FindMinMax(DC1, &VMinVal, &VMaxVal);
	for (k = 0; k < Count; ++k) FindMinMax(DCs[k], &MinVal, &MaxVal);
	if (VMinVal < MinVal && MaxVal < VMaxVal) {
		Full->Height = Full->Width = (float)(VMaxVal - VMinVal + 1);
		Full->Bottom = Full->Left = (float)VMinVal;
	} else {
		if (VMinVal < MinVal) MinVal = VMinVal;
		if (MaxVal < VMaxVal) MaxVal = VMaxVal;
		Full->Height = Full->Width = (MaxVal - MinVal + 1) * 1.06f;
		Full->Bottom = Full->Left = MinVal - 0.03f * (Full->Width / 1.06f);
	}
}

swiDrawingCommands* swiDRNew(SWI_ID ID) {
	swiDrawingCommands* DC = allocate(sizeof(swiDrawingCommands));
	DC->ID = ID;
	DC->Params = 0;
	DC->LineCount = DC->LineAllocated = 0;
	DC->LinePen = 0;
	DC->Lines = 0;
	DC->CircleCount = DC->CircleAllocated = 0;
	DC->CirclePen = 0;
	DC->CircleBrush = 0;
	DC->CircleRadius = 0;
	DC->Circles = 0;
	DC->CircleAbsoluteRadius = 0;
	DC->ArrowCount = DC->ArrowAllocated = 0;
	DC->Arrows = 0;
	DC->ArrowPen = 0;
	DC->ArrowLocation = DC->ArrowHeadSize = 0;
	return DC;
}

void swiDRDelete(swiDrawingCommands* DC) {
	deallocate(DC->Lines);
	deallocate(DC->Circles);
	deallocate(DC->Arrows);
	if (DC->LinePen) DeleteObject(DC->LinePen);
	if (DC->CircleBrush) DeleteObject(DC->CircleBrush);
	if (DC->CirclePen) DeleteObject(DC->CirclePen);
	if (DC->ArrowPen) DeleteObject(DC->ArrowPen);
	if (DC->Params) swiPSDelete(DC->Params);
	if (DC->ID) swiDSUnuse(DC->ID);
	deallocate(DC);
}

void swiDRLineAttributes(swiDrawingCommands* DC, int Color, int Dotted) {
	if (DC->LinePen) DeleteObject(DC->LinePen);
	DC->LinePen = CreatePen(Dotted ? PS_DOT : PS_SOLID, 1, Color);
}

void swiDRLine(swiDrawingCommands* DC, int x1, int y1, int x2, int y2) {
	int k;
	swiDrawLineData* newLD;
	if (DC->LineCount >= DC->LineAllocated) {
		DC->LineAllocated += 512;
		newLD = allocate(sizeof(swiDrawLineData) * DC->LineAllocated);
		if (DC->Lines) {
			for (k = 0; k < DC->LineCount; k++) {
				newLD[k].x1 = DC->Lines[k].x1;
				newLD[k].y1 = DC->Lines[k].y1;
				newLD[k].x2 = DC->Lines[k].x2;
				newLD[k].y2 = DC->Lines[k].y2;
			}
			deallocate(DC->Lines);
		}
		DC->Lines = newLD;
	}
	DC->Lines[DC->LineCount].x1 = x1;
	DC->Lines[DC->LineCount].y1 = y1;
	DC->Lines[DC->LineCount].x2 = x2;
	DC->Lines[DC->LineCount].y2 = y2;
	DC->LineCount++;
}

void swiDRCircleAttributes(swiDrawingCommands* DC, 
	int Color, int Filled, int Radius, int Absolute)
{
	LOGBRUSH B;
	if (DC->CircleBrush) DeleteObject(DC->CircleBrush);
	if (DC->CirclePen) DeleteObject(DC->CirclePen);
	DC->CirclePen = CreatePen(PS_SOLID, 1, Color);
	if (Color) B.lbColor = Color;
	else B.lbColor = 0x0ffffff;
	B.lbHatch = 0;
	B.lbStyle = Filled ? BS_SOLID : BS_HOLLOW;
	DC->CircleBrush = CreateBrushIndirect(&B);
	DC->CircleRadius = Radius;
	DC->CircleAbsoluteRadius = Absolute;
}

void swiDRCircle(swiDrawingCommands* DC, int x, int y) {
	int k;
	swiDrawCircleData* newCD;
	if (DC->CircleCount >= DC->CircleAllocated) {
		DC->CircleAllocated += 512;
		newCD = allocate(sizeof(swiDrawCircleData) * DC->CircleAllocated);
		if (DC->Circles) {
			for (k = 0; k < DC->CircleCount; k++) {
				newCD[k].x = DC->Circles[k].x;
				newCD[k].y = DC->Circles[k].y;
			}
			deallocate(DC->Circles);
		}
		DC->Circles = newCD;
	}
	DC->Circles[DC->CircleCount].x = x;
	DC->Circles[DC->CircleCount].y = y;
	DC->CircleCount++;
}

void swiDRArrowAttributes(swiDrawingCommands* DC,
	int Location, int ArrowHeadSize, int Color, int Dotted, int Thickness)
{
	DC->ArrowLocation = Location;
	DC->ArrowHeadSize = ArrowHeadSize;
	if (DC->ArrowPen) DeleteObject(DC->ArrowPen);
	if (Dotted) DC->ArrowPen = CreatePen(PS_DOT, 1, Color);
	else DC->ArrowPen = CreatePen(PS_SOLID, Thickness, Color);
}

void swiDRArrow(swiDrawingCommands* DC, int x1, int y1, int x2, int y2,
	int ForwardArrow, int BackwardArrow)
{
	swiDrawArrowData* newAD;
	if (DC->ArrowCount >= DC->ArrowAllocated) {
		DC->ArrowAllocated += 512;
		newAD = allocate(sizeof(swiDrawArrowData) * DC->ArrowAllocated);
		if (DC->Arrows) {
			memcpy(newAD, DC->Arrows, 
				sizeof(swiDrawArrowData) * DC->ArrowCount);
			deallocate(DC->Arrows);
		}
		DC->Arrows = newAD;
	}
	DC->Arrows[DC->ArrowCount].x1 = x1;
	DC->Arrows[DC->ArrowCount].y1 = y1;
	DC->Arrows[DC->ArrowCount].x2 = x2;
	DC->Arrows[DC->ArrowCount].y2 = y2;
	DC->Arrows[DC->ArrowCount].Heads =
		(ForwardArrow ? FWD_ARROW : 0) | (BackwardArrow ? BWD_ARROW : 0);
	DC->ArrowCount++;
}

void swiDRDraw(swiDrawingCommands* DC, 
	HDC DeviceContext, DrawArea* SS, int Height)
{
	int k, XRadius, YRadius;
	float OrtoX, OrtoY, ParX, ParY, tmp, LocX, LocY;
	HGDIOBJ OrigPen, OrigBrush;
	OrigPen = GetCurrentObject(DeviceContext, OBJ_PEN);
	OrigBrush = GetCurrentObject(DeviceContext, OBJ_BRUSH);
	if (DC->LineCount) {
		SelectObject(DeviceContext, DC->LinePen);
		for (k = 0; k < DC->LineCount; k++) {
			MoveToEx(DeviceContext,
				(int)((DC->Lines[k].x1 + SS->Left) * SS->Width),
				Height - (int)((DC->Lines[k].y1 + SS->Bottom) * SS->Height),
				0);
			LineTo(DeviceContext,
				(int)((DC->Lines[k].x2 + SS->Left) * SS->Width),
				Height - (int)((DC->Lines[k].y2 + SS->Bottom) * SS->Height));
		}
	}
	if (DC->ArrowCount) {
		SelectObject(DeviceContext, DC->ArrowPen);
		for (k = 0; k < DC->ArrowCount; ++k) {
			MoveToEx(DeviceContext,
				(int)((DC->Arrows[k].x1 + SS->Left) * SS->Width),
				Height - (int)((DC->Arrows[k].y1 + SS->Bottom) * SS->Height),
				0);
			LineTo(DeviceContext,
				(int)((DC->Arrows[k].x2 + SS->Left) * SS->Width),
				Height - (int)((DC->Arrows[k].y2 + SS->Bottom) * SS->Height));
			/*ErrorMessage("line: %i %i %i %i\n",
				DC->Arrows[k].x1, DC->Arrows[k].y1,
				DC->Arrows[k].x2, DC->Arrows[k].y2);*/
			/* Basis vectors. */
			ParX = (float)(DC->Arrows[k].x2 - DC->Arrows[k].x1);
			ParY = (float)(DC->Arrows[k].y2 - DC->Arrows[k].y1);
			if (ParX == 0 && ParY == 0) continue;
			/*ErrorMessage("%f %f %f\n", ParX * ParX, ParY * ParY,
				ParX * ParX + ParY * ParY);*/
			/*tmp = sqrt(ParX * ParX + ParY * ParY);*/
			tmp = (float)pow(ParX * ParX + ParY * ParY, 0.5);
			/*ErrorMessage("%f %f %i %f\n", ParX, ParY, DC->ArrowHeadSize, tmp);*/
			/*ParX *= DC->ArrowHeadSize / tmp;
			ParY *= DC->ArrowHeadSize / tmp;*/
			ParX *= (DC->ArrowHeadSize / 1000) / tmp;
			ParY *= (DC->ArrowHeadSize / 1000) / tmp;
			OrtoX = -ParY * 0.58f;
			OrtoY = ParX * 0.58f;
			if (DC->Arrows[k].Heads & FWD_ARROW) { /* Forward arrow. */
				/*LocX = DC->Arrows[k].x1
					+ (DC->Arrows[k].x2 - DC->Arrows[k].x1)
					* (100 - DC->ArrowLocation) / 100.0f;
				LocY = DC->Arrows[k].y1
					+ (DC->Arrows[k].y2 - DC->Arrows[k].y1)
					* (100 - DC->ArrowLocation) / 100.0f;*/
				LocX = (DC->Arrows[k].x2 + SS->Left) * SS->Width
					- (ParX / (DC->ArrowHeadSize / 1000)) * DC->ArrowLocation;
				LocY = (DC->Arrows[k].y2 + SS->Bottom) * SS->Height
					- (ParY / (DC->ArrowHeadSize / 1000)) * DC->ArrowLocation;
				/*LocX += -ParX + OrtoX;
				LocY += -ParY + OrtoY;*/
				MoveToEx(DeviceContext,	(int)(LocX - ParX + OrtoX),
					Height - (int)(LocY - ParY + OrtoY), 0);
				/*LocX += ParX - OrtoX;
				LocY += ParY - OrtoY;*/
				LineTo(DeviceContext, (int)LocX, Height - (int)LocY);
				/*LocX += -ParX - OrtoX;
				LocY += -ParY - OrtoY;*/
				LineTo(DeviceContext, (int)(LocX - ParX - OrtoX),
					Height - (int)(LocY - ParY - OrtoY));
			}
			if (DC->Arrows[k].Heads & BWD_ARROW) { /* Backward arrow. */
				/*LocX = DC->Arrows[k].x1
					+ (DC->Arrows[k].x2 - DC->Arrows[k].x1)
					* DC->ArrowLocation / 100.0f;
				LocY = DC->Arrows[k].y1
					+ (DC->Arrows[k].y2 - DC->Arrows[k].y1)
					* DC->ArrowLocation / 100.0f;*/
				LocX = (DC->Arrows[k].x1 + SS->Left) * SS->Width
					+ (ParX / (DC->ArrowHeadSize / 1000)) * DC->ArrowLocation;
				LocY = (DC->Arrows[k].y1 + SS->Bottom) * SS->Height
					+ (ParY / (DC->ArrowHeadSize / 1000)) * DC->ArrowLocation;
				/*LocX += ParX + OrtoX;
				LocY += ParY + OrtoY;*/
				MoveToEx(DeviceContext, (int)(LocX + ParX + OrtoX),
					Height - (int)(LocY + ParY + OrtoY), 0);
				/*LocX += -ParX - OrtoX;
				LocY += -ParY - OrtoY;*/
				LineTo(DeviceContext, (int)LocX, Height - (int)LocY);
				/*LocX += ParX - OrtoX;
				LocY += ParY - OrtoY;*/
				LineTo(DeviceContext, (int)(LocX + ParX - OrtoX),
					Height - (int)(LocY + ParY - OrtoY));
			}
		}
	}
	if (DC->CircleCount) {
		XRadius = DC->CircleAbsoluteRadius ? 
			DC->CircleRadius / 1000 : (int)(DC->CircleRadius * SS->Width);
		YRadius = DC->CircleAbsoluteRadius ? 
			DC->CircleRadius / 1000 : (int)(DC->CircleRadius * SS->Height);
		SelectObject(DeviceContext, DC->CirclePen);
		SelectObject(DeviceContext, DC->CircleBrush);
		for (k = 0; k < DC->CircleCount; k++) Ellipse(DeviceContext, 
			(int)((DC->Circles[k].x + SS->Left) * SS->Width - XRadius),
			Height - (int)((DC->Circles[k].y + SS->Bottom) * SS->Height - YRadius),
			(int)((DC->Circles[k].x + SS->Left) * SS->Width + XRadius),
			Height - (int)((DC->Circles[k].y + SS->Bottom) * SS->Height + YRadius));
	}
	SelectObject(DeviceContext, OrigPen);
	SelectObject(DeviceContext, OrigBrush);
}

static int ToColor(int ColorEnum) {
	switch (ColorEnum) {
	case 0: return 0x00ff0000;
	case 1: return 0x0000ff00;
	case 2: return 0x000000ff;
	case 3: return 0x00ff00ff;
	case 4: return 0x00ffff00;
	case 5: return 0x0000ffff;
	case 6: return 0x00ff7f7f;
	case 7: return 0x007fff7f;
	case 8: return 0x007f7fff;
	case 9: return 0x00000000;
	}
	return 0;
}

typedef struct {
	SWI_PVOID_VOID Create;
	SWI_VOID_PVOID Destroy;
	SWI_PSWIDRAWINGCOMMANDS_PADDINFO Add;
	SWI_VOID_PADDINFO Remove;
	SWI_PSWIDRAWINGCOMMANDS_PVOID_PCRITICAL_SECTION Special;
	SWI_ID Params;
	char* Description;
} swiImageFunctions;

static swiImageFunctions IVF;

typedef struct {
	void* ViewData;
	HWND Handle;
	int UpdateCounter;
	CRITICAL_SECTION* CS;
} SpecialInfo;

void __cdecl SpecialWrapper(void* SpecialInfoPtr) {
	SpecialInfo* SI = SpecialInfoPtr;
	swiDrawingCommands* DCs;
	if (IVF.Special) {
		DCs = (*IVF.Special)(SI->ViewData, SI->CS);
		if (IsWindow(SI->Handle)) PostMessage(SI->Handle, 
			SWI_WM_IV_NEW_SPECIAL, (WPARAM) DCs, SI->UpdateCounter);
	}
	deallocate(SI);
}

#define SWI_IMAGE_VIEW_TOOLBAR_BUTTONS 5

static SWI_ID ImageParameters = 0;
static SWI_ID ImageSaveFilter = 0;

void swiIVInitialize() {
	ImageParameters = 0;
	ImageSaveFilter = swiFLNew();
	swiFLAddStrings(ImageSaveFilter, "Bitmaps", "*.pbm");
	swiFLAddStrings(ImageSaveFilter, "Grayscale images", "*.pgm");
	swiFLAddStrings(ImageSaveFilter, "Color images", "*.ppm");
	swiFLAddStrings(ImageSaveFilter, "All files", "*");
}

void swiIVFunctions(SWI_PVOID_VOID CreateFunc, SWI_VOID_PVOID DestroyFunc,
	SWI_PSWIDRAWINGCOMMANDS_PADDINFO AddFunc, SWI_VOID_PADDINFO RemoveFunc,
	SWI_PSWIDRAWINGCOMMANDS_PVOID_PCRITICAL_SECTION SpecialFunc)
{
	if (CreateFunc && DestroyFunc && AddFunc && RemoveFunc) {
		IVF.Create = CreateFunc;
		IVF.Destroy = DestroyFunc;
		IVF.Add = AddFunc;
		IVF.Remove = RemoveFunc;
		IVF.Special = SpecialFunc;
	}
}

void swiIVSetParameters(SWI_ID ParameterSetID, char* SpecialDescription) {
	IVF.Params = ParameterSetID;
	if (SpecialDescription) {
		IVF.Description = allocate(strlen(SpecialDescription) + 1);
		strcpy(IVF.Description, SpecialDescription);
	} else IVF.Description = 0;
}

SWI_ID swiIVParameters() {
	return IVF.Params;
}

#define ORIGNAMELEN 32

typedef struct {
	void* ViewData;
	HWND Toolbar;
	HWND ListView;
	HDC MemoryDC;
	int NeedsRedraw;
	int VertBar;
	int HorzBar;
	int Width, Height;
	int HSPos, VSPos;
	int HSMax, VSMax;
	int HPage, VPage;
	int DrawSpecial;
	int DrawingCommandCount;
	swiDrawingCommands** DCs;
	swiDrawingCommands* Special;
	HGDIOBJ BgBrush;
	HGDIOBJ OrigBitmap;
	HGDIOBJ OrigPen;
	HGDIOBJ EdgePen;
/*	int IDCount;
	SWI_ID* IDs;
	SWI_ID* Params;*/
	CRITICAL_SECTION CS;
	int UpdateCounter, LastUpdated;
	SWI_ID Params;
	DrawArea Visible, Adjusted, Full, ShiftScale;
	char OriginalName[ORIGNAMELEN + 1];
} IVExtra;


static void FillIVExtra(HWND Handle, IVExtra* E) {
	RECT Area;
	LOGBRUSH B;
	E->VertBar = E->HorzBar = 0;
	GetClientRect(Handle, &Area);
	E->HSMax = E->Width = Area.right;
	E->VSMax = E->Height = Area.bottom;
	E->HSPos = E->VSPos = 1;
	E->DrawSpecial = 0;
	E->DrawingCommandCount = 0;
	E->DCs = 0;
	E->Special = 0;
	B.lbColor = 0x00ffffff;
	B.lbHatch = 0;
	B.lbStyle = BS_SOLID;
	E->BgBrush = CreateBrushIndirect(&B);
	InitializeCriticalSection(&(E->CS));
	E->UpdateCounter = 1;
	E->LastUpdated = 0;
	E->Visible.Left = E->Visible.Bottom = 0.0f;
	E->Visible.Width = E->Visible.Height = 1.0f;
	E->Adjusted.Left = E->Adjusted.Bottom = 0.0f;
	E->Adjusted.Width = E->Adjusted.Height = 1.0f;
	E->Full.Left = E->Full.Bottom = 0.0f;
	E->Full.Width = E->Full.Height = 1.0f;
	E->ShiftScale.Left = E->ShiftScale.Bottom = 0.0f;
	E->ShiftScale.Width = E->ShiftScale.Height = 1.0f;
	E->OriginalName[0] = '\0';
}

static void SaveImage(char* Filename, int Format, HDC DC,
	int Width, int Height)
{
	int x, y;
	COLORREF Color;
	IMAGE image;
	FILE* f;
	BYTE PixelVal;
	HBITMAP bmp = GetCurrentObject(DC, OBJ_BITMAP);
	SetImageSize(&image, Width, Height);
	switch (Format) {
	case 1:
		image.FileType = PBM;
		image.MaxValue = 1;
		break;
	case 2:
		image.FileType = PGM;
		image.MaxValue = 255;
		break;
	case 3:
		image.FileType = PPM;
		image.MaxValue = 255;
		break;
	}
	f = fopen(Filename, "wb");
	InitializeImage(f, &image, WHOLEIMAGE, OUTPUT, 0);
	for (y = Height; y > 0; y--) for (x = 1; x <= Width; x++) {
		Color = GetPixel(DC, x - 1, y - 1);
		switch(Format) {
		case 1:
			PutImagePixel(&image, x, y, 0, 
				(BYTE)((Color != 0x00ffffff) ? 0 : 1));
			break;
		case 2:
			PixelVal = min(255, (BYTE) (0.3333 * (Color & 0xff) 
							+ 0.3333 * ((Color >> 8) & 0xff)
							+ 0.3333 * ((Color >> 16) & 0xff)));
			PutImagePixel(&image, x, y, 0, PixelVal);
			break;
		case 3:
			PutImagePixel(&image, x, y, 0, (BYTE) (Color & 0xff));
			PutImagePixel(&image, x, y, 1, (BYTE) ((Color >> 8) & 0xff));
			PutImagePixel(&image, x, y, 2, (BYTE) ((Color >> 16) & 0xff));
		}
	}
	WriteWholeImage(&image);
	FreeImage(&image);
	fclose(f);
	swiInformationMessageBox("Image saved.");
}

static void SetOptionsToDrawingCommands(IVExtra* E, swiDrawingCommands* DC) {
	if (swiPSIsMember(DC->Params, 9)) {
		swiDRArrowAttributes(DC, swiPSValue(DC->Params, 9),
			1000 * swiPSValue(DC->Params, 10), 0, swiPSValue(DC->Params, 11),
			swiPSValue(DC->Params, 12));
	}
	if (swiPSIsMember(DC->Params, 3)) { /* PA: convex hull dotted line */
		swiDRLineAttributes(DC, 0x00404040, swiPSValue(DC->Params, 3));
	} else if (swiPSIsMember(DC->Params, 7)) { /* TS */
		swiDRCircleAttributes(DC, 0, swiPSValue(DC->Params, 7),
			1000 * swiPSValue(DC->Params, 8), swiPSValue(E->Params, 2));
	} else if (swiPSIsMember(DC->Params, 4)) { /* CB: color */
		swiDRCircleAttributes(DC, ToColor(swiPSValue(DC->Params, 4)), 
			swiPSValue(DC->Params, 5), 1000 * swiPSValue(DC->Params, 6), 
			swiPSValue(E->Params, 2));
	}
}

typedef enum {
	SWI_TB_TOOLBAR = 1000,
	SWI_TB_OPTIONS,
	SWI_TB_INSERT,
	SWI_TB_COPY,
	SWI_TB_SAVE,
	SWI_TB_VORONOI
} swiToolbarButtonIDs;

static TBBUTTON ButtonData[SWI_IMAGE_VIEW_TOOLBAR_BUTTONS] = {
	{ 0, SWI_TB_OPTIONS, TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0 },
	{ 1, SWI_TB_INSERT, TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0 },
	{ 2, SWI_TB_COPY, TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0 },
	{ 3, SWI_TB_SAVE, TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0 },
	{ 4, SWI_TB_VORONOI, TBSTATE_ENABLED, TBSTYLE_CHECK, 0, 0 }
};

static LRESULT HandleWMCommand(HWND Handle, UINT Message, 
	WPARAM wParam, LPARAM lParam)
{
	int k, Count, Format, ToolbarHeight;
	SWI_ID* IDs;
	IVExtra* E;
	RECT Area;
	char* Name, *Ext;
	HWND NewView;
	switch (LOWORD(wParam)) {
	case SWI_TB_OPTIONS:
		E = (IVExtra*) GetWindowLong(Handle, 0);
		if (E->Params) {
			swiOptionsDialog(E->Params, Handle);
			if (E->Special) swiDRLineAttributes(
				E->Special, 0x00808080, swiPSValue(E->Params, 1));
			for (k = 0; k < E->DrawingCommandCount; k++)
				SetOptionsToDrawingCommands(E, E->DCs[k]);
			E->NeedsRedraw = 1;
			GetClientRect(Handle, &Area);
			InvalidateRect(Handle, &Area, 0);
		} else swiInformationMessageBox("No options for the image view.");
		return 0;
	case SWI_TB_INSERT:
		Count = swiSDSelectedCount(SWI_SELECTED_ACTIVE);
		if (Count) {
			IDs = allocate(Count * sizeof(SWI_ID));
			swiSDSets(SWI_SELECTED_ACTIVE, 0, Count, IDs);
			for (k = 0; k < Count; k++) 
				PostMessage(Handle, SWI_WM_ADD_TO_VIEW, (WPARAM)IDs[k], 0);
			deallocate(IDs);
		}
		return 0;
	case SWI_TB_COPY:
		E = (IVExtra*) GetWindowLong(Handle, 0);
		GetWindowRect(E->Toolbar, &Area);
		ToolbarHeight = Area.bottom - Area.top;
		GetClientRect(Handle, &Area);
		Area.bottom -= ToolbarHeight;
		Area.right -= ListViewWidth + 2;
		NewView = swiIVCreate(Area.right, Area.bottom);
		E = (IVExtra*) GetWindowLong(Handle, 0);
		for (k = 0; k < E->DrawingCommandCount; k++) SendMessage(
			NewView, SWI_WM_ADD_TO_VIEW, E->DCs[k]->ID, E->DCs[k]->Params);
		return 0;
	case SWI_TB_SAVE:
		Name = swiSaveFilename(0, 0, ImageSaveFilter, &k);
		if (!Name) return 0;
		Ext = strrchr(Name, '.');
		if (Ext) strlwr(Ext);
		if (Ext && !strcmp(Ext, ".pbm")) Format = 1;
		else if (Ext && !strcmp(Ext, ".pgm")) Format = 2;
		else if (Ext && !strcmp(Ext, ".ppm")) Format = 3;
		else if (1 <= k && k <= 3) Format = k;
		else {
			swiErrorMessageBox(
				"Cannot determine file format, "
				"select proper filter or give the extension.\n");
			deallocate(Name);
			return 0;
		}
		E = (IVExtra*) GetWindowLong(Handle, 0);
		//k = min(E->HSMax, E->VSMax);
		SaveImage(Name, Format, E->MemoryDC, E->HSMax, E->VSMax);
		deallocate(Name);
		return 0;
	case SWI_TB_VORONOI:
		E = (IVExtra*) GetWindowLong(Handle, 0);
		E->DrawSpecial = !E->DrawSpecial;
		E->NeedsRedraw = 1;
		GetClientRect(Handle, &Area);
		InvalidateRect(Handle, &Area, 0);
		return 0;
	}
	return DefWindowProc(Handle, Message, wParam, lParam);
}

static int CALLBACK ArrangeFunc(LPARAM P1, LPARAM P2, LPARAM NotUsed) {
	SWI_ID DS1, DS2, C1, C2;
	char* N1, *N2;
	int Cmp;
	DS1 = (SWI_ID) P1;
	DS2 = (SWI_ID) P2;
	C1 = swiDSGetClassID(DS1);
	C2 = swiDSGetClassID(DS2);
	if (C1 < C2) return -1;
	if (C2 < C1) return 1;
	N1 = swiFilename(swiDSName(DS1));
	N2 = swiFilename(swiDSName(DS2));
	Cmp = _stricmp(N1, N2);
	if (Cmp) return Cmp;
	N1 = swiDSName(DS1);
	N2 = swiDSName(DS2);
	Cmp = _stricmp(N1, N2);
	if (Cmp) return Cmp;
	return 0;
}

static LRESULT HandleWMNotify(HWND Handle, UINT Message, 
	WPARAM wParam, LPARAM lParam)
{
	int k;
	LVITEM Item;
	NMLVKEYDOWN* Key;
	LVHITTESTINFO HitInfo;
	RECT Area;
	DWORD Position;
	TOOLTIPTEXT* Tip;
	IVExtra* E;
	NMHDR* Header = (NMHDR*) lParam;
	switch (Header->code) {
	case TTN_NEEDTEXT:
		Tip = (TOOLTIPTEXT*) lParam;
		Tip->hinst = 0;
		Tip->lpszText = Tip->szText;
		if (Tip->hdr.idFrom == SWI_TB_OPTIONS) strcpy(Tip->szText, 
			"Set options that affect the display of data.");
		else if (Tip->hdr.idFrom == SWI_TB_INSERT) strcpy(Tip->szText, 
			"Insert selected data into the image window.");
		else if (Tip->hdr.idFrom == SWI_TB_COPY) strcpy(Tip->szText, 
			"Make a copy of the image window.");
		else if (Tip->hdr.idFrom == SWI_TB_SAVE)
			strcpy(Tip->szText, "Save the image to a file.");
		else if (Tip->hdr.idFrom == SWI_TB_VORONOI) {
			strncpy(Tip->szText, IVF.Description, 79);
			Tip->szText[79] = 0;
		}
		return 0;
	case LVN_INSERTITEM:
		PostMessage(Handle, SWI_WM_IV_SORT, 0, 0);
		return 0;
	case LVN_DELETEALLITEMS:
		return FALSE;
	case LVN_DELETEITEM:
		PostMessage(Handle, SWI_WM_IV_SORT, 0, 0);
		return 0;
	case LVN_KEYDOWN:
		Key = (NMLVKEYDOWN*) lParam;
		if (Key->wVKey == VK_DELETE) {
			E = (IVExtra*) GetWindowLong(Handle, 0);
			Item.iItem = -1;
			Item.iSubItem = 0;
			Item.mask = LVIF_PARAM;
			for (k = 0; k < (int)ListView_GetSelectedCount(E->ListView); k++) {
				Item.iItem = ListView_GetNextItem(
					E->ListView, Item.iItem, LVNI_SELECTED);
				ListView_GetItem(E->ListView, &Item);
				PostMessage(Handle, SWI_WM_REMOVE_FROM_VIEW, Item.lParam, 0);
			}
		}
		return 0;
	case NM_RCLICK:
		Position = GetMessagePos();
		GetWindowRect(Header->hwndFrom, &Area);
		HitInfo.pt.x = LOWORD(Position) - Area.left;
		HitInfo.pt.y = HIWORD(Position) - Area.top;
		HitInfo.flags = 
			LVHT_ONITEMICON | LVHT_ONITEMLABEL | LVHT_ONITEMSTATEICON;
		if (-1 == ListView_HitTest(Header->hwndFrom, &HitInfo)) return 0;
		Item.iItem = HitInfo.iItem;
		Item.iSubItem = 0;
		Item.mask = LVIF_PARAM;
		ListView_GetItem(Header->hwndFrom, &Item);
		E = (IVExtra*) GetWindowLong(Handle, 0);
		for (k = 0; k < E->DrawingCommandCount; k++) 
			if (E->DCs[k]->ID == (SWI_ID)(Item.lParam))
		{
			if (!(E->DCs[k])->Params) return 0;
			swiOptionsDialog((E->DCs[k])->Params, Handle);
			SetOptionsToDrawingCommands(E, E->DCs[k]);
			E->NeedsRedraw = 1;
			GetClientRect(Handle, &Area);
			InvalidateRect(Handle, &Area, 0);
			return 0;
		}
		return 0;
	}
	return DefWindowProc(Handle, Message, wParam, lParam);
}

LRESULT CALLBACK ImageViewWindowProcedure(HWND Handle, UINT Message, 
	WPARAM wParam, LPARAM lParam)
{
	HDC DC;
	RECT Area, WindowArea;
	LPRECT A;
	int k, n, ToolbarHeight, NCWidth, NCHeight, Width, Height, Thread;
	AddInfo AI;
	MSG Msg;
	SpecialInfo* SI;
	HGDIOBJ bmp, bmpOld;
	LVFINDINFO Info;
	LVITEM Item;
	swiDrawingCommands* DCs;
	int MustBeRedrawn;
	char* NewTitle;
	IVExtra* E = (IVExtra*) GetWindowLong(Handle, 0);
	switch (Message) {
	case WM_CREATE:
		DC = GetDC(Handle);
		E = allocate(sizeof(IVExtra));
		SetWindowLong(Handle, 0, (LONG)E);
		E->ViewData = (*(IVF.Create))();
		E->Toolbar = CreateToolbarEx(Handle, 
			WS_CHILD | CCS_TOP | TBSTYLE_TOOLTIPS,
			SWI_TB_TOOLBAR, SWI_IMAGE_VIEW_TOOLBAR_BUTTONS, 
			GetModuleHandle(0), SWI_IMAGE_VIEW_TOOLBAR,
			ButtonData, SWI_IMAGE_VIEW_TOOLBAR_BUTTONS, 20, 20, 20, 20, 
			sizeof(TBBUTTON));
		E->ListView = CreateWindow(WC_LISTVIEW, "", WS_VISIBLE | WS_CHILD 
			| WS_CLIPSIBLINGS | LVS_ICON | LVS_SHAREIMAGELISTS, 
			0, 0, ListViewWidth, 256, Handle, 0, GetModuleHandle(0), 0);
		ListView_SetImageList(
			E->ListView, swiLargeImageList(), LVSIL_NORMAL);
		ListView_SetImageList(
			E->ListView, swiSmallImageList(), LVSIL_SMALL);
		E->MemoryDC = CreateCompatibleDC(DC);
		E->OrigBitmap = GetCurrentObject(E->MemoryDC, OBJ_BITMAP);
		E->OrigPen = GetCurrentObject(DC, OBJ_PEN);
		E->EdgePen = 
			CreatePen(PS_SOLID, 1, GetSysColor(COLOR_WINDOWFRAME));
		SelectObject(DC, E->EdgePen);
		if (IVF.Params) E->Params = swiPSCopy(IVF.Params);
		else E->Params = 0;
		FillIVExtra(Handle, E);
		ShowWindow(E->Toolbar, SW_SHOW);
		SendMessage(swiMainWindowHandle(), 
			SWI_WM_ADD_TO_WINDOW_MENU, 0, (LPARAM) Handle);
		GetWindowText(Handle, E->OriginalName, ORIGNAMELEN);
		return 0;
	case WM_ACTIVATE:
		swiDDEnd();
		return 0;
	case WM_COMMAND:
		return HandleWMCommand(Handle, Message, wParam, lParam);
	case WM_NOTIFY:
		return HandleWMNotify(Handle, Message, wParam, lParam);
	case WM_MOUSEMOVE:
		if (swiDDDragging() && (wParam & MK_LBUTTON != MK_LBUTTON))
			swiDDEnd();
		break;
	case WM_KEYDOWN:
		MustBeRedrawn = 0;
		//if (lParam & (1 << 30)) return 0; /* Already down? */
		if (wParam == VK_ADD) {
			ScaleVisible(&(E->Visible), ZOOM_IN_SCALE);
			MustBeRedrawn = 1;
		} else if (wParam == VK_SUBTRACT) {
			ScaleVisible(&(E->Visible), ZOOM_OUT_SCALE);
			MustBeRedrawn = 1;
		} else if (wParam == VK_LEFT) {
			MoveVisibleRelative(&(E->Visible), -VIEW_MOVE_AMOUNT, 0.0f);
			MustBeRedrawn = 1;
		} else if (wParam == VK_RIGHT) {
			MoveVisibleRelative(&(E->Visible), VIEW_MOVE_AMOUNT, 0.0f);
			MustBeRedrawn = 1;
		} else if (wParam == VK_UP) {
			MoveVisibleRelative(&(E->Visible), 0.0f, VIEW_MOVE_AMOUNT);
			MustBeRedrawn = 1;
		} else if (wParam == VK_DOWN) {
			MoveVisibleRelative(&(E->Visible), 0.0f, -VIEW_MOVE_AMOUNT);
			MustBeRedrawn = 1;
		}
		if (MustBeRedrawn) {
			GetWindowRect(E->Toolbar, &Area);
			ToolbarHeight = Area.bottom - Area.top;
			GetClientRect(Handle, &Area);
			Area.top += ToolbarHeight;
			Area.right -= ListViewWidth;
			InvalidateRect(Handle, &Area, 0);
			E->NeedsRedraw = 1;
			return 0;
		}
		break;
	case WM_PAINT:
		GetWindowRect(E->Toolbar, &Area);
		ToolbarHeight = Area.bottom - Area.top;
		if (GetUpdateRect(Handle, &Area, 0)) {
			if (E->NeedsRedraw) {
				Area.top = Area.left = 0;
				Area.right = E->HSMax;
				Area.bottom = E->VSMax;
				FillRect(E->MemoryDC, &Area, E->BgBrush);
				E->NeedsRedraw = 0;
				FindLimits(E->Special, E->DrawingCommandCount, E->DCs,
					&(E->Full));
				n = E->VSMax;//min(E->HSMax, E->VSMax);
				AdjustVisible(&(E->Visible), &(E->Adjusted),
					E->HSMax, E->VSMax);
				ComputeScale(&(E->Adjusted), &(E->Full), &(E->ShiftScale),
					E->HSMax, E->VSMax);
				if (E->DrawSpecial && E->Special) 
					swiDRDraw(E->Special, E->MemoryDC, &(E->ShiftScale), n);
				for (k = 0; k < E->DrawingCommandCount; k++)
					swiDRDraw(E->DCs[k], E->MemoryDC, &(E->ShiftScale), n);
				/* New window title. */
				NewTitle = allocate(strlen(E->OriginalName) + 60);
				sprintf(NewTitle, "%s (%i, %i) - (%i, %i)",	E->OriginalName,
					(int)(E->Full.Left + E->Full.Width * E->Adjusted.Left),
					(int)(E->Full.Bottom + E->Full.Height * E->Adjusted.Bottom),
					(int)(E->Full.Left + E->Full.Width 
						* (E->Adjusted.Left + E->Adjusted.Width)),
					(int)(E->Full.Bottom + E->Full.Height 
						* (E->Adjusted.Bottom + E->Adjusted.Height)));
				SetWindowText(Handle, NewTitle);
				deallocate(NewTitle);
			}
			GetClientRect(Handle, &Area);
			Area.top += ToolbarHeight;
			Area.right -= ListViewWidth;
			DC = GetDC(Handle);
			ShowCursor(FALSE);
			BitBlt(DC, Area.left, Area.top, 
				Area.right - Area.left, Area.bottom - Area.top,
				E->MemoryDC, E->HSPos - 1, E->VSPos - 1, SRCCOPY);
			MoveToEx(DC, Area.right - 2, Area.top, 0);
			LineTo(DC, Area.right - 2, Area.bottom);
			ShowCursor(TRUE);
			GetUpdateRect(Handle, &Area, 0);
			ValidateRect(Handle, &Area);
		}
		return 0;
	case WM_SIZE:
		if ((wParam != SIZE_MAXHIDE) && (wParam != SIZE_MINIMIZED)) {
			SendMessage(E->Toolbar, WM_SIZE, wParam, lParam);
			E->HSMax = E->Width = LOWORD(lParam);
			E->HSMax -= ListViewWidth + 2;
			E->VSMax = E->Height = HIWORD(lParam);
			GetWindowRect(E->Toolbar, &Area);
			ToolbarHeight = Area.bottom - Area.top;
			E->VSMax -= ToolbarHeight;
			MoveWindow(E->ListView, E->Width - ListViewWidth, ToolbarHeight,
				ListViewWidth, E->Height - ToolbarHeight, 1);
			//if (k != min(E->HSMax, E->VSMax)) {
			E->NeedsRedraw = 1;
			bmpOld = GetCurrentObject(E->MemoryDC, OBJ_BITMAP);
			//	k = min(E->HSMax, E->VSMax);
			DC = GetDC(Handle);
				//bmp = CreateCompatibleBitmap(DC, k, k);
			bmp = CreateCompatibleBitmap(DC, E->HSMax, E->VSMax);
			SelectObject(E->MemoryDC, bmp);
			if (bmpOld != E->OrigBitmap) DeleteObject(bmpOld);
			//}
			GetClientRect(Handle, &Area);
			Area.top += ToolbarHeight;
			Area.right -= ListViewWidth;
			InvalidateRect(Handle, &Area, 0);
		}
		return 0;
	case WM_SIZING:
		A = (LPRECT) lParam;
		Width = A->right - A->left;
		Height = A->bottom - A->top;
		if (Width < 110 + ListViewWidth + 2) {
			Width = (110 + ListViewWidth + 2) - Width;
			switch (wParam) {
			case WMSZ_BOTTOMLEFT: A->left -= Width; break;
			case WMSZ_BOTTOMRIGHT: A->right += Width; break;
			case WMSZ_LEFT: A->left -= Width; break;
			case WMSZ_RIGHT: A->right += Width; break;
			case WMSZ_TOPLEFT: A->left -= Width; break;
			case WMSZ_TOPRIGHT: A->right += Width; break;
			}
		}
		if (Height < 110) {
			Height = 110 - Height;
			switch (wParam) {
			case WMSZ_BOTTOM: A->bottom += Height; break;
			case WMSZ_BOTTOMLEFT: A->bottom += Height; break;
			case WMSZ_BOTTOMRIGHT: A->bottom += Height; break;
			case WMSZ_TOP: A->top -= Height; break;
			case WMSZ_TOPLEFT: A->top -= Height; break;
			case WMSZ_TOPRIGHT: A->top -= Height; break;
			}
		}
		break;
	case WM_CLOSE:
		DestroyWindow(Handle);
		return 0;
	case WM_DESTROY:
		(*IVF.Destroy)(E->ViewData);
		DeleteObject(E->BgBrush);
		SelectObject(GetDC(Handle), E->OrigPen);
		DeleteObject(E->EdgePen);
		bmp = GetCurrentObject(E->MemoryDC, OBJ_BITMAP);
		if (bmp != E->OrigBitmap) {
			SelectObject(E->MemoryDC, E->OrigBitmap);
			DeleteObject(bmp);
		}
		DeleteDC(E->MemoryDC);
		while (E->DrawingCommandCount)
			SendMessage(Handle, SWI_WM_REMOVE_FROM_VIEW, E->DCs[0]->ID, 0);
		deallocate(E->DCs);
		DeleteCriticalSection(&(E->CS));
		if (E->Params) swiPSDelete(E->Params);
		deallocate(E);
		SendMessage(swiMainWindowHandle(), 
			SWI_WM_REMOVE_FROM_WINDOW_MENU, 0, (LPARAM) Handle);
		return 0;
	case SWI_WM_ADD_TO_VIEW:
		AI.ViewData = E->ViewData;
		AI.ID = (SWI_ID) wParam;
		AI.Class = swiDSGetClassID(AI.ID);
		AI.External = swiDSExternal(AI.ID);
		AI.Handle = Handle;
		AI.ErrorString = 0;
		AI.ParamIndex = swiFPSGetFirst() - 1;
		AI.ParamValue = 0;
		if (lParam) AI.Params = (SWI_ID) lParam;
		else AI.Params = swiDCViewParameters(swiDSGetClassID(AI.ID));
		AI.CS = &(E->CS);
		DCs = (*IVF.Add)(&AI);
		if (DCs) {
			EnterCriticalSection(AI.CS);
			E->DCs = reallocate(E->DCs, 
				(E->DrawingCommandCount + 1) * sizeof(swiDrawingCommands*));
			k = E->DrawingCommandCount++;
			E->DCs[k] = DCs;
			Item.iItem = ListView_GetItemCount(E->ListView);
			Item.iSubItem = 0;
			Item.mask = LVIF_TEXT | LVIF_IMAGE | LVIF_PARAM | LVIF_STATE;
			Item.state = 0;
			Item.stateMask = (UINT)-1;
			Item.pszText = swiFilename(swiDSName(AI.ID));
			Item.cchTextMax = 0;
			Item.iImage = swiDSIcon(AI.ID);
			Item.lParam = wParam;
			ListView_InsertItem(E->ListView, &Item);
			E->NeedsRedraw = 1;
			if (swiFPSGetFirst() <= AI.ParamIndex 
				&& AI.ParamIndex <= swiFPSGetLast()
				&& swiPSIsMember(AI.Params, AI.ParamIndex))
			{
				swiPSSetValue(AI.Params, AI.ParamIndex, AI.ParamValue);
			}
			LeaveCriticalSection(AI.CS);
			GetClientRect(Handle, &Area);
			InvalidateRect(Handle, &Area, 0);
			SendMessage(Handle, SWI_WM_IV_COPY_PARAMETERS, 
				(WPARAM) AI.ID, (LPARAM) AI.Params);
			return 1;
		} else {
			swiDSUnuse(AI.ID);
			if (AI.ErrorString) {
				swiErrorMessageBox(AI.ErrorString);
				deallocate(AI.ErrorString);
			}
		}
		return 0;
	case SWI_WM_REMOVE_FROM_VIEW:
		AI.ViewData = E->ViewData;
		AI.ID = (SWI_ID) wParam;
		AI.Class = swiDSGetClassID(AI.ID);
		AI.External = swiDSExternal(AI.ID);
		AI.Handle = Handle;
		AI.CS = &(E->CS);
		AI.ErrorString = 0;
		Info.flags = LVFI_PARAM;
		Info.lParam = wParam;
		k = ListView_FindItem(E->ListView, -1, &Info);
		if (k != -1) {
			if (IVF.Remove) (*IVF.Remove)(&AI);
			ListView_DeleteItem(E->ListView, k);
			swiDSUnuse(AI.ID);
			EnterCriticalSection(AI.CS);
			for (k = 0; k < E->DrawingCommandCount; k++) {
				if (E->DCs[k]->ID == AI.ID) {
					swiDRDelete(E->DCs[k]);
					for (n = k; n < E->DrawingCommandCount - 1; n++)
						E->DCs[n] = E->DCs[n + 1];
					E->DCs[E->DrawingCommandCount - 1] = 0;
					E->DrawingCommandCount--;
					E->NeedsRedraw = 1;
					LeaveCriticalSection(AI.CS);
					GetClientRect(Handle, &Area);
					InvalidateRect(Handle, &Area, 0);
					return 0;
				}
			}
		}
		LeaveCriticalSection(AI.CS);
		return 0;
	case SWI_WM_SET_IMAGE_SIZE:
		GetWindowRect(E->Toolbar, &Area);
		ToolbarHeight = Area.bottom - Area.top;
		GetClientRect(Handle, &Area);
		GetWindowRect(Handle, &WindowArea);
		NCWidth = WindowArea.right - WindowArea.left - Area.right;
		NCHeight = WindowArea.bottom - WindowArea.top - Area.bottom;
		MoveWindow(Handle, WindowArea.left, WindowArea.top, 
			NCWidth + wParam + ListViewWidth + 2, 
			NCHeight + lParam + ToolbarHeight + 4, 0);
		/* NOTE: +4 was determined from an image on screen, 
		why the error is that much I have no idea */
		return 0;
	case SWI_WM_IV_COPY_PARAMETERS:
		for (k = 0; k < E->DrawingCommandCount; k++) if (E->DCs[k]->ID == wParam) {
			if (E->DCs[k]->Params) swiPSDelete(E->DCs[k]->Params);
			if (lParam) {
				E->DCs[k]->Params = swiPSCopy(lParam);
				SetOptionsToDrawingCommands(E, E->DCs[k]);
			} else E->DCs[k]->Params = 0;
			return 0;
		}
		return 0;
	case SWI_WM_IV_UPDATE_SPECIAL:
		if (PeekMessage(&Msg, Handle, SWI_WM_IV_UPDATE_SPECIAL, 
			SWI_WM_IV_UPDATE_SPECIAL, PM_NOREMOVE)) return 0;
		/* messages are consecutive enumerants in swi_Messages.h */
		if (PeekMessage(&Msg, Handle, 
			SWI_WM_ADD_TO_VIEW, SWI_WM_REMOVE_FROM_VIEW, PM_NOREMOVE))
		{
			PostMessage(Handle, Message, wParam, lParam);
			return 0;
		}
		SI = allocate(sizeof(SpecialInfo));
		SI->Handle = Handle;
		SI->ViewData = E->ViewData;
		SI->UpdateCounter = E->UpdateCounter++;
		SI->CS = &(E->CS);
		Thread = _beginthread(&SpecialWrapper, 4096, SI);
		if (Thread == -1) {
			deallocate(SI);
			PostMessage(Handle, Message, wParam, lParam);
			Sleep(100);
		}
		return 0;
	case SWI_WM_IV_NEW_SPECIAL:
		if (lParam > E->LastUpdated) {
			if (E->Special) swiDRDelete(E->Special);
			E->Special = (swiDrawingCommands*) wParam;
			if (E->Params && E->Special) swiDRLineAttributes(
				E->Special, 0x00808080, swiPSValue(E->Params, 1));
			E->LastUpdated = lParam;
			E->NeedsRedraw = 1;
			GetClientRect(Handle, &Area);
			InvalidateRect(Handle, &Area, 0);
		} else if (wParam) swiDRDelete((swiDrawingCommands*) wParam);
		return 0;
	case SWI_WM_IV_SORT:
		ListView_SortItems(E->ListView, &ArrangeFunc, 0);
		return 0;
	case SWI_WM_RENAMED:
		Info.flags = LVFI_PARAM;
		Info.lParam = lParam;
		k = ListView_FindItem(E->ListView, -1, &Info);
		if (k == -1) return 0;
		ListView_SetItemText(E->ListView, 
			k, 0, swiFilename(swiDSName((SWI_ID) lParam)));
		return 0;
	}
	return DefWindowProc(Handle, Message, wParam, lParam);
}



static int IVRegistered = 0;

static void RegisterImageViewClass() {
	WNDCLASS W;
	W.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	W.lpfnWndProc = &ImageViewWindowProcedure;
	W.cbClsExtra = 0;
	W.cbWndExtra = sizeof(IVExtra*);
	W.hInstance = GetModuleHandle(0);
	W.hIcon = swiAppIcon();
	W.hCursor = 0; /*LoadCursor(NULL, IDC_ARROW);*/
	W.hbrBackground = (HBRUSH) (COLOR_WINDOW + 1);
	W.lpszMenuName = 0;
	W.lpszClassName = "swiImageViewWindow";
	RegisterClass(&W);
	IVRegistered = 1;
}

static int WindowNumber = 0;

HWND swiIVCreate(int Width, int Height) {
	HWND Window;
	char ViewName[32];
	if (IVF.Create) {
		if (!IVRegistered) RegisterImageViewClass();
		WindowNumber++;
		sprintf(ViewName, "Image View %i", WindowNumber);
		Window = CreateWindow("swiImageViewWindow", ViewName,
			WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN,
			CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
			swiMainWindowHandle(), 0, GetModuleHandle(0), 0);
		if (Width && Height) SendMessage(Window, 
			SWI_WM_SET_IMAGE_SIZE, (WPARAM) Width, (LPARAM) Height);
		else SendMessage(Window, SWI_WM_SET_IMAGE_SIZE, 256, 256);
		ShowWindow(Window, SW_SHOW);
		UpdateWindow(Window);
	} else Window = 0;
	return Window;
}

void swiIVDestroy(HWND ImageViewWindow) {
	DestroyWindow(ImageViewWindow);
}

int swiIVAdd(HWND ImageViewWindow, SWI_ID DataSetID) {
	return SendMessage(ImageViewWindow, 
		SWI_WM_ADD_TO_VIEW, (WPARAM) DataSetID, 0);
}

int swiIVDefined() {
	return IVF.Create != 0;
}

void swiIVRemove(HWND ImageViewWindow, SWI_ID DataSetID) {
	PostMessage(ImageViewWindow, 
		SWI_WM_REMOVE_FROM_VIEW, (WPARAM) DataSetID, 0);

}

void swiIVUpdateSpecial(HWND ImageViewWindow) {
	PostMessage(ImageViewWindow, SWI_WM_IV_UPDATE_SPECIAL, 0, 0);
}
