/********************************************************************

	swi_MainWindow.c

	Ismo Krkkinen

	Last changed 16.11.1999

	Main message loop and main window procedure.

********************************************************************/

#include "swi_Internal.h"
#include "swi_ProcessAndOptions.h"
#include "swi_ProcessQueue.h"
#include "swi_MainMenuSymbols.h"
#include "swi_WindowsDialogs.h"
#include "swi_OutputFlow.h"
#include <windows.h>
#include <commctrl.h>
#include "memctrl.h"
#include "interfc.h"
#include <string.h>
#include "idxqueue.h"
#include "swi_OwnedWindow.h"
#include "swi_WindowsAccelerator.h"
#include "swi_WindowsMenus.h"
#include "swi_Requirement.h"
#include "swi_SelectedData.h"
#include "swi_DragAndDrop.h"
#include "swi_Process.h"
#include "swi_ImageView.h"
#include "resource.h"
#include "swi_DataClass.h"
#include "swi_Filter.h"
#include "swi_Storage.h"

char dir[2048];

#define SWI_WINDOW_MENU_ITEMS 2

static HMENU Menu = 0;
static HWND ActiveListView = 0;
static HPEN OutlinePenDark, OutlinePenLight;
static HWND MainList, InputList, OutputList, StatusBar;
static char* StatusText = 0;
static COLORREF MainWindowBrushColor = 0;
static HBRUSH MainWindowBackgroundBrush = 0;
static POINT DragBeginPos;

static INDEXEDQUEUE IDs;

/* Note: used by several functions */
int SeekDataSets(void* Data) {
	SWI_ID* Found;
	if (((SWI_ID*)Data)[1] == DATA_SET) {
		Found = allocate(sizeof(SWI_ID));
		*Found = *((SWI_ID*)Data);
		EnqueueIntoIndexedQueue(&IDs, Found);
	}
	return 0;
}

static void ValidateMenus() {
	int Count, k, Save, Text;
	SWI_ID* IDs;
	SWI_ID Class;
	Count = swiSDSelectedCount(SWI_SELECTED_ACTIVE);
	if (Count) {
		EnableMenuItem(Menu,
			SWI_MAINMENU_FILE_CLOSE, MF_BYCOMMAND | MF_ENABLED);
		IDs = allocate(Count * sizeof(SWI_ID));
		swiSDSets(SWI_SELECTED_ACTIVE, 0, Count, IDs);
		if (swiIVDefined()) 
			EnableMenuItem(Menu, 
				SWI_MAINMENU_VIEW_IMAGE, MF_BYCOMMAND | MF_ENABLED);
		else EnableMenuItem(Menu, 
			SWI_MAINMENU_VIEW_IMAGE, MF_BYCOMMAND | MF_GRAYED);
		Save = Text = 1;
		for (k = 0; k < Count; k++) {
			Class = swiDSGetClassID(IDs[k]);
			Save = Save && swiDCSaveDefined(Class);
			Text = Text && swiDCTextViewDefined(Class);
		}
		if (Save) {
			EnableMenuItem(Menu,
				SWI_MAINMENU_FILE_SAVE, MF_BYCOMMAND | MF_ENABLED);
			EnableMenuItem(Menu,
				SWI_MAINMENU_FILE_SAVEAS, MF_BYCOMMAND | MF_ENABLED);
		} else {
			EnableMenuItem(Menu,
				SWI_MAINMENU_FILE_SAVE, MF_BYCOMMAND | MF_GRAYED);
			EnableMenuItem(Menu,
				SWI_MAINMENU_FILE_SAVEAS, MF_BYCOMMAND | MF_GRAYED);
		}
		if (Text)
			EnableMenuItem(Menu,
				SWI_MAINMENU_VIEW_TEXT, MF_BYCOMMAND | MF_ENABLED);
		else EnableMenuItem(Menu,
			SWI_MAINMENU_VIEW_TEXT, MF_BYCOMMAND | MF_GRAYED);
		deallocate(IDs);
	} else {
		EnableMenuItem(Menu,
			SWI_MAINMENU_FILE_SAVE, MF_BYCOMMAND | MF_GRAYED);
		EnableMenuItem(Menu,
			SWI_MAINMENU_FILE_SAVEAS, MF_BYCOMMAND | MF_GRAYED);
		EnableMenuItem(Menu,
			SWI_MAINMENU_FILE_CLOSE, MF_BYCOMMAND | MF_GRAYED);
		EnableMenuItem(Menu,
			SWI_MAINMENU_VIEW_TEXT, MF_BYCOMMAND | MF_GRAYED);
		EnableMenuItem(Menu,
			SWI_MAINMENU_VIEW_IMAGE, MF_BYCOMMAND | MF_GRAYED);
	}
	Count = swiSDCount(SWI_SELECTED_ACTIVE);
	if (Count <= 1) {
		EnableMenuItem(Menu,
			SWI_MAINMENU_VIEW_ARRANGENAME, MF_BYCOMMAND | MF_GRAYED);
		EnableMenuItem(Menu,
			SWI_MAINMENU_VIEW_ARRANGETYPE, MF_BYCOMMAND | MF_GRAYED);
	} else {
		EnableMenuItem(Menu,
			SWI_MAINMENU_VIEW_ARRANGENAME, MF_BYCOMMAND | MF_ENABLED);
		EnableMenuItem(Menu,
			SWI_MAINMENU_VIEW_ARRANGETYPE, MF_BYCOMMAND | MF_ENABLED);
	}
	if (Count) {
		EnableMenuItem(Menu,
			SWI_MAINMENU_VIEW_ARRANGETOP, MF_BYCOMMAND | MF_ENABLED);
		EnableMenuItem(Menu,
			SWI_MAINMENU_VIEW_LINEUP, MF_BYCOMMAND | MF_ENABLED);
	} else {
		EnableMenuItem(Menu,
			SWI_MAINMENU_VIEW_ARRANGETOP, MF_BYCOMMAND | MF_GRAYED);
		EnableMenuItem(Menu,
			SWI_MAINMENU_VIEW_LINEUP, MF_BYCOMMAND | MF_GRAYED);
	}
	DrawMenuBar(swiMainWindowHandle());
}

static void Outline(HDC DC, HWND View)
{
	RECT Area, Main;
	int diff;
	GetWindowRect(View, &Area);
	GetWindowRect(swiMainWindowHandle(), &Main);
	diff = GetSystemMetrics(SM_CXSIZEFRAME) + Main.left;
	Area.left -= diff;
	Area.right -= diff;
	diff = Main.top + 
		GetSystemMetrics(SM_CYSIZEFRAME) + GetSystemMetrics(SM_CYMENU)
		+ GetSystemMetrics(SM_CYCAPTION);
	Area.top -= diff;
	Area.bottom -= diff;
	Area.top -= 2;
	Area.left -= 2;
	Area.right += 2;
	Area.bottom += 2;
	SelectObject(DC, OutlinePenLight);
	MoveToEx(DC, Area.left, Area.bottom, 0);
	LineTo(DC, Area.left, Area.top);
	LineTo(DC, Area.right, Area.top);
	SelectObject(DC, OutlinePenDark);
	MoveToEx(DC, Area.left, Area.bottom, 0);
	LineTo(DC, Area.right, Area.bottom);
	LineTo(DC, Area.right, Area.top);
}

static void swiSetStatusBarText(char* Text) {
	SendMessage(StatusBar, 
		SB_SETTEXT, MAKEWPARAM(0, SBT_NOBORDERS), (LPARAM) Text);
}

static int SaveThese(int Count, SWI_ID* IDs) {
	int k, Index;
	char* foo, *bar, *Name;
	SWI_ID Class;
	for (k = 0; k < Count; k++) {
		Class = swiDSGetClassID(IDs[k]);
		if (!swiDCSaveDefined(Class)) {
			foo = AppendString(
				"Save function hasn't been defined for class\n",
				swiDCDescription(Class));
			swiInformationMessageBox(foo);
			deallocate(foo);
		} else {
			Name = swiDSName(IDs[k]);
			if (!swiDSLocked(IDs[k])) {
				if (swiDSHasBeenSaved(IDs[k])) swiDSSave(IDs[k], Name);
				else {
					Index = 1;
					Name = swiSaveFilename(Name, strlen(Name), 
						swiDCFilter(Class), &Index);
					if (Name) swiDSSave(IDs[k], Name);
					else return 0;
				}
			} else {
				if (Name) {
					foo = AppendString(Name,
						"\nhas been locked and cannot be saved.");
				} else {
					bar = AppendString("<Untitled> of class\n",
						swiDCDescription(Class));
					foo = AppendString(bar,
						"\nhas been locked and cannot be saved.");
					deallocate(bar);
				}
				swiWarningMessageBox(foo);
				deallocate(foo);
			}
		}
	}
	return 1;
}

static int AskSaveForThese(int Count, SWI_ID* IDs) {
	int k;
	char* Name, *foo, *bar;
	SWI_ID Class;
	int SavingOK = 1;
	for (k = 0; k < Count; k++) {
		Name = swiDSName(IDs[k]);
		if (!swiDSLocked(IDs[k]) 
			&& (swiDSHasChanged(IDs[k]) || !swiDSHasBeenSaved(IDs[k])))
		{
			Class = swiDSGetClassID(IDs[k]);
			if (Name && swiDSHasBeenSaved(IDs[k])) {
				foo = AppendString(Name, "\nhas not been saved. Save it?");
			} else {
				if (Name) foo = 
					AppendString(Name, "\nhas not been saved. Save it?");
				else {
					bar = AppendString("<Untitled> of class\n",
						swiDCDescription(Class));
					foo = 
						AppendString(bar, "\nhas not been saved. Save it?");
					deallocate(bar);
				}
			}
			if (swiQuestionMessageBox(foo, 1)) {
				if (!Name || !swiDSHasBeenSaved(IDs[k])) {
					if (Name) Name = swiSaveFilename(
						Name, strlen(Name), swiDCFilter(Class), 0);
					else Name = 
						swiSaveFilename(0, 0, swiDCFilter(Class), 0);
					if (Name) swiDSSave(IDs[k], Name);
					else return 0; /* cancel was pressed */ 
				} else swiDSSave(IDs[k], Name);
			}
			deallocate(foo);
		} else if (swiDSHasChanged(IDs[k]) || !swiDSHasBeenSaved(IDs[k])) {
			if (Name) {
				foo = AppendString(Name,
					"\nhas been locked and cannot be saved.");
			} else {
				bar = AppendString("<Untitled> of class\n",
					swiDCDescription(Class));
				foo = AppendString(bar,
					"\nhas been locked and cannot be saved.");
				deallocate(bar);
			}
			swiWarningMessageBox(foo);
			deallocate(foo);
		}
	}
	return 1;
}

static LRESULT HandleWMNotify(HWND Handle, UINT iMsg, 
	WPARAM wParam, LPARAM lParam)
{
	LVITEM Item;
	NMLVDISPINFO* Info;
	NMLVKEYDOWN* Key;
	LVHITTESTINFO HitInfo;
	RECT Area;
	DWORD Position;
	NMHDR* Header = (NMHDR*) lParam;
	switch (Header->code) {
	case LVN_BEGINDRAG:
		if (((Header->hwndFrom == swiSDGetSource(SWI_SELECTED_MAIN)) 
			&& swiSDSelectedCount(SWI_SELECTED_MAIN))
			|| ((Header->hwndFrom == swiSDGetSource(SWI_SELECTED_OUTPUT)) 
			&& swiSDSelectedCount(SWI_SELECTED_OUTPUT))
			|| ((Header->hwndFrom == swiSDGetSource(SWI_SELECTED_INPUT)) 
			&& swiSDSelectedCount(SWI_SELECTED_INPUT)))
		{
			Position = GetMessagePos();
			DragBeginPos.x = LOWORD(Position);
			DragBeginPos.y = HIWORD(Position);
			swiSDSetSource(SWI_SELECTED_DRAG_SOURCE, Header->hwndFrom);
			swiDDBegin();
		}
		return 0;
	case LVN_KEYDOWN:
		Key = (NMLVKEYDOWN*) lParam;
		if (Key->wVKey == VK_DELETE)
			SendMessage(Handle, SWI_WM_CLOSE_SELECTED, 0, 0);
		return 0;
	case LVN_ENDLABELEDIT:
		Info = (NMLVDISPINFO*) lParam;
		if (Info->item.pszText) {
			Item.iItem = Info->item.iItem;
			Item.iSubItem = 0;
			Item.mask = LVIF_PARAM;
			ListView_GetItem(Header->hwndFrom, &Item);
			swiDSSetName(Item.lParam, Info->item.pszText);
			return TRUE;
		}
		return FALSE;
	case NM_SETFOCUS:
		swiSDSetSource(SWI_SELECTED_ACTIVE, Header->hwndFrom);
		ActiveListView = Header->hwndFrom;
		if (ActiveListView == InputList) 
			swiSetStatusBarText("Input area");
		else if (ActiveListView == OutputList) 
			swiSetStatusBarText("Output area");
		else if (ActiveListView == MainList) 
			swiSetStatusBarText("Main area");
		GetClientRect(Handle, &Area);
		InvalidateRect(Handle, &Area, 1);
		UpdateWindow(Handle);
		ValidateMenus();
		return 0;
	case NM_CLICK:
		ValidateMenus();
		return 0;
	case NM_DBLCLK:
		Position = GetMessagePos();
		GetWindowRect(Header->hwndFrom, &Area);
		HitInfo.pt.x = LOWORD(Position) - Area.left;
		HitInfo.pt.y = HIWORD(Position) - Area.top;
		HitInfo.flags = 
			LVHT_ONITEMICON | LVHT_ONITEMLABEL | LVHT_ONITEMSTATEICON;
		ListView_HitTest(Header->hwndFrom, &HitInfo);
		Item.iItem = 
			ListView_GetNextItem(Header->hwndFrom, -1, LVNI_FOCUSED);
		if (HitInfo.iItem != Item.iItem || Item.iItem == -1) return 0;
		Item.iSubItem = 0;
		Item.mask = LVIF_PARAM;
		ListView_GetItem(Header->hwndFrom, &Item);
		Handle = swiIVCreate(256, 256);
		if (Handle) {
			if (!swiIVAdd(Handle, (SWI_ID) Item.lParam)) {
				swiIVDestroy(Handle);
				swiDSCreateTextWindow((SWI_ID) Item.lParam);
			}
		}
		return 0;
	}
	return DefWindowProc(Handle, iMsg, wParam, lParam);
}

typedef enum {
	SWI_ARRANGE_NAME = 1,
	SWI_ARRANGE_TYPE
} swiArrangeTypes;

static int CALLBACK ArrangeFunc(LPARAM P1, LPARAM P2, LPARAM ArrangeType) {
	SWI_ID DS1, DS2, C1, C2;
	char* N1, *N2;
	int Cmp;
	DS1 = (SWI_ID) P1;
	DS2 = (SWI_ID) P2;
	if (ArrangeType == SWI_ARRANGE_TYPE) {
		C1 = swiDSGetClassID(DS1);
		C2 = swiDSGetClassID(DS2);
		if (C1 < C2) return -1;
		if (C2 < C1) return 1;
		ArrangeType = SWI_ARRANGE_NAME;
	}
	if (ArrangeType == SWI_ARRANGE_NAME) {
		N1 = swiFilename(swiDSName(DS1));
		N2 = swiFilename(swiDSName(DS2));
		Cmp = _stricmp(N1, N2);
		if (Cmp) return Cmp;
		N1 = swiDSName(DS1);
		N2 = swiDSName(DS2);
		Cmp = _stricmp(N1, N2);
		if (Cmp) return Cmp;
	}
	return 0;
}

static int OpenFilter = 1;

static LRESULT HandleWMCommand(int Message, int Control, HWND Handle, 
	UINT OrigMessage, WPARAM wParam, LPARAM lParam)
{
	SWI_ID Process, Class;
	HINSTANCE Exe;
	int OK, Count, k;
	SWI_ID* IDs;
	char* foo, *bar;
	MENUITEMINFO M;
	HMENU WindowMenu;
	HWND View;
	WINDOWPLACEMENT WP;
	swiDDEnd();
	if ((Message < SWI_LAST_PROCESS_MENUITEM) 
		&& (Message >= SWI_FIRST_PROCESS_MENUITEM))
	{
		M.cbSize = sizeof(MENUITEMINFO);
		M.fMask = MIIM_DATA;
		M.dwItemData = 0;
		GetMenuItemInfo(Menu, Message, FALSE, &M);
		Class = 
			swiBestFittingProcessClass(M.dwItemData, SWI_SELECTED_INPUT);
		if (!swiPCParameterSet(Class)) {
			if (swiSDSimilarContents(SWI_SELECTED_INPUT, 
				swiPCInputRequirement(Class)))
			{
				OK = swiPCOverwriteOutput(Class);
				Process = swiPRNew(Class, OK);
				swiEnqueueProcess(Process);
				PostMessage(swiMainWindowHandle(),
					SWI_WM_CHECK_NEW_PROCESSES, 0, 0);
			} else {
				foo = swiPCPreconditionDescription(Class);
				swiInformationMessageBox(foo);
				deallocate(foo);
			}
		} else {
			swiCreateProcessAndOptionsDialog();
			swiSelectProcessAndOptions(Class);
		}
		return 0;
	} else if (SWI_MAINMENU_WINDOW_LISTBEGIN <= Message 
		&& Message < 
			SWI_MAINMENU_WINDOW_LISTBEGIN + SWI_MAINMENU_WINDOW_LISTSIZE)
	{
		M.cbSize = sizeof(MENUITEMINFO);
		M.fMask = MIIM_DATA;
		M.dwItemData = 0;
		GetMenuItemInfo(GetSubMenu(GetMenu(Handle), 3), Message, 0, &M);
		WP.length = sizeof(WINDOWPLACEMENT);
		GetWindowPlacement((HWND) M.dwItemData, &WP);
		if (WP.showCmd == SW_SHOWMINIMIZED) 
			ShowWindow((HWND) M.dwItemData, SW_RESTORE);
		SetForegroundWindow((HWND) M.dwItemData);
		return 0;
	} else switch (Message) {
	case SWI_MAINMENU_FILE_OPEN:
		foo = swiOpenFilename(0, 0, GlobalOpenFilter, &Class, &OpenFilter);
		if (foo) swiDSLoad(Class, foo);
		return 0;
	case SWI_MAINMENU_FILE_SAVE:
		Count = swiSDSelectedCount(SWI_SELECTED_ACTIVE);
		if (Count) {
			IDs = allocate(Count * sizeof(SWI_ID));
			swiSDSets(SWI_SELECTED_ACTIVE, 0, Count, IDs);
			SaveThese(Count, IDs);
			deallocate(IDs);
		}
		return 0;
	case SWI_MAINMENU_FILE_SAVEAS:
		Count = swiSDSelectedCount(SWI_SELECTED_ACTIVE);
		if (Count) {
			IDs = allocate(Count * sizeof(SWI_ID));
			swiSDSets(SWI_SELECTED_ACTIVE, 0, Count, IDs);
			for (k = 0; k < Count; k++) {
				if (!swiDSLocked(IDs[k])) {
					bar = swiDSName(IDs[k]);
					Class = swiDSGetClassID(IDs[k]);
					foo = swiSaveFilename(
						bar, strlen(bar), swiDCFilter(Class), 0);
					if (!foo) {
						deallocate(IDs);
						return 0;
					}
					swiDSSave(IDs[k], foo);
					if (foo != bar) deallocate(foo);
				} else {
					foo = AppendString(swiDSName(IDs[k]), 
						"\nhas been locked and cannot be saved.");
					swiInformationMessageBox(foo);
					deallocate(foo);
				}
			}
			deallocate(IDs);
		}
		return 0;
	case SWI_MAINMENU_FILE_CLOSE:
		SendMessage(Handle, SWI_WM_CLOSE_SELECTED, 0, 0);
		return 0;
	case SWI_MAINMENU_FILE_EXIT:
		PostMessage(Handle, WM_CLOSE, 0, 0);
		return 0;
	case SWI_MAINMENU_PROCESS_OUTPUT:
		if (swiOutputFlowDialogHandle()) {
			ShowWindow(swiOutputFlowDialogHandle(), SW_SHOW);
			SetForegroundWindow(swiOutputFlowDialogHandle());
		}
		return 0;
	case SWI_MAINMENU_PROCESS_QUEUE:
		swiCreateProcessQueueDialog();
		return 0;
	case SWI_MAINMENU_VIEW_TEXT:
		Count = swiSDSelectedCount(SWI_SELECTED_ACTIVE);
		if (Count) {
			IDs = allocate(Count * sizeof(SWI_ID));
			swiSDSets(SWI_SELECTED_ACTIVE, 0, Count, IDs);
			for (k = 0; k < Count; k++) swiDSCreateTextWindow(IDs[k]);
			deallocate(IDs);
		}
		return 0;
	case SWI_MAINMENU_VIEW_IMAGE:
		Count = swiSDSelectedCount(SWI_SELECTED_ACTIVE);
		if (Count) {
			IDs = allocate(Count * sizeof(SWI_ID));
			swiSDSets(SWI_SELECTED_ACTIVE, 0, Count, IDs);
			View = swiIVCreate(256, 256);
			if (View) for (k = 0; k < Count; k++) swiIVAdd(View, IDs[k]);
			deallocate(IDs);
		}
		return 0;
	case SWI_MAINMENU_VIEW_ARRANGENAME:
		ListView_SortItems(swiSDGetSource(SWI_SELECTED_ACTIVE), 
			&ArrangeFunc, SWI_ARRANGE_NAME);
		return 0;
	case SWI_MAINMENU_VIEW_ARRANGETYPE:
		ListView_SortItems(swiSDGetSource(SWI_SELECTED_ACTIVE), 
			&ArrangeFunc, SWI_ARRANGE_TYPE);
		return 0;
	case SWI_MAINMENU_VIEW_ARRANGETOP:
		ListView_Arrange(swiSDGetSource(SWI_SELECTED_ACTIVE), LVA_ALIGNTOP);
		return 0;
	case SWI_MAINMENU_VIEW_LINEUP:
		ListView_Arrange(
			swiSDGetSource(SWI_SELECTED_ACTIVE), LVA_SNAPTOGRID);
		return 0;
	case SWI_MAINMENU_WINDOW_CLOSEALL:
		WindowMenu = GetSubMenu(Menu, 3);
		M.cbSize = sizeof(MENUITEMINFO);
		M.fMask = MIIM_DATA;
		while (GetMenuItemCount(WindowMenu) > SWI_WINDOW_MENU_ITEMS) {
			GetMenuItemInfo(WindowMenu, SWI_WINDOW_MENU_ITEMS, 1, &M);
			SendMessage(Handle, 
				SWI_WM_REMOVE_FROM_WINDOW_MENU, 0, (LPARAM) M.dwItemData);
			DestroyWindow((HWND) M.dwItemData);
		}
		return 0;
	case SWI_MAINMENU_HELP_INTERFACE:
		WinHelp(Handle, "swi.hlp", HELP_FINDER, 0);
		return 0;
	case SWI_MAINMENU_HELP_APPLICATION:
		if (swiHelpFilename()) {
			if (swiExtension(swiHelpFilename(), "hlp")) {
				WinHelp(Handle, swiHelpFilename(), HELP_FINDER, 0);
				return 0;
			}
			OK = GetModuleFileName(GetModuleHandle(0), dir, 2048);
			if (OK) {
				dir[strlen(dir) - strlen(swiFilename(dir))] = 0;
				Exe = ShellExecute(Handle, 
					"open", swiHelpFilename(), 0, dir, SW_SHOWNORMAL);
				if ((int)Exe < 32) switch ((int)Exe) {
				case 0:
					swiErrorMessageBox("System out of resources.");
					break;
				case ERROR_FILE_NOT_FOUND:
					foo = AppendString(
						"Bad help file name. File not found:\n",
						swiHelpFilename());
					swiErrorMessageBox(foo);
					deallocate(foo);
					break;
				case ERROR_PATH_NOT_FOUND:
					foo = AppendString("Path not found:\n", dir);
					swiErrorMessageBox(foo);
					deallocate(foo);
					break;
				case SE_ERR_ASSOCINCOMPLETE:
				case SE_ERR_NOASSOC:
					foo = AppendString(
						"Incomplete or invalid filename association. "
						"Filename:\n",
						swiHelpFilename());
					swiErrorMessageBox(foo);
					deallocate(foo);
					break;
				case SE_ERR_DDEBUSY:
				case SE_ERR_DDEFAIL:
				case SE_ERR_DDETIMEOUT:
					swiErrorMessageBox("DDE failure. Please try again.");
					break;
				case SE_ERR_SHARE:
					foo = AppendString(
						"Sharing violation. Folder:\n", dir);
					bar = AppendString(foo, "\nFile:\n");
					deallocate(foo);
					foo = AppendString(bar, swiHelpFilename());
					deallocate(bar);
					swiErrorMessageBox(foo);
					deallocate(foo);
					break;
				}
			} else
				swiErrorMessageBox(
					"Too short module name buffer, "
					"complain to implementor of swi-library.");
		} else swiInformationMessageBox(
			"No application-specific help available.");
		return 0;
	case SWI_MAINMENU_HELP_ABOUT:
		foo = AppendString(swiApplicationInfo(), swiInternalInfo());
		swiInformationMessageBox(foo);
		deallocate(foo);
		return 0;
	}
	return DefWindowProc(Handle, OrigMessage, wParam, lParam);
}

typedef struct {
	void* Next;
	SWI_ID DataID;
} DataIDStruct;

static DataIDStruct* First = 0;

static int GoThroughDataSets() {
	int Count;
	SWI_ID* Found;
	DataIDStruct* DIS;
	InitIndexedQueue(&IDs);
	swiSIterateWithCallback(&SeekDataSets);
	Found = DequeueFromIndexedQueue(&IDs);
	First = 0;
	Count = 0;
	while (Found) {
		if (!swiDSHasBeenSaved(*Found)) {
			DIS = allocate(sizeof(DataIDStruct));
			DIS->Next = First;
			DIS->DataID = *Found;
			First = DIS;
			Count++;
		}
		deallocate(Found);
		Found = DequeueFromIndexedQueue(&IDs);
	}
	DeleteIndexedQueue(&IDs);
	return Count;
}

static int CheckDataSetSaves() {
	DataIDStruct* Curr;
	int Count;
	SWI_ID* IDs;
	int SavingOK = 1;
	First = 0;
	/* go through all data sets and store unsaved data sets */
	Count = GoThroughDataSets();
	IDs = allocate(Count * sizeof(SWI_ID));
	Curr = First;
	Count = 0;
	while (Curr) {
		IDs[Count++] = Curr->DataID;
		Curr = Curr->Next;
		deallocate(First);
		First = Curr;
	}
	return AskSaveForThese(Count, IDs);
}

static int CheckProcessExiting() {
	SWI_ID Process;
	Process = swiRunningProcess();
	if (!Process) return 1;
	if (swiCurrentProcessCanBeHalted()) {
		swiInformationMessageBox(
			"There is a running process. "
			"Halt it or let it finish before exiting.");
		return 0;
	} else if (swiQuestionMessageBox(
		"There is a running process that cannot be halted, terminate it?",
		0))
	{
		swiKillCurrentProcess();
	} else return 0;
	return 1;
}

static void MakeImageList(HIMAGELIST* Large, HIMAGELIST* Small) {
	HINSTANCE I = GetModuleHandle(0);
	*Large = ImageList_Create(
		32, 32, ILC_COLOR4 | ILC_MASK, SWI_ICON_COUNT, 0);
	*Small = ImageList_Create(
		16, 16, ILC_COLOR4 | ILC_MASK, SWI_ICON_COUNT, 0);
	/* note: the order and count are important, both lists must have the same
	    order and count and they must match the enumeration defined in swi.h */
	ImageList_AddIcon(*Large, LoadIcon(I, MAKEINTRESOURCE(swiTSIcon)));
	ImageList_AddIcon(*Small, LoadIcon(I, MAKEINTRESOURCE(swiTSIconSmall)));
	ImageList_AddIcon(*Large, LoadIcon(I, MAKEINTRESOURCE(swiCBIcon)));
	ImageList_AddIcon(*Small, LoadIcon(I, MAKEINTRESOURCE(swiCBIconSmall)));
	ImageList_AddIcon(*Large, LoadIcon(I, MAKEINTRESOURCE(swiPAIcon)));
	ImageList_AddIcon(*Small, LoadIcon(I, MAKEINTRESOURCE(swiPAIconSmall)));
	ImageList_AddIcon(*Large, LoadIcon(I, MAKEINTRESOURCE(swiGIcon)));
	ImageList_AddIcon(*Small, LoadIcon(I, MAKEINTRESOURCE(swiGIconSmall)));
}

static HWND MakeListView(HIMAGELIST SharedLarge, HIMAGELIST SharedSmall) {
	HWND ListView = CreateWindow(WC_LISTVIEW, "", WS_VISIBLE | WS_CHILD 
		| WS_CLIPSIBLINGS | LVS_EDITLABELS | LVS_ICON | LVS_SHAREIMAGELISTS,
		0, 0, CW_USEDEFAULT, CW_USEDEFAULT, swiMainWindowHandle(), 0, 
		GetModuleHandle(0), 0);
	ListView_SetImageList(ListView, SharedLarge, LVSIL_NORMAL);
	ListView_SetImageList(ListView, SharedSmall, LVSIL_SMALL);
	return ListView;
}

static int GetUnusedID(HMENU WindowMenu) {
	LARGE_INTEGER Time;
	MENUITEMINFO M;
	QueryPerformanceCounter(&Time);
	M.cbSize = sizeof(MENUITEMINFO);
	M.fMask = MIIM_ID;
	M.wID = SWI_MAINMENU_WINDOW_LISTBEGIN 
		+ Time.LowPart % SWI_MAINMENU_WINDOW_LISTSIZE;
	while (GetMenuItemInfo(WindowMenu, M.wID, 0, &M))
		if (++M.wID >= 
			SWI_MAINMENU_WINDOW_LISTBEGIN + SWI_MAINMENU_WINDOW_LISTSIZE) 
		{
			M.wID = SWI_MAINMENU_WINDOW_LISTBEGIN;
		}
	return M.wID;
}

#define MAX_WINDOW_MENU_TEXT 64

LRESULT CALLBACK
MainWindowProcedure(HWND Handle, UINT iMsg, WPARAM wParam, LPARAM lParam) {
	/* only one main window, thus static variables are used */
	static int State;
	static HIMAGELIST LargeImageList, SmallImageList;
	static int XDiv, YDiv;
	int ReturnValue, Width, Height;
	LVITEM Item;
	LPRECT Area;
	RECT Rect;
	int Count, k;
	SWI_ID* IDs;
	HWND View;
	HMENU WindowMenu;
	MENUITEMINFO M;
	POINT P;
	HWND Target;
	DWORD Position;
	HDC DC;
	LVFINDINFO FindInfo;
	char Buffer[128];
	char WindowMenuText[MAX_WINDOW_MENU_TEXT];
	switch (iMsg) {
	case WM_CREATE:
		swiInitializeMainWindowHandle(Handle);
		State = 1;
		StatusBar = CreateStatusWindow(
			WS_CHILD | WS_VISIBLE | SBARS_SIZEGRIP, 0, Handle, 1);
		/* create list views */
		MakeImageList(&LargeImageList, &SmallImageList);
		swiInitializeImageLists(LargeImageList, SmallImageList);
		MainList = MakeListView(LargeImageList, SmallImageList);
		swiSDSetSource(SWI_SELECTED_MAIN, MainList);
		InputList = MakeListView(LargeImageList, SmallImageList);
		swiSDSetSource(SWI_SELECTED_INPUT, InputList);
		OutputList = MakeListView(LargeImageList, SmallImageList);
		swiSDSetSource(SWI_SELECTED_OUTPUT, OutputList);
		OutlinePenDark = CreatePen(PS_SOLID, 3, 0x007f4040);
		OutlinePenLight = CreatePen(PS_SOLID, 3, 0x00ff7f7f);
		SetForegroundWindow(MainList);
		return 0;
	case WM_NOTIFY:
		return HandleWMNotify(Handle, iMsg, wParam, lParam);
	case WM_COMMAND:
		return HandleWMCommand(LOWORD(wParam), HIWORD(wParam), Handle,
			iMsg, wParam, lParam);
	case WM_ACTIVATE:
		swiDDEnd();
		return 0;
	case WM_MOUSEMOVE:
		if (swiDDDragging() && (wParam & MK_LBUTTON != MK_LBUTTON))
			swiDDEnd();
		break;
	case WM_LBUTTONUP:
		if (swiDDDragging()) {
			GetWindowRect(Handle, &Rect);
			Position = GetMessagePos();
			P.x = LOWORD(Position);
			P.y = HIWORD(Position);
			Target = WindowFromPoint(P);
			if (Target == MainList)
				swiSDSetSource(SWI_SELECTED_DRAG_TARGET, MainList);
			else if (Target == InputList)
				swiSDSetSource(SWI_SELECTED_DRAG_TARGET, InputList);
			else if (Target == OutputList)
				swiSDSetSource(SWI_SELECTED_DRAG_TARGET, OutputList);
			else {
				GetClassName(Target, Buffer, 128);
				if (!strcmp("swiImageViewWindow", Buffer)) {
					Count = swiSDSelectedCount(SWI_SELECTED_DRAG_SOURCE);
					if (Count) {
						IDs = allocate(Count * sizeof(SWI_ID));
						swiSDSets(SWI_SELECTED_DRAG_SOURCE, 0, Count, IDs);
						for (k = 0; k < Count; k++) {
							swiDSUse(IDs[k]);
							SendMessage(Target, 
								SWI_WM_ADD_TO_VIEW, (WPARAM)IDs[k], 0);
						}
						deallocate(IDs);
					}
				}
				swiDDEnd();
				return 0;
			}
			if (swiSDGetSource(SWI_SELECTED_DRAG_TARGET) 
				!= swiSDGetSource(SWI_SELECTED_DRAG_SOURCE))
			{
				View = swiSDGetSource(SWI_SELECTED_DRAG_TARGET);
				IDs = 0;
				Count = swiSDRemoveSets(SWI_SELECTED_DRAG_SOURCE, &IDs);
				if (Count) {
					Item.mask = 
						LVIF_TEXT | LVIF_PARAM | LVIF_IMAGE | LVIF_STATE;
					Item.iItem = ListView_GetItemCount(View);
					Item.iSubItem = 0;
					Item.state = 0;
					Item.stateMask = (UINT) -1;
					Item.cchTextMax = 0;
					for (k = 0; k < Count; k++) {
						Item.pszText = swiFilename(swiDSName(IDs[k]));
						Item.iImage = swiDSIcon(IDs[k]);
						Item.lParam = IDs[k];
						ListView_InsertItem(View, &Item);
						Item.iItem++;
					}
					deallocate(IDs);
				}
			} else {
				P.x -= DragBeginPos.x;
				P.y -= DragBeginPos.y;
				swiSDMoveSelected(SWI_SELECTED_DRAG_SOURCE, P.x, P.y);
			}
			swiDDEnd();
			ValidateMenus();
		}
		return 0;
	case WM_SIZE:
		if (wParam == SIZE_MAXIMIZED) {
			if (!State) PostMessage(swiParentWindowHandle(), 
				SWI_WM_CHILD_WINDOW_VISIBILITY, 1, 0);
			State = 2;
		} else if (wParam == SIZE_RESTORED) {
			if (!State) PostMessage(swiParentWindowHandle(), 
				SWI_WM_CHILD_WINDOW_VISIBILITY, 1, 0);
			State = 1;
		} else if (wParam == SIZE_MINIMIZED) {
			State = 0;
			PostMessage(swiParentWindowHandle(), 
				SWI_WM_CHILD_WINDOW_VISIBILITY, 0, 0);
		} else if (wParam == SIZE_MAXHIDE) return 0;
		Width = LOWORD(lParam);
		Height = HIWORD(lParam);
		GetWindowRect(StatusBar, &Rect);
		Height -= (Rect.bottom - Rect.top);
		MoveWindow(StatusBar, 
			0, Height, (Rect.bottom - Rect.top), Width, 1);
		XDiv = (int)(Width * 0.67);
		YDiv = Height - 100;
		MoveWindow(MainList, 3, 3, XDiv - 1, YDiv - 1, 1);
		MoveWindow(InputList, 
			3, YDiv + 6, XDiv - 1, Height - (YDiv + 11), 1);
		MoveWindow(OutputList, 
			XDiv + 6, 3, Width - (XDiv + 11), Height - 8, 1);
		return 0;
	case WM_SIZING:
		Area = (LPRECT) lParam;
		Width = Area->right - Area->left;
		Height = Area->bottom - Area->top;
		if (Width < 300) {
			Width = 300 - Width;
			switch (wParam) {
			case WMSZ_BOTTOMLEFT: Area->left -= Width; break;
			case WMSZ_BOTTOMRIGHT: Area->right += Width; break;
			case WMSZ_LEFT: Area->left -= Width; break;
			case WMSZ_RIGHT: Area->right += Width; break;
			case WMSZ_TOPLEFT: Area->left -= Width; break;
			case WMSZ_TOPRIGHT: Area->right += Width; break;
			}
		}
		if (Height < 270) {
			Height = 270 - Height;
			switch (wParam) {
			case WMSZ_BOTTOM: Area->bottom += Height; break;
			case WMSZ_BOTTOMLEFT: Area->bottom += Height; break;
			case WMSZ_BOTTOMRIGHT: Area->bottom += Height; break;
			case WMSZ_TOP: Area->top -= Height; break;
			case WMSZ_TOPLEFT: Area->top -= Height; break;
			case WMSZ_TOPRIGHT: Area->top -= Height; break;
			}
		}
		break;
	case WM_PAINT:
		if (GetUpdateRect(Handle, &Rect, TRUE)) {
			DC = GetDC(Handle);
			ShowCursor(FALSE);
			Outline(DC, ActiveListView);
			ValidateRect(Handle, &Rect);
			ShowCursor(TRUE);
		}
		return 0;
	case WM_CLOSE:
		if (!CheckDataSetSaves()) return 0;
		if (!CheckProcessExiting()) return 0;
		swiDestroyProcessAndOptionsDialog();
		swiDestroyProcessQueueDialog();
		DestroyWindow(swiParentWindowHandle());
		DestroyWindow(Handle);
		return 0;
	case WM_DESTROY:
		ImageList_Destroy(LargeImageList);
		ImageList_Destroy(SmallImageList);
		DeleteObject(OutlinePenDark);
		DeleteObject(OutlinePenLight);
		DeleteObject(MainWindowBackgroundBrush);
		PostQuitMessage(0);
		return 0;
	case SWI_WM_THREAD_FINISHED:
		ReturnValue = swiPRFinish(swiRunningProcess());
		switch (ReturnValue) {
		case EVERYTHING_OK: break;
		case FATAL_ERROR:
			swiWarningMessageBox(
				"Process terminated with fatal exit value.\n"
				"It is recommended that you save any unsaved "
				"data and exit program.");
			break;
		case EXIT_REQUEST_HONOURED:
			swiInformationMessageBox("Process terminated as requested.");
			break;
		case RECOVERABLE_ERROR:
			swiInformationMessageBox(
				"Process terminated because of internal error.\n"
				"Please try again.");
			break;
		case RECOVERABLE_ERROR_UNCONTROLLED_EXIT:
			swiWarningMessageBox(
				"Process terminated uncontrollably due to internal error.\n"
				"Some resources may have been left unreleased.");
			break;
		}
		swiRunQueuedProcess();
		swiProcessQueueUpdate();
		CloseInfoFlow();
		CloseProgressWindow();
		return 0;
	case SWI_WM_CHECK_NEW_PROCESSES:
		if (!swiRunningProcess() && swiRunQueuedProcess())
			EnableMenuItem(Menu, 
				SWI_MAINMENU_PROCESS_OUTPUT, MF_BYCOMMAND | MF_ENABLED);
		DrawMenuBar(Handle);
		swiProcessQueueUpdate();
		return 0;
	case SWI_WM_NEW_DATASET:
		Item.mask = LVIF_TEXT | LVIF_IMAGE | LVIF_PARAM | LVIF_STATE;
		Item.iItem = ListView_GetItemCount(wParam ? OutputList : MainList);
		Item.iSubItem = 0;
		Item.state = 0;
		Item.stateMask = (UINT)-1;
		Item.pszText = swiFilename(swiDSName((SWI_ID)lParam));
		Item.cchTextMax = 0;
		Item.iImage = swiDSIcon((SWI_ID)lParam);
		Item.lParam = lParam;
		if (wParam) ListView_InsertItem(OutputList, &Item);
		else ListView_InsertItem(MainList, &Item);
		ValidateMenus();
		return 0;
	case SWI_WM_ADD_TO_WINDOW_MENU:
		EnableMenuItem(Menu, 
			SWI_MAINMENU_WINDOW_CLOSEALL, MF_BYCOMMAND | MF_ENABLED);
		GetWindowText((HWND) lParam, WindowMenuText, MAX_WINDOW_MENU_TEXT);
		WindowMenuText[MAX_WINDOW_MENU_TEXT - 1] = 0;
		WindowMenu = GetSubMenu(Menu, 3);
		M.cbSize = sizeof(MENUITEMINFO);
		M.fMask = MIIM_DATA | MIIM_ID | MIIM_STATE | MIIM_TYPE;
		M.fType = MFT_STRING;
		M.fState = MFS_ENABLED;
		M.wID = GetUnusedID(WindowMenu);
		M.hSubMenu = 0;
		M.hbmpChecked = 0;
		M.hbmpUnchecked = 0;
		M.dwItemData = lParam;
		M.dwTypeData = WindowMenuText;
		M.cch = strlen(WindowMenuText);
		InsertMenuItem(WindowMenu, GetMenuItemCount(WindowMenu) + 1, 1, &M);
		DrawMenuBar(Handle);
		return 0;
	case SWI_WM_REMOVE_FROM_WINDOW_MENU:
		WindowMenu = GetSubMenu(Menu, 3);
		M.cbSize = sizeof(MENUITEMINFO);
		M.fMask = MIIM_DATA;
		for (k = SWI_WINDOW_MENU_ITEMS; 
			k < GetMenuItemCount(WindowMenu); k++)
		{
			GetMenuItemInfo(WindowMenu, k, 1, &M);
			if ((long)(M.dwItemData) == lParam) {
				RemoveMenu(WindowMenu, k, MF_BYPOSITION);
				if (GetMenuItemCount(WindowMenu) == SWI_WINDOW_MENU_ITEMS) 
					EnableMenuItem(Menu, SWI_MAINMENU_WINDOW_CLOSEALL, 
						MF_BYCOMMAND | MF_GRAYED);
				DrawMenuBar(Handle);
				return 0;
			}
		}
		return 0;
	case SWI_WM_RENAMED:
		FindInfo.flags = LVFI_PARAM;
		FindInfo.lParam = lParam;
		k = ListView_FindItem(MainList, -1, &FindInfo);
		if (k != -1) View = MainList;
		else {
			k = ListView_FindItem(InputList, -1, &FindInfo);
			if (k != -1) View = InputList;
			else {
				k = ListView_FindItem(OutputList, -1, &FindInfo);
				if (k != -1) View = OutputList;
				else View = 0;
			}
		}
		if (View) {
			ListView_SetItemText(View, 
				k, 0, swiFilename(swiDSName((SWI_ID) lParam)));
		} else return 0;
		WindowMenu = GetSubMenu(Menu, 3);
		M.cbSize = sizeof(MENUITEMINFO);
		M.fMask = MIIM_DATA;
		for (k = SWI_WINDOW_MENU_ITEMS; 
			k < GetMenuItemCount(WindowMenu); k++)
		{
			GetMenuItemInfo(WindowMenu, k, 1, &M);
			GetClassName((HWND) M.dwItemData, Buffer, 128);
			if (!stricmp(Buffer, "swiImageViewWindow"))
				SendMessage((HWND) M.dwItemData, SWI_WM_RENAMED, 0, lParam);
		}
		return 0;
	case SWI_WM_CLOSE_SELECTED:
		Count = swiSDSelectedCount(SWI_SELECTED_ACTIVE);
		if (Count) {
			IDs = allocate(Count * sizeof(SWI_ID));
			swiSDSets(SWI_SELECTED_ACTIVE, 0, Count, IDs);
			if (AskSaveForThese(Count, IDs)) {
				deallocate(IDs);
				IDs = 0;
				swiSDRemoveSets(SWI_SELECTED_ACTIVE, &IDs);
				for (k = 0; k < Count; k++) swiDSDelete(IDs[k]);
				deallocate(IDs);
			} else deallocate(IDs);
		}
		ValidateMenus();
		return 0;
	}
	return DefWindowProc(Handle, iMsg, wParam, lParam);
}


static void RegisterMainWindowClass() {
	WNDCLASSEX WC;
	COLORREF Color, C;
	LOGBRUSH B;
    WC.cbSize = sizeof(WC);
    WC.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
    WC.lpfnWndProc = &MainWindowProcedure;
    WC.cbClsExtra = 0;
    WC.cbWndExtra = 0;
    WC.hInstance = GetModuleHandle(0);
    WC.hIcon = swiAppIcon();
    WC.hCursor = 0;
	Color = GetSysColor(COLOR_WINDOW);
	C = Color & 255;
	C *= 0.9375;
	Color = (Color & 0x00ffff00) | C;
	C = (Color >> 8) & 255;
	C *= 0.9375;
	Color = (Color & 0x00ff00ff) | (C << 8);
	C = (Color >> 16) & 255;
	C *= 0.9375;
	Color = (Color & 0x0000ffff) | (C << 16);
	B.lbStyle = BS_SOLID;
	B.lbColor = Color;
	B.lbHatch = 0;
	MainWindowBrushColor = Color;
    WC.hbrBackground = MainWindowBackgroundBrush = CreateBrushIndirect(&B);
    WC.lpszMenuName = "swiMainMenu";
    WC.lpszClassName = "MainWindow";
    WC.hIconSm = NULL;
    RegisterClassEx(&WC);
}


int SeekWindows(void* Data) {
	SWI_ID WindowID;
	SWI_ID* FoundWindow;
	if (((SWI_ID*)Data)[1] == OWNED_WINDOW) {
		WindowID = *((SWI_ID*)Data);
		FoundWindow = allocate(sizeof(SWI_ID));
		*FoundWindow = WindowID;
		EnqueueIntoIndexedQueue(&IDs, FoundWindow);
	}
	return 0;
}


LRESULT CALLBACK
ParentWindowProcedure(HWND Handle, UINT iMsg, WPARAM wParam, LPARAM lParam) {
	SWI_ID* FoundWindow;
	switch (iMsg) {
	case WM_CREATE:
		swiInitializeParentWindowHandle(Handle);
		return 0;
	case WM_CLOSE:
		PostMessage(swiMainWindowHandle(), WM_CLOSE, 0, 0);
		return 0;
	case SWI_WM_CHILD_WINDOW_VISIBILITY:
		InitIndexedQueue(&IDs);
		swiSIterateWithCallback(&SeekWindows);
		FoundWindow = DequeueFromIndexedQueue(&IDs);
		while (FoundWindow) {
			if (wParam) swiOWShow(*FoundWindow);
			else swiOWHide(*FoundWindow);
			deallocate(FoundWindow);
			FoundWindow = DequeueFromIndexedQueue(&IDs);
		}
		DeleteIndexedQueue(&IDs);
		return 0;
	}
	return DefWindowProc(Handle, iMsg, wParam, lParam);
}

static void RegisterParentWindowClass() {
	WNDCLASSEX WC;
    WC.cbSize = sizeof(WC);
    WC.style = CS_HREDRAW | CS_VREDRAW;
    WC.lpfnWndProc = &ParentWindowProcedure;
    WC.cbClsExtra = 0;
    WC.cbWndExtra = 0;
    WC.hInstance = GetModuleHandle(0);
    WC.hIcon = swiAppIcon();
    WC.hCursor = 0;
    WC.hbrBackground = (HBRUSH) (COLOR_APPWORKSPACE + 1);
    WC.lpszMenuName = 0;
    WC.lpszClassName = "ParentWindow";
    WC.hIconSm = NULL;
    RegisterClassEx(&WC);
}



int swiRun(char* MainWindowTitle, char* ApplicationInfo, char* HelpFilename)
{
	MSG msg;
	HWND WindowHandle;
	HACCEL AcceleratorsHandle;
	HWND QueueHandle, DialogHandle, OutputHandle;

	swiSetAppIcon(
		LoadIcon(GetModuleHandle(0), MAKEINTRESOURCE(IDI_APP_ICON)));

	InitCommonControls();
	AcceleratorsHandle = swiWACreate();

	RegisterParentWindowClass();
	CreateWindow("ParentWindow", MainWindowTitle, 
		WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
		100, 100, 200, 100, 0, 0, GetModuleHandle(0), 0);

	RegisterMainWindowClass();
	WindowHandle = CreateWindow("MainWindow", MainWindowTitle,
		WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
		CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
		0, 0, GetModuleHandle(0), NULL);

	Menu = GetMenu(WindowHandle);
	/* append to Process-menu */
	swiWMAppendToMenu(GetSubMenu(Menu, 1));
	swiInitializeApplicationData(
		MainWindowTitle, ApplicationInfo, HelpFilename);
	if (!swiHelpFilename()) EnableMenuItem(Menu, 
		SWI_MAINMENU_HELP_APPLICATION, MF_BYCOMMAND | MF_GRAYED);
	ShowWindow(WindowHandle, SW_SHOWNORMAL);
	UpdateWindow(WindowHandle);

	while (GetMessage(&msg, NULL, 0, 0)) {
		/* check modeless dialogs and accelerators */
		QueueHandle = swiProcessQueueDialogHandle();
		DialogHandle = swiProcessAndOptionsDialogHandle();
		OutputHandle = swiOutputFlowDialogHandle();
		if (QueueHandle && IsDialogMessage(QueueHandle, &msg));
		else if (DialogHandle && IsDialogMessage(DialogHandle, &msg));
		else if (OutputHandle && IsDialogMessage(OutputHandle, &msg));
		else if (
			TranslateAccelerator(WindowHandle, AcceleratorsHandle, &msg));
		else {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
	swiClearApplicationData();
	return msg.wParam;
}
