/********************************************************************

	swi_OutputFlow.c

	Ismo Krkkinen

	Last changed 12.11.1999

********************************************************************/

#include "swi_OutputFlow.h"
#include <windows.h>
#include <commctrl.h>
#include "memctrl.h"
#include <string.h>
#include "swi_ProcessOutputSymbols.h"
#include "swi_WindowsControls.h"
#include "swi_MainMenuSymbols.h"
#include "swi_OwnedWindow.h"
#include "swi_Messages.h"
#include "swi_Storage.h"

static HWND OutputFlowHandle = 0;
static SWI_ID OutputFlowOWID = 0;

HWND swiOutputFlowDialogHandle() {
	return OutputFlowHandle;
}

typedef struct {
	SWI_ID ID;
	SWIDataTypes Type;
	int Lines;
	HWND Handle;
	char* LeftOverLine;
	int BarValue;
} OutputFlow;


static void AddLine(HWND ListBoxHandle, OutputFlow* OF, char* String) {
	int k;
	k = swiLBFocusIndex(ListBoxHandle);
	if (k + 3 >= swiLBCount(ListBoxHandle)) {
		k = swiLBAdd(ListBoxHandle, String, 0);
		swiLBSelectSingle(ListBoxHandle, k);
		/*swiLBSetFocusIndex(ListBoxHandle, k);*/
	} else swiLBAdd(ListBoxHandle, String, 0);
	deallocate(String);
	OF->Lines++;
	if (OF->Lines >= 999) {
		if (swiLBFocusIndex(ListBoxHandle) < 1)
			swiLBSetFocusIndex(ListBoxHandle, 1);
		OF->Lines--;
		swiLBRemove(ListBoxHandle, 0);
	}
}


BOOL APIENTRY ProcessOutputDialogProcedure(
	HWND Handle, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
	switch (iMsg) {
	case WM_ACTIVATE:
		if ((LOWORD(wParam) == WA_ACTIVE) 
			|| (LOWORD(wParam) == WA_CLICKACTIVE))
		{
			OutputFlowHandle = Handle;
		}
		SetFocus(GetDlgItem(Handle, SWI_PROCESS_OUTPUT_LISTBOX));
		return TRUE;
	case WM_INITDIALOG:
		PostMessage(Handle, WM_SETICON, ICON_BIG, (LPARAM) swiAppIcon());
		EnableMenuItem(GetMenu(swiMainWindowHandle()),
			SWI_MAINMENU_PROCESS_OUTPUT, MF_BYCOMMAND | MF_ENABLED);
		OutputFlowOWID = swiOWNew(Handle);
		return TRUE;
	case WM_COMMAND:
		if ((LOWORD(wParam) == SWI_PROCESS_OUTPUT_HIDE)
			&& (HIWORD(wParam) == BN_CLICKED))
		{
			ShowWindow(Handle, SW_HIDE);
			return TRUE;
		}
		break;
	case WM_SHOWWINDOW:
		if (wParam) swiOWSelfShow(OutputFlowOWID);
		else swiOWSelfHide(OutputFlowOWID);
		break;
	case WM_CLOSE:
		ShowWindow(Handle, SW_HIDE);
		return 0;
	case WM_DESTROY:
		swiOWDelete(OutputFlowOWID);
		EnableMenuItem(GetMenu(swiMainWindowHandle()),
			SWI_MAINMENU_PROCESS_OUTPUT, MF_BYCOMMAND | MF_GRAYED);
		return 0;
	case SWI_WM_ADD_STRING:
		AddLine(GetDlgItem(Handle, SWI_PROCESS_OUTPUT_LISTBOX), 
			(OutputFlow*)wParam, (char*)lParam);
		return 0;
	}
	return FALSE;
}


SWI_ID swiOFNew() {
	OutputFlow* OF = allocate(sizeof(OutputFlow));
	OF->Type = OUTPUT_FLOW;
	OF->BarValue = 0;
	OF->Lines = 0;
	if (swiMainWindowHandle()) {
		OF->Handle = CreateDialog(
			GetModuleHandle(0), "swiDialogProcessOutput",
			swiParentWindowHandle(), &ProcessOutputDialogProcedure);
		OutputFlowHandle = OF->Handle; /* because only one is allowed */
		ShowWindow(OF->Handle, SW_SHOW);
		UpdateWindow(OF->Handle);
	} else OF->Handle = 0;
	OF->LeftOverLine = 0;
	swiSInsert(OF);
	return OF->ID;
}

void swiOFDelete(SWI_ID FlowID) {
	OutputFlow* OF = swiSRemove(FlowID, OUTPUT_FLOW);
	if (OF) {
		if (OF->Handle) DestroyWindow(OF->Handle);
		deallocate(OF);
	}
}


void swiOFAppend(SWI_ID FlowID, char* Message) {
	HWND ListBoxHandle;
	OutputFlow* OF;
	char* Newline, *LineToAppend;
	int NewLen, LeftLen;
	OF = swiSFind(FlowID, OUTPUT_FLOW);
	if (OF && OF->Handle) {
		ListBoxHandle = GetDlgItem(OF->Handle, SWI_PROCESS_OUTPUT_LISTBOX);
		Newline = strchr(Message, '\n');
		while (Newline) {
			*Newline = 0;
			/* did previous message omit newline at the end? */
			if (OF->LeftOverLine) LeftLen = strlen(OF->LeftOverLine);
			else LeftLen = 0;
			NewLen = LeftLen + strlen(Message) + 1;
			LineToAppend = allocate(NewLen);
			*LineToAppend = 0;
			/* if we have something from previous message, 
			copy it and destroy */
			if (LeftLen) {
				strcpy(LineToAppend, OF->LeftOverLine);
				deallocate(OF->LeftOverLine);
				OF->LeftOverLine = 0;
			}
			strcat(LineToAppend, Message);
			SendMessage(OF->Handle, 
				SWI_WM_ADD_STRING, (WPARAM) OF, (LPARAM) LineToAppend);
			*Newline = '\n';
			Message = Newline + 1;
			Newline = strchr(Message, '\n');
		}
		/* we have text without newline at the end */
		if (*Message) {
			if (OF->LeftOverLine) {
				LeftLen = strlen(OF->LeftOverLine);
				/* forgetting newlines systematically? */
				if (LeftLen >= 80) {
					OF->LeftOverLine[LeftLen + 1] = 0;
					SendMessage(OF->Handle, SWI_WM_ADD_STRING, (WPARAM) OF,
						(LPARAM) OF->LeftOverLine);
					OF->LeftOverLine = 0;
				} else {
					Newline = allocate(LeftLen + strlen(Message) + 1);
					strcpy(Newline, OF->LeftOverLine);
					strcat(Newline, Message);
					deallocate(OF->LeftOverLine);
					OF->LeftOverLine = Newline;
				}
			} else {
				OF->LeftOverLine = allocate(strlen(Message) + 1);
				strcpy(OF->LeftOverLine, Message);
			}
		}
	}
}

void swiOFClear(SWI_ID FlowID) {
	OutputFlow* OF = swiSFind(FlowID, OUTPUT_FLOW);
	if (OF && OF->Handle) {
		swiLBRemoveAll(GetDlgItem(OF->Handle, SWI_PROCESS_OUTPUT_LISTBOX));
		if (OF->LeftOverLine) {
			deallocate(OF->LeftOverLine);
			OF->LeftOverLine = 0;
		}
		OF->Lines = 0;
		swiOFClearProgress(FlowID);
	}
}

void swiOFFlush(SWI_ID FlowID) {
	int LeftLen;
	OutputFlow* OF = swiSFind(FlowID, OUTPUT_FLOW);
	if (OF && OF->LeftOverLine) {
		LeftLen = strlen(OF->LeftOverLine);
		OF->LeftOverLine[LeftLen + 1] = 0;
		SendMessage(OF->Handle, SWI_WM_ADD_STRING, (WPARAM) OF, 
			(LPARAM) OF->LeftOverLine);
		OF->LeftOverLine = 0;
	}
}


void swiOFClearProgress(SWI_ID FlowID) {
	swiOFSetProgress(FlowID, 0, 100);
	swiOFLabelProgress(FlowID, 0);
}

void swiOFSetProgress(SWI_ID FlowID, int State, int Total) {
	int Value;
	OutputFlow* OF = swiSFind(FlowID, OUTPUT_FLOW);
	if (OF && OF->Handle) {
		if (Total) {
			Value = (100 * State) / Total;
			if (Value > 100) Value = 100;
			else if (Value < 0) Value = 0;
		} else Value = 100;
		SendMessage(GetDlgItem(OF->Handle, SWI_PROCESS_OUTPUT_PROGRESSBAR),
			PBM_SETPOS, Value, 0);
	}
}

void swiOFLabelProgress(SWI_ID FlowID, char* Label) {
	OutputFlow* OF = swiSFind(FlowID, OUTPUT_FLOW);
	if (OF && OF->Handle) {
		if (Label)
			SendMessage(GetDlgItem(OF->Handle, SWI_PROCESS_OUTPUT_STATIC),
				WM_SETTEXT, 0, (LONG)Label);
		else
			SendMessage(GetDlgItem(OF->Handle, SWI_PROCESS_OUTPUT_STATIC),
				WM_SETTEXT, 0, (LONG)"");
	}
}

