/********************************************************************

	swi_Parameters.c

	Ismo Krkkinen

	Last changed 12.11.1999

********************************************************************/

#include "swi_Parameters.h"
#include "memctrl.h"
#include "swi_Storage.h"
#include "string.h"
#include "swi_Integer.h"
#include "swi_Boolean.h"
#include "swi_Enumerated.h"
#include "resource.h"
#include "swi_WindowsControls.h"

static int First = 0;
static int Last = -1;
static SWI_PCHAR_INT Document = 0;
static SWI_INT_INT Type = 0;
static SWI_INT_INT GetValue = 0;
static SWI_VOID_INT_INT SetValue = 0;
static SWI_INT_INT MaxValue = 0;
static SWI_INT_INT MinValue = 0;
static SWI_INT_INT DefaultValue = 0;
static SWI_PCHAR_INT_INT ValueName = 0;

void swiFullParameterSet(
	int FirstOption, int LastOption,
	SWI_PCHAR_INT DocumentFunc,
	SWI_INT_INT TypeFunc,
	SWI_INT_INT GetValueFunc,
	SWI_VOID_INT_INT SetValueFunc,
	SWI_INT_INT MaxValueFunc,
	SWI_INT_INT MinValueFunc,
	SWI_INT_INT DefaultValueFunc,
	SWI_PCHAR_INT_INT ValueNameFunc)
{
	First = FirstOption;
	Last = LastOption;
	Document = DocumentFunc;
	Type = TypeFunc;
	GetValue = GetValueFunc;
	SetValue = SetValueFunc;
	MaxValue = MaxValueFunc;
	MinValue = MinValueFunc;
	DefaultValue = DefaultValueFunc;
	ValueName = ValueNameFunc;
}

int swiFPSGetFirst() { return First; }

int swiFPSGetLast() { return Last; }

char* swiFPSGetDocumentString(int Option) {
	if (First <= Option && Option <= Last && Document)
		return (*Document)(Option);
	return 0;
}

int swiFPSGetType(int Option) {
	if (First <= Option && Option <= Last && Type)
		return (*Type)(Option);
	return -1;
}

int swiFPSValue(int Option) {
	if (First <= Option && Option <= Last && GetValue)
		return (*GetValue)(Option);
	return -1;
}

void swiFPSSetValue(int Option, int Value) {
	if (First <= Option && Option <= Last && SetValue)
		(*SetValue)(Option, Value);
}

int swiFPSGetMaxValue(int Option) {
	if (First <= Option && Option <= Last && MaxValue)
		return (*MaxValue)(Option);
	return -1;
}

int swiFPSGetMinValue(int Option) {
	if (First <= Option && Option <= Last && MinValue)
		return (*MinValue)(Option);
	return -1;
}

int swiFPSGetDefaultValue(int Option) {
	if (First <= Option && Option <= Last && DefaultValue)
		return (*DefaultValue)(Option);
	return -1;
}

char* swiFPSGetEnumValueName(int Option, int Value) {
	if (First <= Option && Option <= Last && Type && (*Type)(Option) == 2)
		return (*ValueName)(Option, Value);
	return 0;
}

static char Out[1024];

char* swiFPSValueAsString(int Option, int Value) {
	int Failure = 0;
	switch (swiFPSGetType(Option)) {
	case 0: /* BOOL */
		if (Value) strcpy(Out, "Yes");
		else strcpy(Out, "No");
		break;
	case 1: /* INT */
		itoa(Value, Out, 10);
		break;
	case 2: /* ENUM */
		strcpy(Out, swiFPSGetEnumValueName(Option, Value));
		break;
	default:
		swiErrorMessageBox(
			"swi_Parameters::swiFPSValueAsString: Unknown option type");
		Failure = 1;
	}
	if (!Failure) return Out;
	return 0;
}


typedef struct {
	int IsMember;
	int Value;
	int Changed;
} Parameter;

typedef struct {
	SWI_ID ID;
	SWIDataTypes Type;
	Parameter* Parameters;
	int Offset;
} ParameterSet;



SWI_ID swiPSNew(int ParameterCount, int* Parameters) {
	ParameterSet* PS;
	int k;
	PS = allocate(sizeof(ParameterSet));
	PS->Type = PARAMETER_SET;
	PS->Offset = -First;
	PS->Parameters = allocate((Last - First + 1) * sizeof(Parameter));
	for (k = First; k <= Last; k++) {
		PS->Parameters[k + PS->Offset].IsMember = 0;
		PS->Parameters[k + PS->Offset].Value = swiFPSGetDefaultValue(k);
		PS->Parameters[k + PS->Offset].Changed = 0;
	}
	for (k = 0; k < ParameterCount; k++)
		PS->Parameters[Parameters[k] + PS->Offset].IsMember = 1;
	swiSInsert(PS);
	return PS->ID;
}

void swiPSCopyOver(SWI_ID FromID, SWI_ID ToID) {
	int k;
	ParameterSet* From, *To;
	From = swiSFind(FromID, PARAMETER_SET);
	To = swiSFind(ToID, PARAMETER_SET);
	if (From && To) {
		for (k = 0; k < Last - First + 1; k++) {
			To->Parameters[k].IsMember = From->Parameters[k].IsMember;
			To->Parameters[k].Value = From->Parameters[k].Value;
			To->Parameters[k].Changed = From->Parameters[k].Changed;
		}
	}
}

SWI_ID swiPSCopy(SWI_ID ParameterSetID) {
	ParameterSet* PS;
	ParameterSet* NewPS;
	PS = swiSFind(ParameterSetID, PARAMETER_SET);
	if (PS) {
		NewPS = allocate(sizeof(ParameterSet));
		NewPS->Type = PARAMETER_SET;
		NewPS->Offset = PS->Offset;
		NewPS->Parameters = 
			allocate((Last - First + 1) * sizeof(Parameter));
		swiSInsert(NewPS);
		swiPSCopyOver(ParameterSetID, NewPS->ID);
		return NewPS->ID;
	}
	return 0;
}

void swiPSDelete(SWI_ID ParameterSetID) {
	ParameterSet* PS = swiSRemove(ParameterSetID, PARAMETER_SET);
	if (PS) {
		deallocate(PS->Parameters);
		deallocate(PS);
	}
}

int swiPSIsMember(SWI_ID ParameterSetID, int Parameter) {
	ParameterSet* PS = swiSFind(ParameterSetID, PARAMETER_SET);
	if ((First <= Parameter) && (Parameter <= Last) && PS)
		return PS->Parameters[PS->Offset + Parameter].IsMember;
	return 0;
}

int swiPSValue(SWI_ID ParameterSetID, int Parameter) {
	ParameterSet* PS = swiSFind(ParameterSetID, PARAMETER_SET);
	if ((First <= Parameter) && (Parameter <= Last) && PS)
		return PS->Parameters[PS->Offset + Parameter].Value;
	return 0;
}

void swiPSSetValue(SWI_ID ParameterSetID, int Parameter, int Value) {
	ParameterSet* PS = swiSFind(ParameterSetID, PARAMETER_SET);
	if ((First <= Parameter) && (Parameter <= Last) && PS)
		PS->Parameters[PS->Offset + Parameter].Value = Value;
}

void swiPSSetDefaults(SWI_ID ParameterSetID) {
	int k;
	ParameterSet* PS = swiSFind(ParameterSetID, PARAMETER_SET);
	if (PS) {
		for (k = First; k <= Last; k++)
			swiPSSetValue(ParameterSetID, k, swiFPSGetDefaultValue(k));
	}
}

void swiPSModifyFullSet(SWI_ID ParameterSetID) {
	int k;
	ParameterSet* PS = swiSFind(ParameterSetID, PARAMETER_SET);
	if (PS) for (k = 0; k < Last - First + 1; k++)
		swiFPSSetValue(First + k, PS->Parameters[k].Value);
}

int swiPSExists(SWI_ID ParameterSetID) {
	return swiSFind(ParameterSetID, PARAMETER_SET) != 0;
}




int swiChangeOption(SWI_ID ParameterSetID, int Selected, HWND Parent) {
	int Result, Old;
	int Failure = 0;
	Old = swiPSValue(ParameterSetID, Selected);
	switch (swiFPSGetType(Selected)) {
	case 0: /* BOOL */
		Result = swiGetBooleanValue(Selected, Old, Parent);
		break;
	case 1: /* INT */
		Result = swiGetIntegerValue(Selected, Old, Parent);
		break;
	case 2: /* ENUM */
		Result = swiGetEnumeratedValue(Selected, Old, Parent);
		break;
	default:
		swiErrorMessageBox(
			"swi_Parameters::ChangeOption: Unknown option type");
		Result = 0;
		Failure = 1;
	}
	if (!Failure) swiPSSetValue(ParameterSetID, Selected, Result);
	return Result;
}

static void SetParameterList(HWND Handle, SWI_ID ParameterSetID) {
	int k;
	char* foo, *bar;
	HWND ListBox = GetDlgItem(Handle, SWI_OPTIONS_LISTBOX);
	swiLBRemoveAll(ListBox);
	for (k = swiFPSGetFirst(); k <= swiFPSGetLast(); k++) {
		if (swiPSIsMember(ParameterSetID, k)) {
			bar = AppendString(swiFPSGetDocumentString(k), "\t");
			foo = AppendString(bar, 
				swiFPSValueAsString(k, swiPSValue(ParameterSetID, k)));
			deallocate(bar);
			swiLBAdd(ListBox, foo, k);
			deallocate(foo);
		}
	}
}

static void SetValueDisplay(HWND Handle, int Option, int Value) {
	char *foo, *bar;
	HWND Label = GetDlgItem(Handle, SWI_OPTIONS_VALUE);
	if (swiFPSGetType(Option) == 1) {
		foo = AppendString("Min: ", 
			swiFPSValueAsString(Option, swiFPSGetMinValue(Option)));
		bar = AppendString(foo, ", Curr: ");
		deallocate(foo);
		foo = AppendString(bar, swiFPSValueAsString(Option, Value));
		deallocate(bar);
		bar = AppendString(foo, ", Max: ");
		deallocate(foo);
		foo = AppendString(bar, 
			swiFPSValueAsString(Option, swiFPSGetMaxValue(Option)));
		deallocate(bar);
		SetWindowText(Label, foo);
		deallocate(foo);
	} else SetWindowText(Label, swiFPSValueAsString(Option, Value));
}

static SWI_ID ParameterSetIDForDialog;

BOOL ODHandleWMCommand(int ID, int Message, HWND Handle, LPARAM lParam) {
	HWND h;
	int Selected, Result;
	switch (ID) {
	case SWI_OPTIONS_LISTBOX:
		if (Message == LBN_DBLCLK) {
			Selected = swiLBTag(
				(HWND)lParam, swiLBFocusIndex((HWND)lParam));
			Result = swiChangeOption(
				ParameterSetIDForDialog, Selected, Handle);
			swiPSSetValue(ParameterSetIDForDialog, Selected, Result);
			SetValueDisplay(Handle, Selected, Result);
			SetParameterList(Handle, ParameterSetIDForDialog);
			return TRUE;
		} else if (Message == LBN_SELCHANGE) {
			Selected = swiLBTag(
				(HWND)lParam, swiLBFocusIndex((HWND)lParam));
			SetValueDisplay(Handle, Selected,
				swiPSValue(ParameterSetIDForDialog, Selected));
			return TRUE;
		}
		break;
	case SWI_OPTIONS_CHANGE:
		if (Message == BN_CLICKED) {
			h = GetDlgItem(Handle, SWI_OPTIONS_LISTBOX);
			Selected = swiLBTag(h, swiLBFocusIndex(h));
			Result = swiChangeOption(
				ParameterSetIDForDialog, Selected, Handle);
			swiPSSetValue(ParameterSetIDForDialog, Selected, Result);
			SetValueDisplay(Handle, Selected, Result);
			SetParameterList(Handle, ParameterSetIDForDialog);
			return TRUE;
		}
		break;
	case SWI_OPTIONS_DEFAULTS:
		if (Message == BN_CLICKED) {
			swiPSSetDefaults(ParameterSetIDForDialog);
			SetParameterList(Handle, ParameterSetIDForDialog);
			return TRUE;
		}
		break;
	case IDCANCEL:
		if (Message == BN_CLICKED) EndDialog(Handle, 0);
		return TRUE;
	case IDOK:
		if (Message == BN_CLICKED) EndDialog(Handle, 1);
		return TRUE;
	}
	return FALSE;
}


BOOL APIENTRY OptionsDialogProcedure(HWND Handle, UINT Message,
	WPARAM wParam, LPARAM lParam)
{
	int TabStop;
	switch (Message) {
	case WM_INITDIALOG:
		PostMessage(Handle, WM_SETICON, ICON_BIG, (LPARAM) swiAppIcon());
		SetParameterList(Handle, ParameterSetIDForDialog);
		TabStop = 125;
		SendMessage(GetDlgItem(Handle, SWI_OPTIONS_LISTBOX),
			LB_SETTABSTOPS, 1, (LPARAM) &TabStop);
		return 0;
	case WM_COMMAND:
		return ODHandleWMCommand(
			LOWORD(wParam), HIWORD(wParam), Handle, lParam);
	case WM_CLOSE:
		EndDialog(Handle, 0);
		return TRUE;
	}
	return FALSE;
}

void swiOptionsDialog(SWI_ID ParameterSetID, HWND Parent) {
	if (ParameterSetID) {
		ParameterSetIDForDialog = swiPSCopy(ParameterSetID);
		if (DialogBox(GetModuleHandle(0), 
			"swiDialogOptions", Parent, &OptionsDialogProcedure))
		{
			swiPSCopyOver(ParameterSetIDForDialog, ParameterSetID);
		}
		swiPSDelete(ParameterSetIDForDialog);
	}
}
