/********************************************************************

	swi_Process.c

	Ismo Krkkinen

	Last changed 12.11.1999

********************************************************************/

#include "swi_Process.h"
#include "swi_Storage.h"
#include "idxqueue.h"
#include "interfc.h"
#include <windows.h>
#include "swi_Messages.h"
#include "memctrl.h"
#include <process.h>
#include "swi_WindowsAccelerator.h"
#include "swi_WindowsMenus.h"
#include "swi_Requirement.h"
#include "swi_SelectedData.h"
#include "swi_Internal.h"
#include "swi_Parameters.h"
#include "swi_DataClass.h"

INDEXEDQUEUE ProcessQueue;

static int UserRequestedExit = 0;
int LastExitValue = 0;
SWI_ID RunningProcessID = 0;

static int Overwrite = SWI_OVERWRITE_NEVER;

unsigned long ThreadHandle = 0;

void swiInitializeProcess() {
	InitIndexedQueue(&ProcessQueue);
	UserRequestedExit = 0;
	LastExitValue = 0;
	RunningProcessID = 0;
}

int swiEnqueueProcess(SWI_ID ProcessID) {
	SWI_ID* Item;
	Item = allocate(sizeof(SWI_ID));
	if (Item) {
		*Item = ProcessID;
		EnqueueIntoIndexedQueue(&ProcessQueue, Item);
		return NumberOfItemsInIndexedQueue(&ProcessQueue);
	} else return 0;
}

void swiDequeueProcess(SWI_ID ProcessID) {
	SWI_ID* Item;
	int k = 0;
	int Limit = NumberOfItemsInIndexedQueue(&ProcessQueue);
	while ((ProcessID != swiQueueContents(k)) && (k < Limit)) k++;
	Item = IndexedQueueItem(&ProcessQueue, k ,1);
	deallocate(Item);
}

void swiRemoveFromProcessQueue(int QueueSlot) {
	SWI_ID* Item;
	if ((0 <= QueueSlot) && (QueueSlot < swiQueueSlots())) 
		Item = RemoveFromIndexedQueue(&ProcessQueue, QueueSlot);
	deallocate(Item);
}

int swiQueueSlots() {
	return NumberOfItemsInIndexedQueue(&ProcessQueue);
}

SWI_ID swiQueueContents(int QueueSlot) {
	SWI_ID* Item;
	Item = IndexedQueueItem(&ProcessQueue, QueueSlot, 0);
	return *Item;
}

int swiQueueEmpty() {
	return 0 == NumberOfItemsInIndexedQueue(&ProcessQueue);
}


int swiRunQueuedProcess() {
	SWI_ID* Item;
	UserRequestedExit = 0;
	while (!swiQueueEmpty()) {
		Item = DequeueFromIndexedQueue(&ProcessQueue);
		RunningProcessID = *Item;
		deallocate(Item);
		if (swiPRRun(RunningProcessID)) return 1;
	}
	RunningProcessID = 0;
	return 0;
}

int swiFinishQueuedProcess() {
	int ReturnValue = swiPRFinish(RunningProcessID);
	UserRequestedExit = 0;
	return ReturnValue;
}

void swiHaltCurrentProcess() {
	UserRequestedExit = 1;
}

int swiCurrentProcessCanBeHalted() {
	return ExitRequestHonored();
}

int swiLastReturnValue() {
	return LastExitValue;
}

SWI_ID swiRunningProcess() {
	return RunningProcessID;
}

int swiProcessExitRequested() {
	return UserRequestedExit;
}

void swiKillCurrentProcess() {
	if (ThreadHandle) {
		TerminateThread((HANDLE) ThreadHandle, ~0);
		CloseHandle((HANDLE) ThreadHandle);
		ThreadHandle = 0;
	}
}

/* process type functions */

typedef struct {
	SWI_ID ID;
	SWIDataTypes Type;
	SWI_ID ParentID;
	char* Desc;
} ProcessType;

SWI_ID swiPTNew(SWI_ID ParentTypeID, char* Description) {
	ProcessType* PT;
	if (Description) {
		PT = allocate(sizeof(ProcessType));
		PT->Type = PROCESS_TYPE;
		PT->Desc = allocate(strlen(Description) + 1);
		strcpy(PT->Desc, Description);
		PT->ParentID = ParentTypeID;
		swiSInsert(PT);
		return PT->ID;
	}
	return 0;
}

void swiPTDelete(SWI_ID ProcessTypeID) {
	ProcessType* PT = swiSRemove(ProcessTypeID, PROCESS_TYPE);
	if (PT) {
		deallocate(PT->Desc);
		deallocate(PT);
	}
}

SWI_ID swiPTParent(SWI_ID ProcessTypeID) {
	ProcessType* PT = swiSFind(ProcessTypeID, PROCESS_TYPE);
	if (PT) return PT->ParentID;
	return 0;
}

char* swiPTDescription(SWI_ID ProcessTypeID) {
	ProcessType* PT = swiSFind(ProcessTypeID, PROCESS_TYPE);
	if (PT) return PT->Desc;
	return 0;
}


typedef struct {
	SWI_ID ID;
	SWIDataTypes Type;
	char* Desc;
	SWI_INT_INT_PSWI_ID_PPVOID_INT_PSWI_ID_PPVOID_PPVOID Preprocess;
	SWI_INT_INT_PPVOID_INT_PPVOID_PPVOID MainProcess;
	SWI_INT_INT_PSWI_ID_PPVOID_INT_PSWI_ID_PPVOID_PPVOID Postprocess;
	SWI_ID PS_ID;
	SWI_ID WorkingPS_ID;
	SWI_ID InputReq;
	SWI_ID OutputReq;
	int UsageCount;
	int DeleteWhenUseless;
	SWI_ID ProcType;
	SWI_ID Link;
	int MenuItem;
	int HasAccel, AccelShift, AccelCtrl, AccelAlt;
	int AccelVirtual, AccelCode, AccelCommand;
} ProcessClass;

SWI_ID swiPCNew(
	SWI_ID ProcessTypeID,
	char* Description,
	SWI_INT_INT_PSWI_ID_PPVOID_INT_PSWI_ID_PPVOID_PPVOID PreprocessFunc,
	SWI_INT_INT_PPVOID_INT_PPVOID_PPVOID ProcessFunc,
	SWI_INT_INT_PSWI_ID_PPVOID_INT_PSWI_ID_PPVOID_PPVOID PostprocessFunc,
	SWI_ID ParameterSetID,
	SWI_ID InputRequirements,
	SWI_ID OutputRequirements)
{
	ProcessClass* PC;
	PC = allocate(sizeof(ProcessClass));
	PC->Type = PROCESS_CLASS;
	PC->ProcType = ProcessTypeID;
	if (Description) {
		PC->Desc = allocate(strlen(Description) + 1);
		strcpy(PC->Desc, Description);
	} else PC->Desc = 0;
	PC->Preprocess = PreprocessFunc;
	PC->MainProcess = ProcessFunc;
	PC->Postprocess = PostprocessFunc;
	PC->PS_ID = ParameterSetID;
	if (ParameterSetID) {
		swiPSSetDefaults(ParameterSetID);
		PC->WorkingPS_ID = swiPSCopy(ParameterSetID);
	} else PC->WorkingPS_ID = 0;
	PC->InputReq = InputRequirements;
	PC->OutputReq = OutputRequirements;
	PC->UsageCount = 0;
	PC->DeleteWhenUseless = 0;
	PC->Link = 0;
	PC->MenuItem = PC->HasAccel = PC->AccelShift = PC->AccelCtrl = 
		PC->AccelAlt = PC->AccelVirtual = PC->AccelCode = 
		PC->AccelCommand = 0;
	swiSInsert(PC);
	return PC->ID;
}

void swiPCDelete(SWI_ID ProcessClassID) {
	ProcessClass* PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC) {
		if (PC->UsageCount > 0) PC->DeleteWhenUseless = 1;
		else {
			swiSRemove(ProcessClassID, PROCESS_CLASS);
			if (PC->Desc) deallocate(PC->Desc);
			deallocate(PC);
		}
	}
}

SWI_ID swiPCLinked(SWI_ID ProcessClassID) {
	ProcessClass* PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC) return PC->Link;
	return 0;
}

void swiPCLink(SWI_ID ProcessClassID, SWI_ID LinkTo) {
	ProcessClass* PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC) PC->Link = LinkTo;
}

SWI_ID swiPCProcessType(SWI_ID ProcessClassID) {
	ProcessClass* PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC) return PC->ProcType;
	return 0;
}

SWI_ID swiPCParameterSet(SWI_ID ProcessClassID) {
	ProcessClass* PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC) return PC->PS_ID;
	return 0;
}

SWI_ID swiPCWorkingParameterSet(SWI_ID ProcessClassID) {
	ProcessClass* PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC) return PC->WorkingPS_ID;
	return 0;
}

void swiPCParameterSetToWorking(SWI_ID ProcessClassID) {
	ProcessClass* PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC && PC->PS_ID) swiPSCopyOver(PC->PS_ID, PC->WorkingPS_ID);
}

void swiPCWorkingToParameterSet(SWI_ID ProcessClassID) {
	ProcessClass* PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC && PC->PS_ID) swiPSCopyOver(PC->WorkingPS_ID, PC->PS_ID);
}

char* swiPCDescription(SWI_ID ProcessClassID) {
	ProcessClass* PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC) return PC->Desc;
	return 0;
}

char* swiPCPreconditionDescription(SWI_ID ProcessClassID) {
	char *foo, *bar, *desc;
	ProcessClass* PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC) {
		foo = "Process ";
		bar = AppendString(foo, swiPCDescription(ProcessClassID));
		foo = AppendString(bar, " requires");
		deallocate(bar);
		if (PC->InputReq) {
			desc = swiRQDescription(PC->InputReq);
			bar = AppendString(foo, desc);
			deallocate(desc);
		} else bar = AppendString(foo, "\nnothing");
		deallocate(foo);
		foo = AppendString(bar, "\nselected as input data.");
		deallocate(bar);
		return foo;
	}
	return 0;
}

SWI_ID swiPCInputRequirement(SWI_ID ProcessClassID) {
	ProcessClass* PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC) return PC->InputReq;
	return 0;
}

SWI_ID swiPCOutputRequirement(SWI_ID ProcessClassID) {
	ProcessClass* PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC) return PC->OutputReq;
	return 0;
}

int swiPCOverwriteOutput(SWI_ID ProcessClassID) {
	int k;
	char* foo, *bar;
	ProcessClass* PC;
	if (Overwrite == SWI_OVERWRITE_NEVER) return 0;
	PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC) {
		/* don't overwrite if counts don't match */
		if (!PC->OutputReq 
			|| !swiSDSimilarSelected(SWI_SELECTED_OUTPUT, PC->OutputReq)) 
			return 0;
		if (Overwrite == SWI_OVERWRITE_ALWAYS) return 1;
		/* check from user if overwrite is desired */
		bar = swiSDListNames(SWI_SELECTED_OUTPUT);
		foo = AppendString(
			"Do you wish to overwrite the following:\n", bar);
		deallocate(bar);
		k = swiQuestionMessageBox(foo, 0);
		deallocate(foo);
		return k;
	}
	return 0;
}

void swiPCUse(SWI_ID ProcessClassID) {
	ProcessClass* PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC) PC->UsageCount++;
}

void swiPCUnuse(SWI_ID ProcessClassID) {
	ProcessClass* PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC) {
		PC->UsageCount--;
		if ((PC->UsageCount == 0) && PC->DeleteWhenUseless)
			swiPCDelete(ProcessClassID);
	}
}

static int FreeCommand = SWI_FIRST_PROCESS_MENUITEM;

int swiPCSetMenuInfo(SWI_ID ProcessClassID, int Key, int Special, 
	int Ctrl, int Alt, int Shift, SWI_ID InMenuAfterProcess)
{
	ProcessClass* PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC && Key) {
		PC->AccelShift = Shift;
		PC->AccelCtrl = Ctrl;
		PC->AccelAlt = Alt;
		PC->AccelVirtual = Special;
		PC->AccelCode = Key;
		PC->AccelCommand = FreeCommand++;
		swiWMAdd(PC->ID, InMenuAfterProcess, PC->AccelCommand);
		PC->MenuItem = 1;
		return PC->HasAccel = 
			swiWAAdd(Key, Special, Ctrl, Alt, Shift, PC->AccelCommand);
	} else if (PC) {
		PC->MenuItem = 1;
		PC->AccelCommand = FreeCommand++;
		swiWMAdd(PC->ID, InMenuAfterProcess, PC->AccelCommand);
		return 1;
	}
	return 0;
}

int swiPCIsMenuItem(SWI_ID ProcessClassID) {
	ProcessClass* PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC) return PC->MenuItem;
	return 0;
}

int swiPCHasAccelerator(SWI_ID ProcessClassID) {
	ProcessClass* PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC) return PC->HasAccel;
	return 0;
}

int swiPCUsesCtrl(SWI_ID ProcessClassID) {
	ProcessClass* PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC) return PC->AccelCtrl;
	return 0;
}

int swiPCUsesAlt(SWI_ID ProcessClassID) {
	ProcessClass* PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC) return PC->AccelAlt;
	return 0;
}

int swiPCUsesShift(SWI_ID ProcessClassID) {
	ProcessClass* PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC) return PC->AccelShift;
	return 0;
}

int swiPCUsesVirtual(SWI_ID ProcessClassID) {
	ProcessClass* PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC) return PC->AccelVirtual;
	return 0;
}

int swiPCUsesKey(SWI_ID ProcessClassID) {
	ProcessClass* PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC) return PC->AccelCode;
	return 0;
}

int swiPCUsesCommand(SWI_ID ProcessClassID) {
	ProcessClass* PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC) return PC->AccelCommand;
	return 0;
}


typedef struct {
	SWI_ID ID;
	SWIDataTypes Type;
	int ExitValue;
	SWI_ID ClassID;
	char* Desc;
	SWI_INT_INT_PSWI_ID_PPVOID_INT_PSWI_ID_PPVOID_PPVOID Preprocess;
	SWI_INT_INT_PPVOID_INT_PPVOID_PPVOID MainProcess;
	SWI_INT_INT_PSWI_ID_PPVOID_INT_PSWI_ID_PPVOID_PPVOID Postprocess;
	SWI_ID PS_ID;
	int PDCount;
	SWI_ID* PD;
	void** Externals;
	int ODCount;
	SWI_ID* OD;
	void** ODExternals;
	void* Aux;
	int Overwrite;
} Process;

SWI_ID swiPRNew(SWI_ID ProcessClassID, int OutputOverwritesExisting) {
	int k;
	Process* P;
	ProcessClass* PC = swiSFind(ProcessClassID, PROCESS_CLASS);
	if (PC) {
		P = allocate(sizeof(Process));
		P->Type = PROCESS;
		P->ClassID = ProcessClassID;
		swiPCUse(P->ClassID);
		P->ExitValue = 0;
		if (PC->Desc) {
			P->Desc = allocate(strlen(PC->Desc) + 1);
			strcpy(P->Desc, PC->Desc);
		} else P->Desc = 0;
		P->Preprocess = PC->Preprocess;
		P->MainProcess = PC->MainProcess;
		P->Postprocess = PC->Postprocess;
		if (PC->PS_ID) P->PS_ID = swiPSCopy(PC->PS_ID);
		else P->PS_ID = 0;
		if (PC->InputReq) P->PDCount = swiRQList(PC->InputReq, 0);
		else P->PDCount = 0;
		if (P->PDCount) {
			P->PD = allocate(P->PDCount * sizeof(SWI_ID));
			P->Externals = allocate(P->PDCount * sizeof(void*));
			swiSDAllSets(SWI_SELECTED_INPUT, 
				PC->InputReq, P->PDCount, P->PD);
			for (k = 0; k < P->PDCount; k++) {
				P->Externals[k] = swiDSExternal(P->PD[k]);
				swiDSUse(P->PD[k]);
			}
		} else {
			P->PD = 0;
			P->Externals = 0;
		}
		P->Overwrite = OutputOverwritesExisting;
		if (PC->OutputReq) P->ODCount = swiRQList(PC->OutputReq, 0);
		else P->ODCount = 0;
		if (P->ODCount) {
			P->OD = allocate(P->ODCount * sizeof(SWI_ID));
			P->ODExternals = allocate(P->ODCount * sizeof(void*));
			/* test if overwrite fulfilled and if so, fill, otherwise zero */
			/*if (P->Overwrite 
				&& swiSDSets(SWI_SELECTED_OUTPUT, PC->OutputReq, P->ODCount, P->OD))
			{
				for (k = 0; k < P->ODCount; k++) {
					P->ODExternals[k] = swiDSExternal(P->OD[k]);
					swiDSUse(P->OD[k]);
				}
			} else */
			for (k = 0; k < P->ODCount; k++) {
				P->OD[k] = 0;
				P->ODExternals[k] = 0;
			}
		} else {
			P->OD = 0;
			P->ODExternals = 0;
		}
		P->Aux = 0;
		swiSInsert(P);
		return P->ID;
	}
	return 0;
}

void swiPRDelete(SWI_ID ProcessID) {
	int k;
	Process* P = swiSRemove(ProcessID, PROCESS);
	if (P) {
		swiPCUnuse(P->ClassID);
		if (P->Desc) deallocate(P->Desc);
		if (P->PDCount) {
			for (k = 0; k < P->PDCount; k++) swiDSUnuse(P->PD[k]);
			deallocate(P->PD);
			deallocate(P->Externals);
		}
		if (P->ODCount) {
			deallocate(P->OD);
			deallocate(P->ODExternals);
		}
		deallocate(P);
	}
}

static Process* Pr;
static SWI_INT_INT_PPVOID_INT_PPVOID_PPVOID PrFunc = 0;

void __cdecl RunProcessHelper(void* unused) {
	Process* P = Pr;
	P->ExitValue = (*PrFunc)
		(P->PDCount, P->Externals, P->ODCount, P->ODExternals, &(P->Aux));
	PostMessage(swiMainWindowHandle(), 
		SWI_WM_THREAD_FINISHED, P->ExitValue, 0);
	_endthread();
}

int swiPRRun(SWI_ID ProcessID) {
	char* foo;
	Process* P = swiSFind(ProcessID, PROCESS);
	if (P) {
		if (P->PS_ID) swiPSModifyFullSet(P->PS_ID);
		foo = swiPCDescription(P->ClassID);
		if (foo) PrintMessage("Begin process %s\n", foo);
		else PrintMessage("Begin process <unnamed>\n");
		/* preprocess failure? */
		if (P->Preprocess 
			&& (*(P->Preprocess))(
				P->PDCount, P->PD, P->Externals, 
				P->ODCount, P->OD, P->ODExternals, &(P->Aux)))
		{
			if (foo) PrintMessage("\nFinished process %s\n", foo);
			else PrintMessage("\nFinished process <unnamed>\n");
			if (P->Postprocess) (*(P->Postprocess))(
				P->PDCount, P->PD, P->Externals, P->ODCount, 
				P->OD, P->ODExternals, &(P->Aux));
			return 0;
		}
		P->ExitValue = 0;
		Pr = P;
		PrFunc = P->MainProcess;
		ThreadHandle = _beginthread(&RunProcessHelper, 8192, 0);
		if (ThreadHandle == ~0) {
			if (P->Postprocess) (*(P->Postprocess))(
				P->PDCount, P->PD, P->Externals, P->ODCount, 
				P->OD, P->ODExternals, &(P->Aux));
			swiErrorMessageBox("Failed to generate thread of execution.");
			if (foo) PrintMessage("\nFinished process %s\n", foo);
			else PrintMessage("\nFinished process <unnamed>\n");
			ThreadHandle = 0;
			return 0;
		}
		return 1;
	}
	return 0;
}

int swiPRFinish(SWI_ID ProcessID) {
	char* foo;
	int ReturnValue, k;
	Process* P = swiSFind(ProcessID, PROCESS);
	SetExitRequestHonored(0);
	if (P->Postprocess) (*(P->Postprocess))(P->PDCount, P->PD, 
		P->Externals, P->ODCount, P->OD, P->ODExternals, &(P->Aux));
	foo = swiPCDescription(P->ClassID);
	if (foo) PrintMessage("\nFinished process %s\n", foo);
	else PrintMessage("\nFinished process <unnamed>\n");
	for (k = 0; k < P->ODCount; k++) if (P->OD[k])
		PostMessage(swiMainWindowHandle(), 
			SWI_WM_NEW_DATASET, 1, P->OD[k]);
	ReturnValue = P->ExitValue;
	swiPRDelete(ProcessID);
	return ReturnValue;
}

int swiPRRunning(SWI_ID ProcessID) {
	return RunningProcessID == ProcessID;
}

int swiPRDataSetCount(SWI_ID ProcessID) {
	Process* P = swiSFind(ProcessID, PROCESS);
	if (P) return P->PDCount;
	return 0;
}

SWI_ID swiPRDataSet(SWI_ID ProcessID, int DataSetIndex) {
	Process* P = swiSFind(ProcessID, PROCESS);
	if (P && (DataSetIndex >= 0) && (P->PDCount > DataSetIndex))
		return P->PD[DataSetIndex];
	return 0;
}

SWI_ID swiPRParameterSet(SWI_ID ProcessID) {
	Process* P = swiSFind(ProcessID, PROCESS);
	if (P) return P->PS_ID;
	return 0;
}

SWI_ID swiPRClass(SWI_ID ProcessID) {
	Process* P = swiSFind(ProcessID, PROCESS);
	if (P) return P->ClassID;
	return 0;
}

void swiPRExitValue(SWI_ID ProcessID, int ReturnValue) {
	Process* P = swiSFind(ProcessID, PROCESS);
	if (P) P->ExitValue = ReturnValue;
}
