/********************************************************************

	swi_ProcessAndOptions.c

	Ismo Krkkinen

	Last changed 12.11.1999

********************************************************************/


#include "swi_ProcessAndOptions.h"
#include "swi_ProcessAndOptionsSymbols.h"
#include <windows.h>
#include "swi_WindowsControls.h"
#include "swi_Boolean.h"
#include "swi_Enumerated.h"
#include "swi_Integer.h"
#include "memctrl.h"
#include <string.h>
#include "idxqueue.h"
#include "swi_OwnedWindow.h"
#include "swi_SelectedData.h"
#include "swi_Parameters.h"
#include "swi_Process.h"
#include "swi_Storage.h"


static HWND ProcessAndOptionsDialogHandle = 0;
static SWI_ID OWID = 0;

static void SetParameterList(HWND Handle, SWI_ID ProcessClassID) {
	int k;
	char* foo, *bar;
	SWI_ID ParameterSetID;
	HWND ListBox = GetDlgItem(Handle, SWI_PROCESS_AND_OPTIONS_LISTBOX);
	swiLBRemoveAll(ListBox);
	ParameterSetID = swiPCWorkingParameterSet(ProcessClassID);
	if (ParameterSetID) {
		for (k = swiFPSGetFirst(); k <= swiFPSGetLast(); k++) {
			if (swiPSIsMember(ParameterSetID, k)) {
				bar = AppendString(swiFPSGetDocumentString(k), "\t");
				foo = AppendString(bar, 
					swiFPSValueAsString(k, swiPSValue(ParameterSetID, k)));
				deallocate(bar);
				swiLBAdd(ListBox, foo, k);
				deallocate(foo);
			}
		}
	}
}

static void SetValueDisplay(HWND Handle, int Option, int Value) {
	char *foo, *bar;
	HWND Label = GetDlgItem(Handle, SWI_PROCESS_AND_OPTIONS_VALUE);
	if (swiFPSGetType(Option) == 1) {
		foo = AppendString("Min: ", 
			swiFPSValueAsString(Option, swiFPSGetMinValue(Option)));
		bar = AppendString(foo, ", Curr: ");
		deallocate(foo);
		foo = AppendString(bar, swiFPSValueAsString(Option, Value));
		deallocate(bar);
		bar = AppendString(foo, ", Max: ");
		deallocate(foo);
		foo = AppendString(bar, 
			swiFPSValueAsString(Option, swiFPSGetMaxValue(Option)));
		deallocate(bar);
		SetWindowText(Label, foo);
		deallocate(foo);
	} else SetWindowText(Label, swiFPSValueAsString(Option, Value));
}

BOOL PAOHandleWMCommand(int ID, int Message, HWND Handle, LPARAM lParam) {
	HWND h;
	int Selected, Result;
	char* Description;
	SWI_ID ProcessClassID, ParameterSetID, ProcessID;
	switch (ID) {
	case SWI_PROCESS_AND_OPTIONS_DROPDOWNLIST:
		if (Message == CBN_SELCHANGE) {
			Selected = swiCBSelectedIndex((HWND)lParam);
			ProcessClassID = swiCBTag((HWND)lParam, Selected);
			SetParameterList(Handle, ProcessClassID);
			return TRUE;
		}
		break;
	case SWI_PROCESS_AND_OPTIONS_LISTBOX:
		if (Message == LBN_DBLCLK) {
			h = GetDlgItem(Handle, SWI_PROCESS_AND_OPTIONS_DROPDOWNLIST);
			Selected = swiCBSelectedIndex(h);
			ProcessClassID = swiCBTag(h, Selected);
			ParameterSetID = swiPCWorkingParameterSet(ProcessClassID);
			Selected = 
				swiLBTag((HWND)lParam, swiLBFocusIndex((HWND)lParam));
			Result = swiChangeOption(ParameterSetID, Selected, Handle);
			swiPSSetValue(ParameterSetID, Selected, Result);
			SetValueDisplay(Handle, Selected, Result);
			SetParameterList(Handle, ProcessClassID);
			return TRUE;
		} else if (Message == LBN_SELCHANGE) {
			h = GetDlgItem(Handle, SWI_PROCESS_AND_OPTIONS_DROPDOWNLIST);
			Selected = swiCBSelectedIndex(h);
			ProcessClassID = swiCBTag(h, Selected);
			ParameterSetID = swiPCWorkingParameterSet(ProcessClassID);
			Selected = 
				swiLBTag((HWND)lParam, swiLBFocusIndex((HWND)lParam));
			SetValueDisplay(Handle, Selected,
				swiPSValue(ParameterSetID, Selected));
			return TRUE;
		}
		break;
	case SWI_PROCESS_AND_OPTIONS_START:
		if (Message == BN_CLICKED) {
			h = GetDlgItem(Handle, SWI_PROCESS_AND_OPTIONS_DROPDOWNLIST);
			ProcessClassID = swiCBTag(h, swiCBSelectedIndex(h));
			if (swiSDSimilarContents(SWI_SELECTED_INPUT, 
				swiPCInputRequirement(ProcessClassID)))
			{
				swiPCWorkingToParameterSet(ProcessClassID);
				Result = swiPCOverwriteOutput(ProcessClassID);
				ProcessID = swiPRNew(ProcessClassID, Result);
				swiEnqueueProcess(ProcessID);
				PostMessage(swiMainWindowHandle(), 
					SWI_WM_CHECK_NEW_PROCESSES, 0, 0);
				PostMessage(ProcessAndOptionsDialogHandle, WM_CLOSE, 0, 0);
			} else {
				Description = swiPCPreconditionDescription(ProcessClassID);
				swiInformationMessageBox(Description);
				deallocate(Description);
			}
			return TRUE;
		}
		break;
	case SWI_PROCESS_AND_OPTIONS_CHANGE:
		if (Message == BN_CLICKED) {
			h = GetDlgItem(Handle, SWI_PROCESS_AND_OPTIONS_DROPDOWNLIST);
			Selected = swiCBSelectedIndex(h);
			ProcessClassID = swiCBTag(h, Selected);
			ParameterSetID = swiPCWorkingParameterSet(ProcessClassID);
			h = GetDlgItem(Handle, SWI_PROCESS_AND_OPTIONS_LISTBOX);
			Selected = swiLBTag(h, swiLBFocusIndex(h));
			Result = swiChangeOption(
				ParameterSetID, Selected, GetWindow(Handle, GW_OWNER));
			swiPSSetValue(ParameterSetID, Selected, Result);
			SetValueDisplay(Handle, Selected, Result);
			SetParameterList(Handle, ProcessClassID);
			return TRUE;
		}
		break;
	case SWI_PROCESS_AND_OPTIONS_DEFAULTS:
		if (Message == BN_CLICKED) {
			h = GetDlgItem(Handle, SWI_PROCESS_AND_OPTIONS_DROPDOWNLIST);
			Selected = swiCBSelectedIndex(h);
			ProcessClassID = swiCBTag(h, Selected);
			ParameterSetID = swiPCWorkingParameterSet(ProcessClassID);
			swiPSSetDefaults(ParameterSetID);
			SetParameterList(Handle, ProcessClassID);
			return TRUE;
		}
		break;
	case IDCANCEL:
		if (Message == BN_CLICKED) {
			SendMessage(ProcessAndOptionsDialogHandle, WM_CLOSE, 0, 0);
			return TRUE;
		}
		break;
	}
	return FALSE;
}


static INDEXEDQUEUE Processes;

int SeekProcesses(void* Data) {
	SWI_ID* FoundProcess;
	if (((SWIDataTypes*)Data)[1] == PROCESS_CLASS) {
		FoundProcess = allocate(sizeof(SWI_ID));
		*FoundProcess = *((SWI_ID*)Data);
		EnqueueIntoIndexedQueue(&Processes, FoundProcess);
	}
	return 0;
}

static int FindProcesses(HWND Handle) {
	HWND h;
	SWI_ID* ID;
	h = GetDlgItem(Handle, SWI_PROCESS_AND_OPTIONS_DROPDOWNLIST);
	EnableWindow(h, TRUE);
	swiCBRemoveAll(h);
	InitIndexedQueue(&Processes);
	swiSIterateWithCallback(&SeekProcesses);
	ID = DequeueFromIndexedQueue(&Processes);
	while (ID) {
		swiCBAdd(h, swiPCDescription(*ID), *ID);
		deallocate(ID);
		ID = DequeueFromIndexedQueue(&Processes);
	}
	DeleteIndexedQueue(&Processes);
	if (swiCBCount(h) > 0) {
		swiCBSetSelectedIndex(h, 0);
		h = GetDlgItem(Handle, SWI_PROCESS_AND_OPTIONS_LISTBOX);
		EnableWindow(h, TRUE);
		h = GetDlgItem(Handle, SWI_PROCESS_AND_OPTIONS_CHANGE);
		EnableWindow(h, TRUE);
		h = GetDlgItem(Handle, SWI_PROCESS_AND_OPTIONS_DEFAULTS);
		EnableWindow(h, TRUE);
		h = GetDlgItem(Handle, SWI_PROCESS_AND_OPTIONS_START);
		EnableWindow(h, TRUE);
		return 1;
	} else {
		h = GetDlgItem(Handle, SWI_PROCESS_AND_OPTIONS_LISTBOX);
		EnableWindow(h, FALSE);
		h = GetDlgItem(Handle, SWI_PROCESS_AND_OPTIONS_CHANGE);
		EnableWindow(h, FALSE);
		h = GetDlgItem(Handle, SWI_PROCESS_AND_OPTIONS_DEFAULTS);
		EnableWindow(h, FALSE);
		h = GetDlgItem(Handle, SWI_PROCESS_AND_OPTIONS_START);
		EnableWindow(h, FALSE);
		h = GetDlgItem(Handle, SWI_PROCESS_AND_OPTIONS_DROPDOWNLIST);
		EnableWindow(h, FALSE);
	}
	return 0;
}



BOOL APIENTRY ProcessAndOptionsDialogProcedure(HWND Handle, UINT Message,
	WPARAM wParam, LPARAM lParam)
{
	HWND h;
	int k;
	SWI_ID ProcessClassID;
	switch (Message) {
	case WM_INITDIALOG:
		ProcessAndOptionsDialogHandle = Handle;
		PostMessage(Handle, WM_SETICON, ICON_BIG, (LPARAM) swiAppIcon());
		OWID = swiOWNew(Handle);
		if (FindProcesses(Handle)) {
			h = GetDlgItem(Handle, SWI_PROCESS_AND_OPTIONS_DROPDOWNLIST);
			k = swiCBSelectedIndex(h);
			ProcessClassID = swiCBTag(h, k);
			SetParameterList(Handle, ProcessClassID);
		}
		k = 125;
		SendMessage(GetDlgItem(Handle, SWI_PROCESS_AND_OPTIONS_LISTBOX), 
			LB_SETTABSTOPS, 1, (LPARAM) &k);
		return TRUE;
	case WM_COMMAND:
		return PAOHandleWMCommand(
			LOWORD(wParam), HIWORD(wParam), Handle, lParam);
	case WM_SHOWWINDOW:
		if (wParam) swiOWSelfShow(OWID);
		else swiOWSelfHide(OWID);
		return 0;
	case WM_CLOSE:
		h = GetDlgItem(Handle, SWI_PROCESS_AND_OPTIONS_DROPDOWNLIST);
		for (k = 0; k < swiCBCount(h); k++) {
			ProcessClassID = swiCBTag(h, k);
			swiPCWorkingToParameterSet(ProcessClassID);
		}
		DestroyWindow(Handle);
		return TRUE;
	case WM_DESTROY:
		swiOWDelete(OWID);
		OWID = 0;
		ProcessAndOptionsDialogHandle = 0;
		return 0;
	case SWI_WM_SELECT_PROCESS_CLASS:
		h = GetDlgItem(Handle, SWI_PROCESS_AND_OPTIONS_DROPDOWNLIST);
		for (k = 0; k < swiCBCount(h); k++) if (swiCBTag(h, k) == wParam) {
			swiCBSetSelectedIndex(h, k);
			SendMessage(Handle, WM_COMMAND, 
				MAKEWPARAM(
					SWI_PROCESS_AND_OPTIONS_DROPDOWNLIST, CBN_SELCHANGE),
				(LONG)h);
			return 0;
		}
		return 0;
	}
	return FALSE;
}

void swiCreateProcessAndOptionsDialog() {
	if (!ProcessAndOptionsDialogHandle)
		CreateDialog(GetModuleHandle(0), "swiDialogProcessAndOptions",
			swiParentWindowHandle(), &ProcessAndOptionsDialogProcedure);
	else SetForegroundWindow(ProcessAndOptionsDialogHandle);
}

void swiDestroyProcessAndOptionsDialog() {
	if (ProcessAndOptionsDialogHandle) {
		SendMessage(ProcessAndOptionsDialogHandle, WM_CLOSE, 0, 0);
	}
}

void swiHideProcessAndOptionsDialog() {
	if (ProcessAndOptionsDialogHandle)
		ShowWindow(ProcessAndOptionsDialogHandle, SW_HIDE);
}

void swiShowProcessAndOptionsDialog() {
	if (ProcessAndOptionsDialogHandle)
		ShowWindow(ProcessAndOptionsDialogHandle, SW_SHOW);
}

HWND swiProcessAndOptionsDialogHandle() {
	return ProcessAndOptionsDialogHandle;
}

void swiSelectProcessAndOptions(SWI_ID ProcessClassID) {
	if (ProcessAndOptionsDialogHandle) SendMessage(
		ProcessAndOptionsDialogHandle, SWI_WM_SELECT_PROCESS_CLASS,
		ProcessClassID, 0);
}