/********************************************************************

	swi_ProcessQueue.c

	Ismo Krkkinen

	Last changed 12.11.1999

********************************************************************/


#include "swi_Internal.h"
#include "swi_ProcessQueue.h"
#include "resource.h"
#include "swi_ProcessAndOptions.h"
#include "swi_ProcessAndOptionsSymbols.h"
#include <windows.h>
#include "swi_WindowsControls.h"
#include "memctrl.h"
#include <string.h>
#include "idxqueue.h"
#include "swi_Process.h"
#include "swi_Parameters.h"
#include "swi_OwnedWindow.h"
#include "swi_DataClass.h"


static int FillProcessList(HWND Handle) {
	int k, Added;
	SWI_ID ProcessID;
	char* foo;
	HWND h = GetDlgItem(Handle, SWI_PROCESS_QUEUE_LIST);
	swiLBRemoveAll(h);
	ProcessID = swiRunningProcess();
	if (ProcessID) {
		foo = AppendString(swiPCDescription(swiPRClass(ProcessID)), 
			" (running)");
		swiLBAdd(h, foo, ProcessID);
		deallocate(foo);
		Added = 1;
	} else Added = 0;
	if (!swiQueueEmpty()) for (k = 0; k < swiQueueSlots(); k++) {
		ProcessID = swiQueueContents(k);
		swiLBAdd(h, swiPCDescription(swiPRClass(ProcessID)), ProcessID);
		Added++;
	}
	return Added;
}

static int FillProcessInfo(HWND Handle, SWI_ID ProcessID) {
	int k;
	char* foo, *bar;
	char* Text;
	SWI_ID Parameters, DataSet;
	HWND h = GetDlgItem(Handle, SWI_PROCESS_QUEUE_INFO);
	swiLBRemoveAll(h);
	if (ProcessID) {
		swiLBAdd(h, swiPCDescription(swiPRClass(ProcessID)), 0);
		Parameters = swiPRParameterSet(ProcessID);
		if (Parameters) {
			swiLBAdd(h, "Parameters:", 0);
			for (k = swiFPSGetFirst(); k <= swiFPSGetLast(); k++) 
				if (swiPSIsMember(Parameters, k))
			{
				Text = swiFPSValueAsString(k, swiPSValue(Parameters, k));
				if (Text) {
					bar = AppendString(swiFPSGetDocumentString(k), " = ");
					foo = AppendString(bar, Text);
					deallocate(bar);
					swiLBAdd(h, foo, ProcessID);
					deallocate(foo);
				}
			}
		} else swiLBAdd(h, "No parameters.", 0);
		if (swiPRDataSetCount(ProcessID)) {
			swiLBAdd(h, "Input data:", 0);
			for (k = 0; k < swiPRDataSetCount(ProcessID); k++) {
				DataSet = swiPRDataSet(ProcessID, k);
				swiLBAdd(h, swiDSName(DataSet), DataSet);
			}
		}
		return 1;
	} 
	return 0;
}

static int LastFocus = -1;

BOOL PQHandleWMCommand(int ID, int Message, HWND Handle, LPARAM lParam) {
	int *Indices, k, Count;
	HWND h;
	switch (ID) {
	case SWI_PROCESS_QUEUE_NEW:
		swiCreateProcessAndOptionsDialog();
		return TRUE;
	case SWI_PROCESS_QUEUE_REMOVE:
		h = GetDlgItem(Handle, SWI_PROCESS_QUEUE_LIST);
		Indices = 0;
		Count = swiLBSelectedIndices(h, &Indices);
		if (Count) {
			for (k = Count - 1; k >= 0; k--) 
				if ((k == 0) && (Indices[0] == 0))
			{
				/* stop running process */
				if (swiCurrentProcessCanBeHalted())
					swiHaltCurrentProcess();
				else swiInformationMessageBox(
					"Running process cannot be halted, "
					"exit program to terminate it.");
			} else {
				swiRemoveFromProcessQueue(Indices[k] - 1);
				swiLBRemove(h, Indices[k]);
			}
			deallocate(Indices);
		}
		return TRUE;
	case SWI_PROCESS_QUEUE_CLOSE:
		SendMessage(Handle, WM_CLOSE, 0, 0);
		return TRUE;
	case SWI_PROCESS_QUEUE_LIST:
		if (Message != LBN_KILLFOCUS) {
			h = GetDlgItem(Handle, SWI_PROCESS_QUEUE_LIST);
			k = swiLBFocusIndex(h);
			if (k != LastFocus) {
				LastFocus = k;
				FillProcessInfo(Handle, swiLBTag(h, k));
			}
		}
		return TRUE;
	}
	return FALSE;
}

static HWND ProcessQueueDialogHandle = 0;
static SWI_ID OWID = 0;

BOOL APIENTRY ProcessQueueDialogProcedure(HWND Handle, UINT Message,
	WPARAM wParam, LPARAM lParam)
{
	HWND h;
	switch (Message) {
	case WM_INITDIALOG:
		OWID = swiOWNew(Handle);
		PostMessage(Handle, SWI_WM_UPDATE_PROCESS_LISTS, 0, 0);
		return TRUE;
	case WM_COMMAND:
		return PQHandleWMCommand(
			LOWORD(wParam), HIWORD(wParam), Handle, lParam);
	case WM_SHOWWINDOW:
		if (wParam) swiOWSelfShow(OWID);
		else swiOWSelfHide(OWID);
		return 0;
	case WM_CLOSE:
		DestroyWindow(Handle);
		return TRUE;
	case WM_DESTROY:
		swiOWDelete(OWID);
		ProcessQueueDialogHandle = 0;
		return 0;
	case SWI_WM_UPDATE_PROCESS_LISTS:
		if (FillProcessList(Handle)) {
			h = GetDlgItem(Handle, SWI_PROCESS_QUEUE_LIST);
			EnableWindow(
				GetDlgItem(Handle, SWI_PROCESS_QUEUE_REMOVE), TRUE);
			swiLBSetFocusIndex(h, 0);
			if (FillProcessInfo(Handle, swiLBTag(h, 0)))
				EnableWindow(
					GetDlgItem(Handle, SWI_PROCESS_QUEUE_INFO), TRUE);
			else EnableWindow(
				GetDlgItem(Handle, SWI_PROCESS_QUEUE_INFO), FALSE);
		} else {
			EnableWindow(
				GetDlgItem(Handle, SWI_PROCESS_QUEUE_REMOVE), FALSE);
			EnableWindow(
				GetDlgItem(Handle, SWI_PROCESS_QUEUE_INFO), FALSE);
		}
		return 0;
	}
	return FALSE;
}

void swiCreateProcessQueueDialog() {
	if (!ProcessQueueDialogHandle) ProcessQueueDialogHandle =
		CreateDialog(GetModuleHandle(0), "swiDialogProcessQueue",
			swiParentWindowHandle(), &ProcessQueueDialogProcedure);
	else SetForegroundWindow(ProcessQueueDialogHandle);
}

void swiDestroyProcessQueueDialog() {
	if (ProcessQueueDialogHandle) {
		SendMessage(ProcessQueueDialogHandle, WM_CLOSE, 0, 0);
	}
}

void swiHideProcessQueueDialog() {
	if (ProcessQueueDialogHandle)
		ShowWindow(ProcessQueueDialogHandle, SW_HIDE);
}

void swiShowProcessQueueDialog() {
	if (ProcessQueueDialogHandle)
		ShowWindow(ProcessQueueDialogHandle, SW_SHOW);
}

HWND swiProcessQueueDialogHandle() {
	return ProcessQueueDialogHandle;
}

void swiProcessQueueUpdate() {
	if (ProcessQueueDialogHandle) SendMessage(
		ProcessQueueDialogHandle, SWI_WM_UPDATE_PROCESS_LISTS, 0, 0);
}
