/********************************************************************

	swi_Requirement.c

	Ismo Krkkinen

	Last changed 12.11.1999

********************************************************************/

#include "swi_Requirement.h"
#include "swi_Storage.h"
#include "swi_Internal.h"
#include "memctrl.h"
#include <stdio.h>
#include "swi_DataClass.h"

typedef struct {
	SWI_ID Class;
	int Count;
} ClassAndCount;

typedef struct {
	SWI_ID ID;
	SWIDataTypes Type;
	int Count;
	ClassAndCount* CAC;
} Requirement;


SWI_ID swiRQNew() {
	Requirement* R = allocate(sizeof(Requirement));
	R->Type = REQUIREMENT;
	R->Count = 0;
	R->CAC = 0;
	swiSInsert(R);
	return R->ID;
}

void swiRQAdd(SWI_ID RequirementID, SWI_ID DataClassID, int Count) {
	int k;
	ClassAndCount* NewCAC;
	Requirement* R = swiSFind(RequirementID, REQUIREMENT);
	NewCAC = allocate((R->Count + 1) * sizeof(ClassAndCount));
	for (k = 0; k < R->Count; k++) {
		NewCAC[k].Class = R->CAC[k].Class;
		NewCAC[k].Count = R->CAC[k].Count;
	}
	deallocate(R->CAC);
	R->CAC = NewCAC;
	R->CAC[R->Count].Class = DataClassID;
	R->CAC[R->Count].Count = Count;
	R->Count++;
}

int swiRQList(SWI_ID RequirementID, SWI_ID** DataClassIDList) {
	int Count, k, n;
	Requirement* R = swiSFind(RequirementID, REQUIREMENT);
	Count = 0;
	for (k = 0; k < R->Count; k++) Count += R->CAC[k].Count;
	if (DataClassIDList) {
		*DataClassIDList = allocate(Count * sizeof(SWI_ID));
		Count = 0;
		for (k = 0; k < R->Count; k++) for (n = 0; n < R->CAC[k].Count; n++)
			(*DataClassIDList)[Count++] = R->CAC[k].Class;
	}
	return Count;
}

int swiRQFulfills(SWI_ID RequirementID, int Count, SWI_ID* List) {
	int ClassCount = swiRQList(RequirementID, 0);
	if (ClassCount != Count) return 0;
	return swiRQOrder(RequirementID, Count, List);
}

char* swiRQDescription(SWI_ID RequirementID) {
	char Buffer[40];
	char* foo, *bar;
	int k;
	Requirement* R = swiSFind(RequirementID, REQUIREMENT);
	foo = 0;
	for (k = 0; k < R->Count; k++) {
		if (R->CAC[k].Count > 1) {
			sprintf(Buffer, "\n%i items of type ");
		} else strcpy(Buffer, "\n1 item of type ");
		bar = AppendString(foo, Buffer);
		deallocate(foo);
		foo = AppendString(bar, swiDCDescription(R->CAC[k].Class));
		deallocate(bar);
	}
	return foo;
}

int swiRQOrder(SWI_ID RequirementID, int Count, SWI_ID* IDs) {
	int k, n;
	SWI_ID* ClassList, *OrderedIDs;
	int ClassCount = swiRQList(RequirementID, &ClassList);
	if (ClassCount != Count) {
		deallocate(ClassList);
		return 0;
	}
	OrderedIDs = allocate(Count * sizeof(SWI_ID));
	for (k = 0; k < Count; k++) {
		OrderedIDs[k] = 0;
		n = 0;
		while (n < Count) {
			if (IDs[n]) {
				if (swiDSGetClassID(IDs[n]) == ClassList[k]) {
					OrderedIDs[k] = IDs[n];
					IDs[n] = 0;
					n = Count;
				}
			}
			n++;
		}
		if (!OrderedIDs[k]) {
			deallocate(ClassList);
			deallocate(OrderedIDs);
			return 0;
		}
	}
	for (k = 0; k < Count; k++) IDs[k] = OrderedIDs[k];
	deallocate(ClassList);
	deallocate(OrderedIDs);
	return 1;
}
