/********************************************************************

	swi_SelectedData.c

	Ismo Krkkinen

	Last changed 12.11.1999

********************************************************************/

#include "swi_SelectedData.h"
#include <windows.h>
#include <commctrl.h>
#include "memctrl.h"
#include "swi_Requirement.h"
#include "swi_DataClass.h"

HWND ListViews[SWI_SELECTED_LAST_DONT_USE];

void swiSDInitialize() {
	int k;
	for (k = 0; k < SWI_SELECTED_LAST_DONT_USE; k++) ListViews[k] = 0;
}

void swiSDSetSource(int SelectionSource, HWND Window) {
	ListViews[SelectionSource] = Window;
}

HWND swiSDGetSource(int SelectionSource) {
	return ListViews[SelectionSource];
}

static void GetSelected(HWND ListView, int Count, SWI_ID* IDs) {
	int k;
	LVITEM Item;
	Item.iItem = -1;
	Item.mask = LVIF_PARAM;
	for (k = 0; k < Count; k++) {
		Item.iItem = 
			ListView_GetNextItem(ListView, Item.iItem, LVNI_SELECTED);
		ListView_GetItem(ListView, &Item);
		IDs[k] = Item.lParam;
	}
}

int swiSDSets(int Source, SWI_ID RequirementID, int Count, SWI_ID* Table) {
	HWND View = ListViews[Source];
	if (Count != (int)(ListView_GetSelectedCount(View))) return 0;
	GetSelected(View, Count, Table);
	if (RequirementID) return swiRQOrder(RequirementID, Count, Table);
	return 1;
}

int swiSDSimilarSelected(int Source, SWI_ID RequirementID) {
	int Count, Ok;
	SWI_ID* IDs;
	HWND View = ListViews[Source];
	Count = ListView_GetSelectedCount(View);
	if (!RequirementID) return Count == 0;
	if (!Count) return 0;
	IDs = allocate(Count * sizeof(SWI_ID));
	Ok = swiSDSets(Source, RequirementID, Count, IDs);
	deallocate(IDs);
	return Ok;
}

char* swiSDListNames(int Source) {
	char* foo, *bar;
	int Count, k;
	SWI_ID* IDs;
	HWND View = ListViews[Source];
	Count = swiSDSelectedCount(Source);
	if (Count) {
		IDs = allocate(Count * sizeof(SWI_ID));
		GetSelected(View, Count, IDs);
		foo = 0;
		for (k = 0; k < Count; k++) {
			bar = AppendString(foo, swiDSName(IDs[k]));
			deallocate(foo);
			foo = AppendString(bar, "\n");
			deallocate(bar);
		}
		deallocate(IDs);
	} else foo = 0;
	return foo;
}

int swiSDRemoveSets(int Source, SWI_ID** pIDs) {
	int k;
	LVITEM Item;
	HWND View;
	int Count;
	SWI_ID* IDs;
	int* Items;
	View = ListViews[Source];
	Count = swiSDSelectedCount(Source);
	if (!Count) return 0;
	IDs = *pIDs = allocate(Count * sizeof(SWI_ID));
	Items = allocate(Count * sizeof(int));
	Item.iItem = -1;
	Item.mask = LVIF_PARAM;
	for (k = 0; k < Count; k++) {
		Item.iItem = ListView_GetNextItem(View, Item.iItem, LVNI_SELECTED);
		ListView_GetItem(View, &Item);
		IDs[k] = Item.lParam;
		Items[k] = Item.iItem;
		/* clear IDs so that main window doesn't delete the sets */
		Item.lParam = 0;
		ListView_SetItem(View, &Item);
	}
	for (k = Count - 1; k >= 0; k--) ListView_DeleteItem(View, Items[k]);
	return Count;
}

int swiSDSelectedCount(int Source) {
	return ListView_GetSelectedCount(ListViews[Source]);
}

int swiSDCount(int Source) {
	return ListView_GetItemCount(ListViews[Source]);
}

void swiSDMoveSelected(int Source, int XOffset, int YOffset) {
	POINT Loc;
	int Index;
	HWND View = ListViews[Source];
	Index = ListView_GetNextItem(View, -1, LVNI_SELECTED);
	while (Index != -1) {
		ListView_GetItemPosition(View, Index, &Loc);
		ListView_SetItemPosition(View, Index, 
			Loc.x + XOffset, Loc.y + YOffset);
		Index = ListView_GetNextItem(View, Index, LVNI_SELECTED);
	}
}

int swiSDAllSets(int Source, SWI_ID RequirementID, int Count, SWI_ID* Table)
{
	int k;
	LVITEM Item;
	HWND View = ListViews[Source];
	if (Count != swiSDCount(Source)) return 0;
	Item.iItem = -1;
	Item.mask = LVIF_PARAM;
	for (k = 0; k < Count; k++) {
		Item.iItem = ListView_GetNextItem(View, Item.iItem, LVNI_ALL);
		ListView_GetItem(View, &Item);
		Table[k] = Item.lParam;
	}
	if (RequirementID) return swiRQOrder(RequirementID, Count, Table);
	return 1;
}

int swiSDSimilarContents(int Source, SWI_ID RequirementID) {
	int Count, Ok;
	SWI_ID* IDs;
	HWND View = ListViews[Source];
	Count = swiSDCount(Source);
	if (!RequirementID) return Count == 0;
	if (!Count) return 0;
	IDs = allocate(Count * sizeof(SWI_ID));
	Ok = swiSDAllSets(Source, RequirementID, Count, IDs);
	deallocate(IDs);
	return Ok;
}
