/********************************************************************

	swi_WindowsControls.c

	Ismo Krkkinen

	Last changed 12.11.1999

********************************************************************/


#include <windows.h>
#include "swi_Internal.h"
#include "swi_WindowsControls.h"
#include <string.h>
#include "memctrl.h"

static CRITICAL_SECTION LBCriticalSection;

void swiWCInitialize() {
	InitializeCriticalSection(&LBCriticalSection);
}

void swiWCFinalize() {
	DeleteCriticalSection(&LBCriticalSection);
}

/* List box */

int swiLBAdd(HWND ListBoxHandle, char* Text, SWI_ID Tag) {
	int Index;
	EnterCriticalSection(&LBCriticalSection);
	Index = SendMessage(ListBoxHandle, LB_ADDSTRING, 0, (LPARAM) Text);
	if ((Index == LB_ERR) || (Index == LB_ERRSPACE)) {
		LeaveCriticalSection(&LBCriticalSection);
		return -1;
	}
	SendMessage(ListBoxHandle, LB_SETITEMDATA, Index, (LPARAM) Tag);
	LeaveCriticalSection(&LBCriticalSection);
	return Index;
}

int swiLBCount(HWND ListBoxHandle) {
	int r;
	EnterCriticalSection(&LBCriticalSection);
	r = SendMessage(ListBoxHandle, LB_GETCOUNT, 0, 0);
	LeaveCriticalSection(&LBCriticalSection);
	return r;
}

void swiLBRemove(HWND ListBoxHandle, int Index) {
	EnterCriticalSection(&LBCriticalSection);
	SendMessage(ListBoxHandle, LB_DELETESTRING, Index, 0);
	LeaveCriticalSection(&LBCriticalSection);
}

void swiLBRemoveAll(HWND ListBoxHandle) {
	EnterCriticalSection(&LBCriticalSection);
	SendMessage(ListBoxHandle, LB_RESETCONTENT, 0, 0);
	LeaveCriticalSection(&LBCriticalSection);
}

void swiLBSetFocusIndex(HWND ListBoxHandle, int Index) {
	if (Index < 0) Index = 0;
	if (Index >= swiLBCount(ListBoxHandle)) 
		Index = swiLBCount(ListBoxHandle) - 1;
	EnterCriticalSection(&LBCriticalSection);
	SendMessage(ListBoxHandle, LB_SETTOPINDEX, Index, 0);
	SendMessage(ListBoxHandle, 
		LB_SETCARETINDEX, Index, MAKELPARAM(FALSE, 0));
	LeaveCriticalSection(&LBCriticalSection);
}

int swiLBFocusIndex(HWND ListBoxHandle) {
	int r;
	EnterCriticalSection(&LBCriticalSection);
	r = SendMessage(ListBoxHandle, LB_GETCARETINDEX, 0, 0);
	LeaveCriticalSection(&LBCriticalSection);
	return r;
}

SWI_ID swiLBTag(HWND ListBoxHandle, int Index) {
	SWI_ID r;
	if ((0 <= Index) && (Index < swiLBCount(ListBoxHandle))) {
		EnterCriticalSection(&LBCriticalSection);
		r = (SWI_ID) SendMessage(ListBoxHandle, LB_GETITEMDATA, Index, 0);
		LeaveCriticalSection(&LBCriticalSection);
		return r;
	}
	return 0;
}

void swiLBSelectSingle(HWND ListBoxHandle, int Index) {
	if (Index < 0) Index = 0;
	if (Index >= swiLBCount(ListBoxHandle))
		Index = swiLBCount(ListBoxHandle) - 1;
	EnterCriticalSection(&LBCriticalSection);
	SendMessage(ListBoxHandle, LB_SETCURSEL, Index, 0);
	LeaveCriticalSection(&LBCriticalSection);
}

int swiLBSelectedCount(HWND ListBoxHandle) {
	int Count;
	EnterCriticalSection(&LBCriticalSection);
	Count = SendMessage(ListBoxHandle, LB_GETSELCOUNT, 0, 0);
	LeaveCriticalSection(&LBCriticalSection);
	if (Count != LB_ERR) return Count;
	return 0;
}

int swiLBSelectedIndices(HWND ListBoxHandle, int** Buffer) {
	int Result, Allocated;
	int Count = swiLBSelectedCount(ListBoxHandle);
	if (Count && !(*Buffer)) {
		*Buffer = allocate(sizeof(int) * Count);
		Allocated = 1;
	} else Allocated = 0;
	if (Count) {
		EnterCriticalSection(&LBCriticalSection);
		Result = SendMessage(ListBoxHandle, 
			LB_GETSELITEMS, Count, (LONG)*Buffer);
		LeaveCriticalSection(&LBCriticalSection);
		if (Result == LB_ERR) {
			if (Allocated) {
				deallocate(*Buffer);
				*Buffer = 0;
			} else **Buffer = -1;
			return 0;
		}
		return Count;
	}
	return 0;
}

/* Combo box (drop down listbox) */

int swiCBAdd(HWND ComboBoxHandle, char* Text, SWI_ID Tag) {
	int Index;
	Index = SendMessage(ComboBoxHandle, CB_ADDSTRING, 0, (LONG)Text);
	if ((Index == CB_ERR) || (Index == CB_ERRSPACE)) {
		return 0;
	}
	SendMessage(ComboBoxHandle, CB_SETITEMDATA, Index, Tag);
	return 1;
}

int swiCBCount(HWND ComboBoxHandle) {
	return SendMessage(ComboBoxHandle, CB_GETCOUNT, 0, 0);
}

void swiCBRemoveAll(HWND ComboBoxHandle) {
	SendMessage(ComboBoxHandle, CB_RESETCONTENT, 0, 0);
}

void swiCBSetSelectedIndex(HWND ComboBoxHandle, int Index) {
	SendMessage(ComboBoxHandle, CB_SETCURSEL, Index, 0);
}

int swiCBSelectedIndex(HWND ComboBoxHandle) {
	return SendMessage(ComboBoxHandle, CB_GETCURSEL, 0, 0);
}

SWI_ID swiCBTag(HWND ComboBoxHandle, int Index) {
	return (SWI_ID) SendMessage(ComboBoxHandle, CB_GETITEMDATA, Index, 0);
}

/* Scroll bar */

int swiSBMove(HWND ScrollBarHandle, int Bar, int ScrollCode,
	int LineOffset, int PageOffset, int ThumbPosition)
{
	SCROLLINFO SI;
	int Offset;
	SI.cbSize = sizeof(SCROLLINFO);
	SI.fMask = SIF_PAGE | SIF_POS | SIF_RANGE;
	GetScrollInfo(ScrollBarHandle, Bar, &SI);
	switch (ScrollCode) {
	case SB_BOTTOM: Offset = SI.nMax - SI.nPos; break;
/*	case SB_LINEUP:*/
	case SB_LINELEFT: Offset = -LineOffset; break;
/*	case SB_LINEDOWN:*/
	case SB_LINERIGHT: Offset = LineOffset; break;
/*	case SB_PAGEUP:*/
	case SB_PAGELEFT: Offset = -PageOffset; break;
/*	case SB_PAGEDOWN:*/
	case SB_PAGERIGHT: Offset = PageOffset; break;
	case SB_THUMBPOSITION: Offset = ThumbPosition - SI.nPos; break;
	case SB_TOP: Offset = -SI.nPos; break;
	default: Offset = 0;
	}
	if (Offset) {
		SI.fMask = SIF_POS;
		SI.nPos += Offset;
		if (SI.nPos < SI.nMin) SI.nPos = SI.nMin;
		else if ((int)SI.nPos > (int)(SI.nMax - SI.nPage + 1))
			SI.nPos = SI.nMax - SI.nPage + 1;
		SetScrollInfo(ScrollBarHandle, Bar, &SI, TRUE);
	}
	return SI.nPos;
}

int swiSBSize(HWND ScrollBarHandle, int Bar, 
	int Minimum, int Maximum, int Page)
{
	SCROLLINFO SI;
	double Location;
	SI.cbSize = sizeof(SCROLLINFO);
	SI.fMask = SIF_PAGE | SIF_POS | SIF_RANGE;
	GetScrollInfo(ScrollBarHandle, Bar, &SI);
	Location = (double) (SI.nPos - SI.nMin) / 
		max(1.0, (double) (SI.nMax - SI.nMin));
	SI.nMin = max(0, Minimum);
	SI.nMax = max(SI.nMin, Maximum);
	if (SI.nMax - SI.nMin < Page) Page = SI.nMax - SI.nMin + 1;
	SI.nPage = max(0, Page);
	SI.nPos = SI.nMin + (int) (Location * (double)(SI.nMax - SI.nMin));
	if (SI.nPos < SI.nMin) SI.nPos = SI.nMin;
	else if ((int)SI.nPos > (int)(SI.nMax - SI.nPage + 1))
		SI.nPos = SI.nMax - SI.nPage + 1;
	SetScrollInfo(ScrollBarHandle, Bar, &SI, TRUE);
	return 1;
}

int swiSBPosition(HWND ScrollBarHandle, int Bar) {
	SCROLLINFO SI;
	SI.cbSize = sizeof(SCROLLINFO);
	SI.fMask = SIF_POS;
	SI.nPos = 0;
	GetScrollInfo(ScrollBarHandle, Bar, &SI);
	return SI.nPos;
}
