/********************************************************************

	swi_WindowsDialogs.c

	Ismo Krkkinen

	Last Changed 13.11.1999

	Covers Windows common dialogs and makes them easier to use.
	If buffer is of inadequate length, it will be empty and the
	pointer returned will point to a string containing the filename.
	It is the responsibility of the user to free this data.

********************************************************************/

#include "swi_WindowsDialogs.h"
#include <windows.h>
#include "swi_Filter.h"
#include "memctrl.h"
#include <memory.h>

static OPENFILENAME OFN;
static char Filename[2048];

static void wdSetOFN(SWI_ID FilterID) {
	OFN.lStructSize = sizeof(OPENFILENAME);
	OFN.hwndOwner = swiMainWindowHandle();
	OFN.hInstance = 0;
	OFN.lpstrFilter = swiFLFilterString(FilterID);
	OFN.lpstrCustomFilter = 0;
	OFN.nMaxCustFilter = 0;
	OFN.nFilterIndex = 1;
	Filename[0] = 0;
	OFN.lpstrFile = Filename;
	OFN.nMaxFile = 2048;
	OFN.lpstrFileTitle = 0;
	OFN.nMaxFileTitle = 0;
	OFN.lpstrInitialDir = 0;
	OFN.lpstrTitle = 0;
	OFN.Flags = OFN_EXPLORER | OFN_HIDEREADONLY;
	OFN.nFileOffset = 0;
	OFN.nFileExtension = 0;
	OFN.lpstrDefExt = swiFLDefaultExtension(FilterID);
	OFN.lCustData = 0;
	OFN.lpfnHook = 0;
	OFN.lpTemplateName = 0;
}

static char* wdCopyFilename(char* BufferAddr, int BufferLength) {
	char* Result;
	if (OFN.nMaxFile > 0) {
		if (BufferAddr && (BufferLength >= (int)OFN.nMaxFile)) {
			strcpy(BufferAddr, Filename);
			return BufferAddr;
		} else {
			Result = allocate(OFN.nMaxFile + 1);
			strcpy(Result, Filename);
			return Result;
		}
	} else return BufferAddr;
} 

char* swiOpenFilename(char* BufferAddr, int BufferLength, SWI_ID FilterID, 
	SWI_ID* FilterOwnerID, int* OpenFilterIndex)
{
	char* FilterString;
	int k;
	wdSetOFN(FilterID);
	OFN.Flags |= OFN_FILEMUSTEXIST;
	OFN.nFilterIndex = *OpenFilterIndex;
	if (GetOpenFileName(&OFN) == TRUE) {
		FilterString = (char*)(OFN.lpstrFilter);
		*OpenFilterIndex = OFN.nFilterIndex;
		for (k = 1; k < (int)OFN.nFilterIndex; k++) {
			FilterString += strlen(FilterString) + 1;
			FilterString += strlen(FilterString) + 1;
		}
		*FilterOwnerID = swiFLFindOwnerOfFilterWithString(FilterString);
		return wdCopyFilename(BufferAddr, BufferLength);
	}
	return 0;
}

char* swiSaveFilename(char* BufferAddr, int BufferLength, SWI_ID FilterID,
	int* SelectedFilterIndex)
{
	wdSetOFN(FilterID);
	OFN.Flags |= OFN_OVERWRITEPROMPT;
	if (BufferAddr) strcpy(Filename, BufferAddr);
	if (GetSaveFileName(&OFN) == TRUE) {
		if (SelectedFilterIndex) *SelectedFilterIndex = OFN.nFilterIndex;
		return wdCopyFilename(BufferAddr, BufferLength);
	}
	if (SelectedFilterIndex) *SelectedFilterIndex = 0;
	return 0;
}
