/********************************************************************

	swi_WindowsMenus.c

	Ismo Krkkinen

	Last changed 12.11.1999

********************************************************************/

#include "swi_WindowsMenus.h"
#include <string.h>
#include "memctrl.h"
#include "swi_Storage.h"
#include "swi_Process.h"

typedef struct {
	SWI_ID ItemID;
	int Command;
} MenuItemData;

#define SWI_MENU_MAXCOUNT 1024
static int Count = 0;
static MenuItemData Items[SWI_MENU_MAXCOUNT];

void swiWMAdd(SWI_ID ItemToAdd, SWI_ID AddAfter, int Command) {
	int k, Place;
	if (Count >= SWI_MENU_MAXCOUNT) return;
	Place = 0;
	if (AddAfter) {
		while ((Place < Count) && (Items[Place].ItemID != AddAfter)) 
			Place++;
		Place++;
	}
	for (k = Count; k > Place; k--) {
		Items[k].ItemID = Items[k - 1].ItemID;
		Items[k].Command = Items[k - 1].Command;
	}
	Items[Place].ItemID = ItemToAdd;
	Items[Place].Command = Command;
	Count++;
}

static char KeyNameBuffer[32];

static char* KeyText(int Code) {
	if (0x30 <= Code && Code <= 0x39) {
		/* VK_0 thru VK_9 are the same as ASCII '0' thru '9' 
		(0x30 - 0x39) */
		KeyNameBuffer[0] = (char) Code;
		KeyNameBuffer[1] = 0;
	} else if (0x41 <= Code && Code <= 0x5A) {
		/* VK_A thru VK_Z are the same as ASCII 'A' thru 'Z' 
		(0x41 - 0x5A) */
		KeyNameBuffer[0] = (char) Code;
		KeyNameBuffer[1] = 0;
	} else switch (Code) {
	case VK_NUMPAD0: strcpy(KeyNameBuffer, "Numpad 0"); break;
	case VK_NUMPAD1: strcpy(KeyNameBuffer, "Numpad 1"); break;
	case VK_NUMPAD2: strcpy(KeyNameBuffer, "Numpad 2"); break;
	case VK_NUMPAD3: strcpy(KeyNameBuffer, "Numpad 3"); break;
	case VK_NUMPAD4: strcpy(KeyNameBuffer, "Numpad 4"); break;
	case VK_NUMPAD5: strcpy(KeyNameBuffer, "Numpad 5"); break;
	case VK_NUMPAD6: strcpy(KeyNameBuffer, "Numpad 6"); break;
	case VK_NUMPAD7: strcpy(KeyNameBuffer, "Numpad 7"); break;
	case VK_NUMPAD8: strcpy(KeyNameBuffer, "Numpad 8"); break;
	case VK_NUMPAD9: strcpy(KeyNameBuffer, "Numpad 9"); break;
	case VK_MULTIPLY: strcpy(KeyNameBuffer, "Numpad *"); break;
	case VK_ADD: strcpy(KeyNameBuffer, "Numpad +"); break;
	case VK_SUBTRACT: strcpy(KeyNameBuffer, "Numpad -"); break;
	case VK_DIVIDE: strcpy(KeyNameBuffer, "Numpad /"); break;
	case VK_F1: strcpy(KeyNameBuffer, "F1"); break;
	case VK_F2: strcpy(KeyNameBuffer, "F2"); break;
	case VK_F3: strcpy(KeyNameBuffer, "F3"); break;
	case VK_F4: strcpy(KeyNameBuffer, "F4"); break;
	case VK_F5: strcpy(KeyNameBuffer, "F5"); break;
	case VK_F6: strcpy(KeyNameBuffer, "F6"); break;
	case VK_F7: strcpy(KeyNameBuffer, "F7"); break;
	case VK_F8: strcpy(KeyNameBuffer, "F8"); break;
	case VK_F9: strcpy(KeyNameBuffer, "F9"); break;
	case VK_F10: strcpy(KeyNameBuffer, "F10"); break;
	case VK_F11: strcpy(KeyNameBuffer, "F11"); break;
	case VK_F12: strcpy(KeyNameBuffer, "F12"); break;
	case VK_F13: strcpy(KeyNameBuffer, "F13"); break;
	case VK_F14: strcpy(KeyNameBuffer, "F14"); break;
	case VK_F15: strcpy(KeyNameBuffer, "F15"); break;
	case VK_F16: strcpy(KeyNameBuffer, "F16"); break;
	case VK_F17: strcpy(KeyNameBuffer, "F17"); break;
	case VK_F18: strcpy(KeyNameBuffer, "F18"); break;
	case VK_F19: strcpy(KeyNameBuffer, "F19"); break;
	case VK_F20: strcpy(KeyNameBuffer, "F20"); break;
	case VK_F21: strcpy(KeyNameBuffer, "F21"); break;
	case VK_F22: strcpy(KeyNameBuffer, "F22"); break;
	case VK_F23: strcpy(KeyNameBuffer, "F23"); break;
	case VK_F24: strcpy(KeyNameBuffer, "F24"); break;
	}
	return KeyNameBuffer;
}

static SWI_ID ItemType(SWI_ID ItemID) {
	SWIDataTypes DT = swiSType(ItemID);
	if (DT == PROCESS_CLASS) return swiPCProcessType(ItemID);
	return ItemID;
}

static void AddToMenu(HMENU Menu, SWI_ID ProcessType, 
	int MenuPosition, int* Position)
{
	char* foo, *bar;
	char Key[2];
	SWI_ID Type;
	MENUITEMINFO M;
	M.cbSize = sizeof(MENUITEMINFO);
	M.fType = MFT_STRING;
	M.fState = MFS_ENABLED;
	M.hbmpChecked = 0;
	M.hbmpUnchecked = 0;
	Type = ItemType(Items[*Position].ItemID);
	do {
		M.fMask = MIIM_DATA | MIIM_ID | MIIM_STATE | MIIM_TYPE;
		if ((Type != ProcessType) && (Type == Items[*Position].ItemID) 
			&& (swiPTParent(Type) == ProcessType))
		{
			/* make a submenu */
			M.fMask |= MIIM_SUBMENU;
			M.wID = 0;
			M.hSubMenu = CreatePopupMenu();
			(*Position)++;
			AddToMenu(M.hSubMenu, Type, 0, Position);
			M.dwItemData = 0;
			M.dwTypeData = swiPTDescription(Type);
			M.cch = strlen(swiPTDescription(Type));
			InsertMenuItem(Menu, MenuPosition, TRUE, &M);
		} else if (Type == ProcessType) {
			/* add item to this menu */
			M.wID = Items[*Position].Command;
			M.hSubMenu = 0;
			M.dwItemData = Items[*Position].ItemID;
			if (swiPCHasAccelerator(Items[*Position].ItemID)) {
				foo = AppendString(
					swiPCDescription(Items[*Position].ItemID), "\t");
				if (swiPCUsesCtrl(Items[*Position].ItemID)) {
					bar = AppendString(foo, "Ctrl+");
					deallocate(foo);
				} else bar = foo;
				if (swiPCUsesAlt(Items[*Position].ItemID)) {
					foo = AppendString(bar, "Alt+");
					deallocate(bar);
				} else foo = bar;
				if (swiPCUsesShift(Items[*Position].ItemID)) {
					bar = AppendString(foo, "Shift+");
					deallocate(foo);
				} else bar = foo;
				if (swiPCUsesVirtual(Items[*Position].ItemID)) {
					foo = AppendString(bar, 
						KeyText(swiPCUsesKey(Items[*Position].ItemID)));
					deallocate(bar);
				} else {
					Key[1] = 0;
					Key[0] = (char) swiPCUsesKey(Items[*Position].ItemID);
					foo = AppendString(bar, Key);
					deallocate(bar);
				}
				M.dwTypeData = foo;
			} else {
				M.dwTypeData = swiPCDescription(Items[*Position].ItemID);
				foo = 0;
			}
			M.cch = strlen((char*) M.dwTypeData);
			(*Position)++;
			InsertMenuItem(Menu, MenuPosition, TRUE, &M);
			if (foo) deallocate(foo);
		}
		MenuPosition++;
		if (*Position < Count) Type = ItemType(Items[*Position].ItemID);
	} while ((*Position < Count) && ((Type == ProcessType) 
		|| ((Type == Items[*Position].ItemID) 
		&& (swiPTParent(Type) == ProcessType))));
}


void swiWMAppendToMenu(HMENU Parent) {
	int Position = 0;
	if (Count) 
		AddToMenu(Parent, 0, GetMenuItemCount(Parent) + 1, &Position);
}
