/*--------------------------------------------------------------------*/
/* INTERFC.C	   Ismo Krkkinen  								  */
/*																	  */
/* This module provides User interface functions.                     */
/* Modified for use with SWI and nothing else.                        */
/*																	  */
/*--------------------------------------------------------------------*/

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <interfc.h>

#include "swi.h"
#include "swi_OutputFlow.h"
#include "swi_Process.h"
#include <windows.h>
#include <process.h>


/* --------------------------------------------------------------------*/


static int   __ShowProgress    = 1;   /* Controlled by ProgressVisibility */
static int   __ShowProgressBar = 1;   /* controlled by FreezeProgrssBar   */
static int   __ProgressCreated = 0;

static SWI_ID OutputFlow = 0;

/* ---------------------- INFORMATION OUTPUT --------------------------*/


int PrintMessage(char *format, ...)
{
	va_list ap;             /* points to each unnamed arg in turn */
	int retval=0;
	char msg[255];
	
	va_start(ap, format);   /* make ap point to 1st unnamed arg */
	vsprintf(msg, format, ap);
	va_end(ap);             /* clean up when done */
    if (!OutputFlow) OpenInfoFlow(0);
    swiOFAppend(OutputFlow, msg);
    retval = strlen(msg);
    return(retval);
}


/* --------------------------------------------------------------------*/


int PrintProgramMessage(char *format, ...)
{
	va_list ap;             /* points to each unnamed arg in turn */
	int retval=0;
	char msg[255];
	
	va_start(ap, format);   /* make ap point to 1st unnamed arg */
	vsprintf(msg, format, ap);
	va_end(ap);             /* clean up when done */
	if (!OutputFlow) OpenInfoFlow(0);
	swiOFAppend(OutputFlow, msg);
	retval = strlen(msg);
    return(retval);
	
}


/* --------------------------------------------------------------------*/


int ErrorMessage(char *format, ...)
{
	va_list	ap;		/* points to each unnamed arg in turn */
	char msg[1024];
	
	va_start(ap, format);	/* make ap point to 1st unnamed arg */
	vsprintf(msg, format, ap);
	va_end(ap);		/* clean up when done */
	swiErrorMessageBox(msg);
	return strlen(msg);
}


/* --------------------------------------------------------------------*/


int OpenInfoFlow(int id)
{
	/* GUI:     everything goes to "Process Output" -window   */
	if (OutputFlow) swiOFClear(OutputFlow);
	else OutputFlow = swiOFNew();
	return(0);
}


/* --------------------------------------------------------------------*/


int CloseInfoFlow(void)
{
	if (OutputFlow) swiOFFlush(OutputFlow);
	return(0);
}


/* --------------------- PROGRAM TERMINATION --------------------------*/

int ExitRequestHonoredStatus = 0;

/* returns non-zero, if program calls ExitRequested, zero otherwise */
int ExitRequestHonored() {
	return ExitRequestHonoredStatus;
}


/* --------------------------------------------------------------------*/


/* tells UI if program calls ExitRequested, default is not honored */
void SetExitRequestHonored(int Honored) {
	ExitRequestHonoredStatus = Honored;
}


/* --------------------------------------------------------------------*/


/* checks if stopping of process has been requested, returns non-zero if yes */
int ExitRequested()
{
	return swiProcessExitRequested();
}


/* --------------------------------------------------------------------*/


/* exits the program or thread in a proper manner */
void ExitProcessing(int ReturnValue)
{
	if (GetCurrentThreadId() == ThreadHandle) {
		swiPRExitValue(swiRunningProcess(), ReturnValue);
		PostMessage(swiMainWindowHandle(), SWI_WM_THREAD_FINISHED, ReturnValue, 0);
	}
	_endthread();
}



/* ---------------------- PROGRESS INDICATOR --------------------------*/


/* iak: I presume this will be called before progress bar is created */
int SetProgressVisibility(int value)
{
	__ShowProgress = value;
	return(value);
}


/* --------------------------------------------------------------------*/


int OpenProgressWindow(void)
{
	if (__ShowProgress && !__ProgressCreated)
    {
		__ProgressCreated = 1;
        if (OutputFlow) swiOFClearProgress(OutputFlow);
    }
	return(0);
}


/* --------------------------------------------------------------------*/


int CloseProgressWindow(void)
{
	if (__ShowProgress || __ProgressCreated) {
		if (OutputFlow) swiOFClearProgress(OutputFlow);
	}
	__ProgressCreated = 0;
	return(0);
}


/* --------------------------------------------------------------------*/


int UpdateProgressBar(int state, int total)
{
	if (__ShowProgress && __ShowProgressBar)
    {
		if (OutputFlow) swiOFSetProgress(OutputFlow, state, total);
		else return 1;
    }
	return(0);
}


/* --------------------------------------------------------------------*/


int UpdateFancyProgressBar(int state, int total)
{
	static int x=0;
	
	if (__ShowProgress && __ShowProgressBar)
    {
		/* iak: this is just about as fancy as I care */
		if (OutputFlow) swiOFSetProgress(OutputFlow, state, total);
		else return 1;
    }
	return(0);
}


/* --------------------------------------------------------------------*/


int UpdateProgressLabel(char *format, ...)
{
	va_list ap;             /* points to each unnamed arg in turn */
	int retval=0;
	char msg[255];
	
	if (__ProgressCreated)
    {
		
        va_start(ap, format);   /* make ap point to 1st unnamed arg */
        vsprintf(msg, format, ap);
        va_end(ap);             /* clean up when done */
        if (OutputFlow) swiOFLabelProgress(OutputFlow, msg);
        else return 0;
        retval = strlen(msg);
    }
	return(retval);
}


/* --------------------------------------------------------------------*/


int FreezeProgressBar(value)
{
	__ShowProgressBar = !value;
	return(value);
}


/*                                                                    */
/* ============================= END ================================ */
